/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L1.AtmsFileFilter.java
*
* NAME
*   atms.L1.AtmsFileFilter.java
*
* PURPOSE
*   Match ATMS file names based on naming patterns. 
* 
* DESCRIPTION
*	This class is used to match ATMS geolocation and TDR file names
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package atms.L1;


import java.io.File;
import java.io.FilenameFilter;
import atms.etc.JPSSConstant;

/**
 * This class defines file filters to match ATMS geolocation and TDR files with specified time tag.
 * @author Xianjun Hao, George Mason University
 */
public class AtmsFileFilter implements FilenameFilter {

    final public static int ATMS_FILE_TYPE_GEO = 0;
    final public static int ATMS_FILE_TYPE_TDR = 1;

    private int atmsFileType = 0;
    private String timeTag = "";
    private String filePatt = "";

    /**
     * CONSTRUCTOR.
     * @param filePatt, file pattern.
     */
    public AtmsFileFilter(String filePatt) {
        this.filePatt = filePatt;
    }

    /**
     * Constructor. 
     * @param timeTag, time tag for ATMS geolocation and TDR files.
     * @param atmsFileType, file type (0 for GATMO and 1 for TATMS)
     * @param satName, satellite name 
     */
    public AtmsFileFilter(String timeTag, int atmsFileType, String satName) {
        String filePrefix = "";
        int index = JPSSConstant.getJpssSatelliteIndex(satName);
        String fileTag = JPSSConstant.getJpssFileTag(index);
        
        this.timeTag = timeTag;
        this.atmsFileType = atmsFileType;
        if (this.atmsFileType == ATMS_FILE_TYPE_GEO) {
            filePrefix = "GATMO_" + fileTag +"_";
        }
        if (this.atmsFileType == ATMS_FILE_TYPE_TDR) {
            filePrefix = "TATMS_" + fileTag +"_";
        }

        this.filePatt = String.format("%s%s", filePrefix, timeTag);
    }

    /**
     * Check if a file matches the specified filter.
     * @param dir, file directory.
     * @param name, file name.
     * @return 
     */
    public boolean accept(File dir, String name) {
        boolean ret = false;
        int len = filePatt.length();

        if (name.endsWith(".h5")) {

            if (name.length() <= len) {
                ret = false;
            } else {
                String fileStr = name.substring(0, len);
                if (fileStr.equals(filePatt)) {
                    ret = true;
                } else {
                    ret = false;
                }
            }

        } else {
            ret = false;
        }
        return ret;
    }
    
}
