/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L1.AtmsLimbCorrector.java
*
* NAME
*   atms.L1.AtmsLimbCorrector.java
*
* PURPOSE
*   Conduct limb adjustment for ATMS TDR measurements. 
* 
* DESCRIPTION
*	This class is used for limb-adjustment of ATMS TDR measurements.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package atms.L1;


import ucar.ma2.ArrayByte;
import ucar.ma2.ArrayDouble;
import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import atms.etc.*;
import atms.util.OSValidator;

/**
 * This class does limb correction for ATMS TDR data
 * @author Xianjun Hao, George Mason University
 */
public class AtmsLimbCorrector {
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    /**
     * Name of the limb correction LUT file 
     */
    private String filenameLUT=null;
    
    /**
     * Name of the land/sea mask file
     */
    private String filenameMask=null;
    
    /**
     * Number of rows of mask data (lon)
     */
    final private int nrow = 7200;
    
    /**
     * Number of columns of mask data (lat)
     */
    final private int ncol = 3600;

    /**
     *  Number of channels in LUT
     */
    final private int nchannel = 22;
    
    /**
     * Number of FOVs in each scan
     */
    final private int nfov = 96; 
    
    /**
     * mean temperature over land in the limb-correction look-up table
     */
    private double[] btm_land = null;
    
    /**
     * mean temperature over ocean in the limb-correction look-up table
     */
    private double[] btm_sea = null;
    
    /** 
     * limb correction values over land, in dimensions: ass_channel, scan_position, channel
     */
    private double[][][] btm_ass_land = null;
    
    /** 
     * limb correction values over ocean, in dimensions: ass_channel, scan_position, channel
     */
    private double[][][] btm_ass_sea = null;

    /** 
     * limb correction coefficients over land, in dimensions: ass_channel, scan_position, channel
     */
    private double[][][] coef_ass_land = null;
    
    /** 
     * limb correction coefficients over ocean, in dimensions: ass_channel, scan_position, channel
     */
    private double[][][] coef_ass_sea = null;
    
    /**
     * Land/sea mask
     */
    private byte[][] landseaMask = null;
    
    /**
     * Flag to indicate is LUT table is loaded correctly.
     */
    private boolean isLutReady = true;
        
    /**
     * Flag to indicate is Land/sea mask data is loaded correctly.
     */
    private boolean isMaskReady = true;
    
    /**
     * Flag to indicate it it's running in debug mode.
     */
    private boolean isDebugMode = false;

    /**
     * Constructor. 
     */
    public AtmsLimbCorrector() {
        if (OSValidator.isWindows()) {
            this.filenameLUT = ".\\etc\\atms_limbcorr_lut.nc";
            this.filenameMask = ".\\etc\\landsea_mask_05.nc";
        } else {
            this.filenameLUT = "./etc/atms_limbcorr_lut.nc";
            this.filenameMask = "./etc/landsea_mask_05.nc";            
        }
        this.init();
        this.loadLut();
        this.loadLandseaMask();
    }
    
    /**
     * Constructor
     * @param filenameLUT, name of LUT file for limb adjustment.
     * @param filenameMask, name of Land/Sea mask file. 
     */
    public AtmsLimbCorrector(String filenameLUT, String filenameMask) {
        this.filenameLUT = filenameLUT;
        this.filenameMask = filenameMask;        
        this.init();
        this.loadLut();
        this.loadLandseaMask();
    }
    
    /**
     * Initialization
     */
    private void init(){
        this.isLutReady = true;
        this.isMaskReady = true;
        this.btm_land = new double[this.nchannel];
        this.btm_sea = new double[this.nchannel];
        this.btm_ass_land = new double[this.nchannel][this.nfov][this.nchannel];
        this.btm_ass_sea = new double[this.nchannel][this.nfov][this.nchannel];
        this.coef_ass_land = new double[this.nchannel][this.nfov][this.nchannel];
        this.coef_ass_sea = new double[this.nchannel][this.nfov][this.nchannel];
        this.landseaMask = new byte[this.nrow][this.ncol];
    }
    
    /**
     * Load LUT for limb correction
     */
    private void loadLut(){
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayDouble.D1 darr1 = null;
        ArrayDouble.D3 darr3 = null;
        int[] shape = new int[3];
        int[] origin = new int[3];
        int dim1, dim2, dim3;
        int i, j, k;
        
        for(i=0; i<3; i++){
            origin[i] = 0;
        }
        
        try {
            fileNc = NetcdfFile.open(this.filenameLUT);

            // Read btm_land
            dataVar = fileNc.findVariable(AtmsLimbConstant.LUT_VARNAME_BTM_LAND);
            if (dataVar==null) {
                this.isLutReady = false;
                errLogger.error("Failed to read btm_land in LUT for limb correction.");
            } else {
                shape = dataVar.getShape();
                darr1 = (ArrayDouble.D1) dataVar.read(origin, shape);
                dim1 = shape[0];
                for(i=0; i<dim1; i++){
                    this.btm_land[i] = (double)darr1.get(i);
                }
            }

            // Read btm_sea
            dataVar = fileNc.findVariable(AtmsLimbConstant.LUT_VARNAME_BTM_SEA);
            if (dataVar==null) {
                this.isLutReady = false;
                errLogger.error("Failed to read btm_sea in LUT for limb correction.");
            } else {
                shape = dataVar.getShape();
                darr1 = (ArrayDouble.D1) dataVar.read(origin, shape);
                dim1 = shape[0];
                for(i=0; i<dim1; i++){
                    this.btm_sea[i] = (double)darr1.get(i);
                }
            }

            // Read btm_ass_land
            dataVar = fileNc.findVariable(AtmsLimbConstant.LUT_VARNAME_BTM_ASS_LAND);
            if (dataVar==null) {
                this.isLutReady = false;
                errLogger.error("Failed to read btm_ass_land in LUT for limb correction.");
            } else {
                shape = dataVar.getShape();
                darr3 = (ArrayDouble.D3) dataVar.read(origin, shape);
                dim1 = shape[0];
                dim2 = shape[1];
                dim3 = shape[2];
                for(i=0; i<dim1; i++){
                    for(j=0; j<dim2; j++){
                        for(k=0; k<dim3; k++){
                            this.btm_ass_land[i][j][k] = (double)darr3.get(i,j,k);
                        }
                    }
                }
            }

            // Read btm_ass_sea
            dataVar = fileNc.findVariable(AtmsLimbConstant.LUT_VARNAME_BTM_ASS_SEA);
            if (dataVar==null) {
                this.isLutReady = false;
                errLogger.error("Failed to read btm_ass_sea in LUT for limb correction.");
            } else {
                shape = dataVar.getShape();
                darr3 = (ArrayDouble.D3) dataVar.read(origin, shape);
                dim1 = shape[0];
                dim2 = shape[1];
                dim3 = shape[2];
                for(i=0; i<dim1; i++){
                    for(j=0; j<dim2; j++){
                        for(k=0; k<dim3; k++){
                            this.btm_ass_sea[i][j][k] = (double)darr3.get(i,j,k);
                        }
                    }
                }
            }

            // Read coef_limbcorr_land
            dataVar = fileNc.findVariable(AtmsLimbConstant.LUT_VARNAME_COEF_LIMBCORR_LAND);
            if (dataVar==null) {
                this.isLutReady = false;
                errLogger.error("Failed to read coef_limbcorr_land in LUT for limb correction.");
            } else {
                shape = dataVar.getShape();
                darr3 = (ArrayDouble.D3) dataVar.read(origin, shape);
                dim1 = shape[0];
                dim2 = shape[1];
                dim3 = shape[2];
                for(i=0; i<dim1; i++){
                    for(j=0; j<dim2; j++){
                        for(k=0; k<dim3; k++){
                            this.coef_ass_land[i][j][k] = (double)darr3.get(i,j,k);
                        }
                    }
                }
            }
            
            // Read coef_limbcorr_sea
            dataVar = fileNc.findVariable(AtmsLimbConstant.LUT_VARNAME_COEF_LIMBCORR_SEA);
            if (dataVar==null) {
                this.isLutReady = false;
                errLogger.error("Failed to read coef_limbcorr_sea in LUT for limb correction.");
            } else {
                shape = dataVar.getShape();
                darr3 = (ArrayDouble.D3) dataVar.read(origin, shape);
                dim1 = shape[0];
                dim2 = shape[1];
                dim3 = shape[2];
                for(i=0; i<dim1; i++){
                    for(j=0; j<dim2; j++){
                        for(k=0; k<dim3; k++){
                            this.coef_ass_sea[i][j][k] = (double)darr3.get(i,j,k);
                        }
                    }
                }
            }            
            
        }catch (Exception e) {
            this.isLutReady = false;
            e.printStackTrace();
            errLogger.error("Unable to load LUT for limb correction, please check LUT files.");
        }                
    }
    
    /**
     * set Debug mode
     * @param isDebugMode 
     */
    public void setDebugMode(boolean isDebugMode) {
        this.isDebugMode = isDebugMode;
    }
    
    
    /**
     * Load Land/Sea Mask
     */
    private void loadLandseaMask(){
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayByte.D2 barr2 = null;
        int[] shape = new int[2];
        int[] origin = new int[2];
        int dim1, dim2;
        int i, j;
        
        for(i=0; i<2; i++){
            origin[i] = 0;
        }
        
        try {
            fileNc = NetcdfFile.open(this.filenameMask);

            // Read land/sea mask
            dataVar = fileNc.findVariable(AtmsLimbConstant.LUT_VARNAME_LAND_SEA_MASK);
            if (dataVar==null) {
                this.isMaskReady = false;
                errLogger.error("Failed to read land/sea mask  for limb correction.");
            } else {
                shape = dataVar.getShape();
              //  System.out.println("Mask shape:" +shape[0]+","+shape[1]);
                barr2 = (ArrayByte.D2) dataVar.read(origin, shape);                
            }
        } catch (Exception e) {
            e.printStackTrace();
            this.isMaskReady = false;
            errLogger.error("Unable to read land/sea mask for limb correction");
        }                
    }
    
    /**
     * Do limb correction for ATMS TDR data
     * @param lat, latitude of TDR data
     * @param lon, longitude of TDR data
     * @param bt, brightness temperature of TDR dataS
     * @return 0, if success; -1, if failed 
     */
    public double[][][] doLimbCorrection(double[][] lat, double[][] lon, double[][][] bt){
        int mrow = lat.length;     //22
        int mcol = lat[0].length;  // 96, scan position
        int mchannel = AtmsLimbConstant.ATMS_NUMBER_CHANNELS;
        double[][][] btn = new double[mrow][mcol][mchannel];
        double sum = 0.0d;
        
        // bt dimension: (mrow, mcol, mchannel)
        int i, j, k, k1, k2;
        int count_invalid = 0;
        double tmp_coef = 0.0d;
                
        // get land sea mask
        boolean[][] island = this.isLand(lat, lon);
        
        for(i=0; i<mrow; i++) {  // along track direction
            for(j=0; j<mcol; j++) { // mcol=96, cross-track direction
                
                if (island[i][j]) {   //land
                   for(k=0; k<mchannel; k++) { // for each channel
                        sum = 0.0d;
                        count_invalid = 0;
                        for(k1=0; k1<mchannel; k1++){ // associated channels
                            tmp_coef = this.coef_ass_land[k1][j][k];
                            if((bt[i][j][k1]>=AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE)& (bt[i][j][k1]<AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE)) {
                                sum = sum + tmp_coef*(bt[i][j][k1] - this.btm_ass_land[k1][j][k]);
                            } else {
                                if (Math.abs(tmp_coef)>0) {
                                    count_invalid = count_invalid + 1;
                                }
                            }
                        }
                        if (count_invalid > 0) { // Either this channel, or associated channels have invalid value, 
                                                 // limb correction fails, fill with invalid value
                            btn[i][j][k] = AtmsGridConstant.FILLED_VALUE_BRIGHTNESS_TEMPERATURE;
                        } else {
                            btn[i][j][k] = sum + this.btm_land[k];                            
                        }
                    }                        
                } else {    // sea
                   for(k=0; k<mchannel; k++) {
                        sum = 0.0d;
                        count_invalid = 0;
                        for(k1=0; k1<mchannel; k1++){ // associated channels
                            tmp_coef = this.coef_ass_sea[k1][j][k];
                            if((bt[i][j][k1]>=AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE)& (bt[i][j][k1]<AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE)) {
                                sum = sum + tmp_coef*(bt[i][j][k1] - this.btm_ass_sea[k1][j][k]);
                            } else {
                                if (Math.abs(tmp_coef)>0) {
                                    count_invalid = count_invalid + 1;
                                }
                            }
                        }
                        if (count_invalid > 0) { // Either this channel, or associated channels have invalid value, 
                                                 // limb correction fails, fill with invalid value
                            btn[i][j][k] = AtmsGridConstant.FILLED_VALUE_BRIGHTNESS_TEMPERATURE;
                        } else {
                            btn[i][j][k] = sum + this.btm_sea[k];                            
                        }  
                    }                        
                }
                
            }       
        }
                
        return btn;
    }
    
    /**
     * Check if the specified locations are land or not. 
     * @param lat, array of latitude.
     * @param lon, array of longitude
     * @return land mask, true for land, false for ocean. 
     */
    private boolean[][] isLand(double[][] lat, double[][] lon) {
        int mrow = lat.length;
        int mcol = lat[0].length;
        int i, j, k1, k2;
        boolean[][] ret = new boolean[mrow][mcol];
        double pland = 0.0d;
        
        double res = AtmsLimbConstant.ATMS_RES_LAND_SEA_FRACTION;
        double rratio = res/AtmsLimbConstant.EARTH_RADIUS;
        double rad2deg = 180.0d/Math.PI;
        double deg2rad = Math.PI/180.0d;
        double resolat = 1.0d/(double)AtmsLimbConstant.LAND_SEA_MASK_NRLA;
        double resolon = 1.0d/(double)AtmsLimbConstant.LAND_SEA_MASK_NRLO;
        double slope = AtmsLimbConstant.LAND_SEA_MASK_SLOPE;
        double threshold = AtmsLimbConstant.THRESHOLD_LAND_SEA_FRACTION;
        
        double dlon, dlat, tlon, tlat, lon2, lat2;
        int nlat, nlon, klat, klon;
        
        
        for(i=0; i<mrow; i++){
            for(j=0; j<mcol; j++){
                if (lon[i][j]>180.0d){
                    tlon = lon[i][j];
                } else {
                    tlon = lon[i][j] - 360.0d;
                }
                tlat=lat[i][j];
                dlon = (rratio/Math.cos(tlat*deg2rad))*rad2deg; // delta lon
                dlat = rratio*rad2deg; // delta lat
                nlat = (int)Math.floor(dlat/resolat);
                nlon = (int)Math.floor(dlon/resolon);
                pland = 0.0d;
                for(k1=0; k1<nlon; k1++) {
                    lon2 = tlon - (dlon/2.0d)+ k1*resolon;
                    klon = (int)Math.floor(slope*(lon2+180.0d));
                    if (klon<0) {
                        klon =0;
                    }
                    if(klon > (this.nrow -1)) {
                        klon = this.nrow - 1 ;
                    }
                    
                    for (k2=0; k2<nlat; k2++){
                        lat2 = tlat - (dlat/2.0d) + k2*resolat;
                        klat = (int)Math.floor(slope*(90.0d-lat2));
                        if (klat<0) {
                            klat  = 0;
                        }
                        if (klat > (this.ncol -1 )) {
                            klat = this.ncol - 1;
                        }    
                        
                        if (this.landseaMask[klon][klat]>0){
                            pland = pland + 1.0d;
                        }                        
                    }  // end k2
                } // end k1
                pland = pland / ((double)nlat) /((double)nlon);
                if (pland>= threshold){
                    ret[i][j] = true;
                } else {
                    ret[i][j] = false;
                }
            }
        }                
        return ret;
    }
            
}
