/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L1.AtmsNamingConvention.java
*
* NAME
*   atms.L1.AtmsNamingConvention.java
*
* PURPOSE
*   Defined naming conventions for ATMS TDR files. 
* 
* DESCRIPTION
*	This class is used for naming convention of ATMS data files.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package atms.L1;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.List;
import java.util.ArrayList;

import atms.util.*;

/**
 * ATMS file naming convention. 
 *
 * @author Xianjun Hao, George Mason University 
 */
public class AtmsNamingConvention {

    /**
     * file prefix for Geolocation file
     */
    final public static String FILE_NAMING_PREFIX_GEO = "GATMO_";

    /**
     * file prefix for SDR file
     */
    final public static String FILE_NAMING_PREFIX_SDR = "SATMS_";

    /**
     * file prefix for TDR file
     */
    final public static String FILE_NAMING_PREFIX_TDR = "TATMS_";

    /**
     * Length of time tag
     */
    final public static int FILE_NAMING_LENGTH_TIMETAG = 27;

    /**
     * Index of year in file name
     */
    final public static int FILE_NAMING_INDEX_YEAR = 11;

    /**
     * Index of month in file name
     */
    final public static int FILE_NAMING_INDEX_MONTH = 15;

    /**
     * Index of day in file name
     */
    final public static int FILE_NAMING_INDEX_DAY = 17;

    /**
     * Index of start hour in file name
     */
    final public static int FILE_NAMING_INDEX_HOUR_START = 21;

    /**
     * Index of start minute in file name
     */
    final public static int FILE_NAMING_INDEX_MINUTE_START = 23;

    /**
     * Index of start second in file name
     */
    final public static int FILE_NAMING_INDEX_SECOND_START = 25;

    /**
     * Index of end hour in file name
     */
    final public static int FILE_NAMING_INDEX_HOUR_END = 30;

    /**
     * Index of end minute in file name
     */
    final public static int FILE_NAMING_INDEX_MINUTE_END = 32;

    /**
     * Index of end second in file name
     */
    final public static int FILE_NAMING_INDEX_SECOND_END = 34;

    /**
     * Index of orbit ID in file name
     */
    final public static int FILE_NAMING_INDEX_ORBIT = 40;

    /**
     * Get time tag of ATMS TDR file
     *
     * @param tdrFileName name of ATMS TDR file
     * @return time tag of ATMS TDR file
     */
    public static String getAtmsTdrTimeTag(String tdrFileName) {
        String fileTag = null;
        String shortName = AtmsNamingConvention.getFileShortname(tdrFileName);
        int k1 = AtmsNamingConvention.FILE_NAMING_PREFIX_TDR.length() + 4;
        int k2 = k1 + AtmsNamingConvention.FILE_NAMING_LENGTH_TIMETAG;
        fileTag = shortName.substring(k1, k2);
        return fileTag;
    }

    /**
     * Get time tag of ATMS GEO file
     *
     * @param geoFileName name of ATMS GEO file
     * @return time tag of ATMS GEO file
     */
    public static String getAtmsGeoTimeTag(String geoFileName) {
        String fileTag = null;
        String shortName = AtmsNamingConvention.getFileShortname(geoFileName);
        int k1 = AtmsNamingConvention.FILE_NAMING_PREFIX_GEO.length() + 4;
        int k2 = k1 + AtmsNamingConvention.FILE_NAMING_LENGTH_TIMETAG;
        fileTag = shortName.substring(k1, k2);
        return fileTag;
    }

    /**
     * Get prefix of ATMS GEO file
     *
     * @param geoFileName name of ATMS GEO file
     * @return prefix of ATMS GEO file
     */
    public static String getAtmsGeoFilePrefix(String geoFileName) {
        String filePrefix = null;
        String shortName = AtmsNamingConvention.getFileShortname(geoFileName);
        int k1 = AtmsNamingConvention.FILE_NAMING_PREFIX_GEO.length() + 4;
        int k2 = k1 + AtmsNamingConvention.FILE_NAMING_LENGTH_TIMETAG;
        filePrefix = shortName.substring(0, k2);
        return filePrefix;
    }

    /**
     * Remove path in filename to get short name
     *
     * @param fileFullname
     * @return short name of the specified file
     */
    public static String getFileShortname(String fileFullname) {
        String str = null;
        int index = fileFullname.lastIndexOf(File.separator);
        if (index < 0) {
            index = 0;
        } else {
            index = index + 1;
        }
        str = fileFullname.substring(index);
        return str;
    }


    /**
     * Get list of geolocation file for a specified day
     * 
     * @param rootDirGeo root directory
     * @param year year
     * @param month month
     * @param day day
     * @return Array of geolocation file
     */
    public static File[] getAtmsDailyGeoFileList(String rootDirGeo, int year, int month, int day) {
        
        ArrayList<String> tagList = new ArrayList<String>();
        ArrayList<File> selectedList = new ArrayList<File>();
        
        String fileTimeTag;
        
        String geoPath;
        String geoPath2;
        FilenameFilter geoFilter;
        int k = 0;
        
        if (rootDirGeo.endsWith(File.separator)) {
            geoPath = String.format("%s%04d%s%04d-%02d-%02d%s", rootDirGeo, year, File.separator,
                    year, month, day, File.separator);
            geoPath2 = String.format("%s%04d%s%04d-%02d-%02d.SCDR%s", rootDirGeo, year, File.separator,
                    year, month, day, File.separator);

        } else {
            geoPath = String.format("%s%s%04d%s%04d-%02d-%02d%s", rootDirGeo, File.separator, year, File.separator,
                    year, month, day, File.separator);
            geoPath2 = String.format("%s%s%04d%s%04d-%02d-%02d.SCDR%s", rootDirGeo, File.separator, year, File.separator,
                    year, month, day, File.separator);
        }
        
        // add .SCDR file first (no such issue for reprocessed version)
        File dirGeo = new File(geoPath2);
        if (dirGeo.exists()) {
            // if exist, list and add files
            geoFilter = new FilenameFilter() {
                public boolean accept(File dir, String name) {
                    int[] dt = getTdrFileDay(name);
                    if (dt==null) {
                        return false;
                    } else {
                        return (dt[0] == year) && (dt[1] == month) && (dt[2] == day);
                    }
                }
            };
            File[] files = dirGeo.listFiles(geoFilter);
            int numFiles = files.length;
            for(k=0; k<numFiles; k++) {
                fileTimeTag=AtmsNamingConvention.getAtmsGeoTimeTag(files[k].getAbsolutePath());
                if(!tagList.contains(fileTimeTag)) {
                    tagList.add(fileTimeTag);
                    selectedList.add(files[k]);
                }
            }            
        } 
        // add original folder
        dirGeo = new File(geoPath);
        if (dirGeo.exists()) {
            // if exist, list and add files
            geoFilter = new FilenameFilter() {
                public boolean accept(File dir, String name) {
                    int[] dt = getTdrFileDay(name);
                    if (dt==null) {
                        return false;
                    } else {
                        return (dt[0] == year) && (dt[1] == month) && (dt[2] == day);
                    }
                }
            };
            File[] files = dirGeo.listFiles(geoFilter);
            int numFiles = files.length;
            for(k=0; k<numFiles; k++) {
                fileTimeTag=AtmsNamingConvention.getAtmsGeoTimeTag(files[k].getAbsolutePath());
                if(!tagList.contains(fileTimeTag)) {
                    tagList.add(fileTimeTag);
                    selectedList.add(files[k]);
                }
            }            
        } 
        
//   prepare output
        int fileCount = selectedList.size();
        File[] files = new File[fileCount];
        for(k=0; k<fileCount; k++) {
            files[k] = selectedList.get(k);
        }
        
        return files;
    }
    
    /**
     * Pair ATMS geolocation and TDR data files
     * @param filenameGeo, ATMS geolocation file.
     * @param dirRootTdr, root directory of TDR data.
     * @param satName, satellite name.
     * @return file name of TDR file
     */
    public static String pairAtmsTdrFile(String filenameGeo, String dirRootTdr, String satName) {
        String filenameTdr = null;

        String fileTag = getAtmsGeoTimeTag(filenameGeo);
        int[] dt = getTdrFileDay(filenameGeo);

        String tdrPath = null;
        String tdrPath2 = null;
        if (dirRootTdr.endsWith(File.separator)) {
            tdrPath = String.format("%s%04d%s%04d-%02d-%02d%s", dirRootTdr, dt[0], File.separator, dt[0], dt[1], dt[2], File.separator);
            tdrPath2 = String.format("%s%04d%s%04d-%02d-%02d.SCDR%s", dirRootTdr, dt[0], File.separator, dt[0], dt[1], dt[2], File.separator);
        } else {
            tdrPath = String.format("%s%s%04d%s%04d-%02d-%02d%s", dirRootTdr, File.separator, dt[0], File.separator, dt[0], dt[1], dt[2], File.separator);
            tdrPath2 = String.format("%s%s%04d%s%04d-%02d-%02d.SCDR%s", dirRootTdr, File.separator, dt[0], File.separator, dt[0], dt[1], dt[2], File.separator);
        }
        
        AtmsFileFilter fileFilter = new AtmsFileFilter(fileTag, AtmsFileFilter.ATMS_FILE_TYPE_TDR, satName);
        File dirTdr2 = new File(tdrPath2);
        File dirTdr = new File(tdrPath);
        int flag = 0;
        File[] files = null;
        
        if (dirTdr2.exists()) {
            files = dirTdr2.listFiles(fileFilter); 
            if (files != null) {
                int nf = files.length;
                if (nf > 0) {
                    filenameTdr = files[0].getAbsolutePath();
                    flag = 1;
                }
            }   
        }
        
        if (flag==0) {
            
            if(!dirTdr.exists()){
                dirTdr = new File(tdrPath);
            }
            files = dirTdr.listFiles(fileFilter);
            if (files != null) {
                int nf = files.length;
                if (nf > 0) {
                    filenameTdr = files[0].getAbsolutePath();
                    flag = 1;
                }
            }            
        }
                
        return filenameTdr;
    }

    /**
     * Get day of a GEO/TDR file
     *
     * @param filename filename
     * @return day represented with a vector of year, month, day
     */
    public static int[] getTdrFileDay(String filename) {
        int[] dt = new int[3];
        // remove path first to get short name
        String str = null;
        int index = filename.lastIndexOf(File.separator);
        if (index < 0) {
            index = 0;
        } else {
            index = index + 1;
        }

        str = filename.substring(index);

        try {
            dt[0] = Integer.parseInt(str.substring(FILE_NAMING_INDEX_YEAR, FILE_NAMING_INDEX_YEAR + 4));
            dt[1] = Integer.parseInt(str.substring(FILE_NAMING_INDEX_MONTH, FILE_NAMING_INDEX_MONTH + 2));
            dt[2] = Integer.parseInt(str.substring(FILE_NAMING_INDEX_DAY, FILE_NAMING_INDEX_DAY + 2));
        } catch (Exception e) {
            dt = null;
        }

        return dt;
    }

}

