/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L1.AtmsTDR.java
*
* NAME
*   atms.L1.AtmsTDR.java
*
* PURPOSE
*   Defines ATMS TDR data object
* 
* DESCRIPTION
*	This class is used to define object for TDR measurements.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package atms.L1;


import java.io.File;
import ucar.ma2.ArrayShort;
import ucar.ma2.ArrayLong;
import ucar.ma2.ArrayFloat;
import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import atms.util.*;
import atms.etc.*;

/**
 * Class to read ATMS TDRdata
 *
 * @author Xianjun Hao, George Mason University <Email: xianjun.hao@noaa.gov, xhao1@gmu.edu >
 */
public class AtmsTDR{
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    
    /**
     * Limb correction
     */
    private AtmsLimbCorrector limbCorrector = null;
        
    /**
     * Name of ATMS Geolocation file
     */
    private String filenameAtmsGeo = null;

    /**
     * Name of ATMS TDRdata file
     */
    private String filenameAtmsTdr = null;

    /**
     * Status of ATMS TDRfiles
     */
    private int[] fileStatus = new int[2];

    /**
     * Flag to indicate if TDRfiles exist
     */
    private boolean isSdrReady = false;
    
    /**
     *  file tag with time, for example: d20160901_t0001236_e0001553
     */
    private String sdrFileTag = null;

    /**
     *  orbit Id
     */
    private int orbitId = 0;
    
    /**
     *  orbit type (ascending/descending)
     */
    private int orbitType = -1;
    
    /**
     * Beam time
     */
    private double[][] beamTime = null;

    /**
     * Latitude
     */
    private double[][] lat = null;

    /**
     * Longitude
     */
    private double[][] lon = null;

    /**
     * Solar zenith angle
     */
    private double[][] solarZenithAngle = null;

    /**
     * Solar azimuth angle
     */
    private double[][] solarAzimuthAngle = null;

    /**
     * satellite zenith angle
     */
    private double[][] satelliteZenithAngle = null;

    /**
     * Satellite azimuth angle
     */
    private double[][] satelliteAzimuthAngle = null;

    /**
     * TDRbrightness temperature
     */
    private double[][][] Temperature = null;

    /**
     * Start time of each scan
     */
    private long[] startTime = null;
    
    /**
     * Shape of geolocation
     */
    private int[] shapeGeo = null;

    /**
     * Shape of brightness temperature
     */
    private int[] shapeBT = null;

    /**
     * Start time in file name
     */
    private int[] startTimeNaming= null;
    

    /**
     * Start time in file name
     */
    private int[] endTimeNaming= null;
    
    
    private long fileTimeStart = 0;
    
    private long fileTimeEnd = 0;
        
    private boolean isDebugMode = false;
    
    /**
     *  Construct an empty object. 
     */
    public AtmsTDR(AtmsLimbCorrector limbCorrector ){
        // construct an empty object first, then set files, and load data.
        this.limbCorrector = limbCorrector;
    }
    
    
    /**
     * Constructor.
     *
     * @param filenameAtmsGeo Name of ATMS geolocation file
     * @param filenameAtmsTdr Name of ATMS TDRdata file
     */
    public AtmsTDR(String filenameAtmsGeo, String filenameAtmsTdr, AtmsLimbCorrector limbCorrector) {
        this.limbCorrector = limbCorrector;
        this.filenameAtmsGeo = filenameAtmsGeo;
        this.filenameAtmsTdr = filenameAtmsTdr;
        this.init();
    }

    /**
     * Initialize reader
     */
    private void init() {
        this.beamTime = null;
        this.lat = null;
        this.lon = null;
        this.solarZenithAngle = null;
        this.solarAzimuthAngle = null;
        this.satelliteZenithAngle = null;
        this.satelliteAzimuthAngle = null;
        this.Temperature = null;
        this.startTime = null;
        this.startTimeNaming = null;
        this.endTimeNaming = null;
        this.shapeGeo = new int[2];
        this.shapeBT = new int[3];
        
        // get short name
        String sname = AtmsNamingConvention.getFileShortname(this.filenameAtmsGeo);
        
        this.setFileTime();
        
        // set debug mode
        this.isDebugMode = false;
                
        // read geolocation data
        this.loadGeoData();
        
        // read TDR_BT data
        this.loadTdrData();
        
        this.isSdrReady = (this.fileStatus[0] == ErrorAtmsSDR.SUCCESS_FILE_READ) && 
                (this.fileStatus[1] == ErrorAtmsSDR.SUCCESS_FILE_READ);
        
    }
    
    /**
     * Set debug mode
     * @param isDebugMode 
     */
    public void setDebugMode(boolean isDebugMode) {
        this.isDebugMode = isDebugMode;
    }

    /**
     * Read Geolocation data
     */
    private void loadGeoData() {
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayFloat.D2 farr = null;
        int[] shape = new int[2];
        int[] origin = new int[2];
        int dim1, dim2;
        int i, j;
        String errMsg = null;
        
        try {
            fileNc = NetcdfFile.open(this.filenameAtmsGeo);
            // Read Latitude
            dataVar = fileNc.findVariable("/All_Data/ATMS-SDR-GEO_All/Latitude");
            
            if (dataVar == null ) {
                System.out.println("Unable to find Latitude");
            }
            
            shape = dataVar.getShape();
            farr = (ArrayFloat.D2) dataVar.read(origin, shape);
            dim1 = shape[0];
            dim2 = shape[1];
            this.shapeGeo[0] = dim1;
            this.shapeGeo[1] = dim2;
           // System.out.println("Lat shape :" + dim1 +", " + dim2);
            this.lat = new double[dim1][dim2];
            for (i = 0; i < dim1; i++) {
                for (j = 0; j < dim2; j++) {
                    this.lat[i][j] = (double) farr.get(i, j);
                }
            }

            // Longitude
            dataVar = fileNc.findVariable("/All_Data/ATMS-SDR-GEO_All/Longitude");
            shape = dataVar.getShape();
            farr = (ArrayFloat.D2) dataVar.read(origin, shape);
            dim1 = shape[0];
            dim2 = shape[1];
            //System.out.println("Lon shape:" + dim1 +", " + dim2);
            this.lon = new double[dim1][dim2];
            for (i = 0; i < dim1; i++) {
                for (j = 0; j < dim2; j++) {
                    this.lon[i][j] = (double) farr.get(i, j);
                }
            }

            // /All_Data/ATMS-SDR-GEO_All/SolarZenithAngle
            dataVar = fileNc.findVariable("/All_Data/ATMS-SDR-GEO_All/SolarZenithAngle");
            shape = dataVar.getShape();
            farr = (ArrayFloat.D2) dataVar.read(origin, shape);
            dim1 = shape[0];
            dim2 = shape[1];
            //System.out.println("SZA shape:" + dim1 +", " + dim2);
            this.solarZenithAngle = new double[dim1][dim2];
            for (i = 0; i < dim1; i++) {
                for (j = 0; j < dim2; j++) {
                    this.solarZenithAngle[i][j] = (double) farr.get(i, j);
                }
            }

            // /All_Data/ATMS-SDR-GEO_All/SolarAzimuthAngle
            dataVar = fileNc.findVariable("/All_Data/ATMS-SDR-GEO_All/SolarAzimuthAngle");
            shape = dataVar.getShape();
            farr = (ArrayFloat.D2) dataVar.read(origin, shape);
            dim1 = shape[0];
            dim2 = shape[1];
            //System.out.println("SZA shape:" + dim1 +", " + dim2);
            this.solarAzimuthAngle = new double[dim1][dim2];
            for (i = 0; i < dim1; i++) {
                for (j = 0; j < dim2; j++) {
                    this.solarAzimuthAngle[i][j] = (double) farr.get(i, j);
                }
            }

            // /All_Data/ATMS-SDR-GEO_All/SatelliteZenithAngle
            dataVar = fileNc.findVariable("/All_Data/ATMS-SDR-GEO_All/SatelliteZenithAngle");
            shape = dataVar.getShape();
            farr = (ArrayFloat.D2) dataVar.read(origin, shape);
            dim1 = shape[0];
            dim2 = shape[1];
            //System.out.println("SZA shape:" + dim1 +", " + dim2);
            this.satelliteZenithAngle = new double[dim1][dim2];
            for (i = 0; i < dim1; i++) {
                for (j = 0; j < dim2; j++) {
                    this.satelliteZenithAngle[i][j] = (double) farr.get(i, j);
                }
            }

            // /All_Data/ATMS-SDR-GEO_All/SatelliteAzimuthAngle
            dataVar = fileNc.findVariable("/All_Data/ATMS-SDR-GEO_All/SatelliteAzimuthAngle");
            shape = dataVar.getShape();
            farr = (ArrayFloat.D2) dataVar.read(origin, shape);
            dim1 = shape[0];
            dim2 = shape[1];
            //System.out.println("SZA shape:" + dim1 +", " + dim2);
            this.satelliteAzimuthAngle = new double[dim1][dim2];
            for (i = 0; i < dim1; i++) {
                for (j = 0; j < dim2; j++) {
                    this.satelliteAzimuthAngle[i][j] = (double) farr.get(i, j);
                }
            }
            
            // /All_Data/ATMS-SDR-GEO_All/StartTime
            dataVar = fileNc.findVariable("/All_Data/ATMS-SDR-GEO_All/StartTime");
            shape = dataVar.getShape();
            ArrayLong.D1 varr = (ArrayLong.D1) dataVar.read(origin, shape);
            dim1 = shape[0];
            //System.out.println("SZA shape:" + dim1 +", " + dim2);
            this.startTime = new long[dim1];
            for (i = 0; i < dim1; i++) {
                    this.startTime[i] = (long) varr.get(i);
            }
                        
            
            this.fileStatus[0] = ErrorAtmsSDR.SUCCESS_FILE_READ;
            
        } catch (Exception e) { 
            e.printStackTrace();
            this.fileStatus[0] = ErrorAtmsSDR.ERROR_FILE_FAILED_TO_READ;
            this.isSdrReady = false;
            errMsg = String.format("Error in loading geolocation data from %s", fileNc);
            errLogger.error(errMsg);
        } finally {
            try {
                fileNc.close();
            } catch (Exception ce) {
            }
        }
    }

    /**
     * Read TDRdata from files.
     */
    private void loadTdrData() {
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayFloat.D1 carr = null;
        ArrayLong.D2 farr = null;
        ArrayShort.D3 sarr = null;

        int[] shape = new int[3];
        int[] origin = new int[3];
        int dim1, dim2, dim3;
        int i, j, k;
        int vtmp;

        double scaleFactor;

        try {
            fileNc = NetcdfFile.open(this.filenameAtmsTdr);

            // Beam Time
            dataVar = fileNc.findVariable("/All_Data/ATMS-TDR_All/BeamTime");
            shape = dataVar.getShape();
            farr = (ArrayLong.D2) dataVar.read(origin, shape);
            dim1 = shape[0];
            dim2 = shape[1];
            this.shapeGeo[0] = dim1;
            this.shapeGeo[1] = dim2;
            this.beamTime = new double[dim1][dim2];
            for (i = 0; i < dim1; i++) {
                for (j = 0; j < dim2; j++) {
                    this.beamTime[i][j] = ((double) farr.get(i, j)) * 1.0e-6;  // change to seconds
                }
            }

            // READ scale factor
            dataVar = fileNc.findVariable("/All_Data/ATMS-TDR_All/AntennaTemperatureFactors");
            shape = dataVar.getShape();
            carr = (ArrayFloat.D1) dataVar.read(origin, shape);
            scaleFactor = (double) carr.get(0);

            // TDR
            dataVar = fileNc.findVariable("/All_Data/ATMS-TDR_All/AntennaTemperature");
            shape = dataVar.getShape();
            sarr = (ArrayShort.D3) dataVar.read(origin, shape);
            dim1 = shape[0];
            dim2 = shape[1];
            dim3 = shape[2];

            this.shapeBT[0] = dim1;
            this.shapeBT[1] = dim2;
            this.shapeBT[2] = dim3;

            //System.out.println("TDR shape:" + dim1 +", " + dim2 +", "+ dim3);
            // 12, 96, 22
            this.Temperature = new double[dim1][dim2][dim3];
            for (i = 0; i < dim1; i++) {
                for (j = 0; j < dim2; j++) {
                    for (k = 0; k < dim3; k++) {
                        vtmp = DataUtility.convertShort(sarr.get(i, j, k));
                        if ((vtmp<=0)  || (vtmp>=AtmsGridConstant.ATMS_TDR_BRIGHTNESS_TEMPERATURE_FILLED_VALUE)) {
                            this.Temperature[i][j][k] = AtmsGridConstant.FILLED_VALUE_BRIGHTNESS_TEMPERATURE;
                        } else {
                            this.Temperature[i][j][k] = ((double) vtmp) * scaleFactor;                           
                        }
                    }
                }
            }
            this.fileStatus[1] = ErrorAtmsSDR.SUCCESS_FILE_READ;
        } catch (Exception e) {
            e.printStackTrace();
            this.fileStatus[1] = ErrorAtmsSDR.ERROR_FILE_FAILED_TO_READ;
            this.isSdrReady = false;
        } finally {
            try {
                fileNc.close();
            } catch (Exception ce) {
            }
        }
    }

    /**
     * Get beam time
     *
     * @return beam time
     */
    public double[][] getBeamTime() {
        return this.beamTime;
    }

    /**
     * Get latitude
     *
     * @return latitude
     */
    public double[][] getLatitude() {
        return this.lat;
    }

    /**
     * Get longitude
     *
     * @return longitude
     */
    public double[][] getLongitude() {
        return this.lon;
    }

    /**
     * Get solar zenith angle
     *
     * @return
     */
    public double[][] getSolarZenithAngle() {
        return this.solarZenithAngle;
    }

    /**
     * Get solar azimuth angle
     *
     * @return solar azimuth angle
     */
    public double[][] getSolarAzimuthAngle() {
        return this.solarAzimuthAngle;
    }

    /**
     * Get satellite zenith angle
     *
     * @return satelliteZenithAngle
     */
    public double[][] getSatelliteZenithAngle() {
        return this.satelliteZenithAngle;
    }

    /**
     * Get satellite azimuth angle
     *
     * @return satellite azimuth angle
     */
    public double[][] getSatelliteAzimuthAngle() {
        return this.satelliteAzimuthAngle;
    }

    /**
     * Get TDRbrightness temperature
     *
     * @return TDRbrightness temperature
     */
    public double[][][] getTdrBT() {
        return this.Temperature;
    }

    /**
     * Get limb-corrected brightness temperature
     * @return  limb-corrected brightness temperature
     */
    public double[][][] getLimbcorrTdrBT(){
        double[][][] btn = limbCorrector.doLimbCorrection(this.lat, this.lon, this.Temperature);

        return btn;
    }
    
    
    /**
     * Check if TDRdata ready
     * @return true if read TDRdata successfully, false, otherwise.
     */
    public boolean isSdrReady() {
        return this.isSdrReady;
    }

    /**
     * Get name of ATMS geolocation file
     *
     * @return name of ATMS geolocation file
     */
    public String getFilenameAtmsGeo() {
        return this.filenameAtmsGeo;
    }

    /**
     * Get name of ATMS TDRdata file
     *
     * @return name of ATMS TDRdata file
     */
    public String getFilenameAtmsTdr() {
        return this.filenameAtmsTdr;
    }

    /**
     * Get TDRfile tag
     * @return tag for TDRfile
     */
    public String getSdrFileTag(){
        return this.sdrFileTag;
    }
    
    /**
     * Get shape of Geolocation data
     * @return shape of geolocation data
     */
    public int[] getShapeGeo(){
        return this.shapeGeo;
    }
    
    /**
     * Get shape of brightness temperature data
     * @return shape of brightness temperature 
     */
    public int[] getShapeBt(){
        return this.shapeBT;
    }
    
    /**
     * Get orbit Id
     * @return orbit id number
     */
    public int getOrbitId(){
        return this.orbitId;
    }
    

    /**
     * Convert time string  in file name to time.
     */
    public void setFileTime(){
        this.startTimeNaming = new int[7];
        this.endTimeNaming = new int[7];
        int tmp = 0;
        int ten = 10;
        String str = AtmsNamingConvention.getAtmsGeoTimeTag(this.filenameAtmsGeo);
        // d20160901_t0046436_e0047153
        //System.out.println(str);
        this.startTimeNaming[0] = Integer.parseInt(str.substring(1, 5));
        this.startTimeNaming[1] = Integer.parseInt(str.substring(5, 7));
        this.startTimeNaming[2] = Integer.parseInt(str.substring(7, 9));
        this.startTimeNaming[3] = Integer.parseInt(str.substring(11, 13));
        this.startTimeNaming[4] = Integer.parseInt(str.substring(13, 15));
        tmp = Integer.parseInt(str.substring(15, 18));
        this.startTimeNaming[5] = tmp/ten;
        this.startTimeNaming[6] = (tmp - (tmp/ten)*ten);

        int[] tv = this.startTimeNaming;        
        // 2016 9 1 23 52 11 6  
        this.fileTimeStart = TimeUtility.getStartSecondSince58(tv[0], tv[1], tv[2], tv[3], tv[4], tv[5], tv[6]);
        
        
        String info1 = String.format("***** %04d-%02d-%02d %02d:%02d:%02d%01d", tv[0], tv[1], tv[2], tv[3], tv[4], tv[5], tv[6]);
        
        for(int k=0; k<3; k++){
            this.endTimeNaming[k] = this.startTimeNaming[k];
        }
         this.endTimeNaming[3] = Integer.parseInt(str.substring(20, 22));
         this.endTimeNaming[4] = Integer.parseInt(str.substring(22, 24));
        tmp = Integer.parseInt(str.substring(24, 27));
        this.endTimeNaming[5] = tmp/ten;
        this.endTimeNaming[6] = (tmp - (tmp/ten)*ten);
        
        tv = this.endTimeNaming;        
        this.fileTimeEnd = TimeUtility.getStartSecondSince58(tv[0], tv[1], tv[2], tv[3], tv[4], tv[5], tv[6]);
        String info2 = String.format("--> %04d-%02d-%02d %02d:%02d:%02d%01d", tv[0], tv[1], tv[2], tv[3], tv[4], tv[5], tv[6]);        
    }
    
    
    /** 
     * Check beam time
     */
    public void checkBeamTime(){
        if (this.beamTime == null ){
            return;
        }
        // check date and time
        int dim1 = this.shapeGeo[0];
        int dim2 = this.shapeGeo[1];
        
        long tm1 = (long) Math.round(this.beamTime[0][0]);
        long tm2 = (long) Math.round(this.beamTime[dim1-1][dim2-1]);
        
        TimeUtility tmApp = new TimeUtility();
        int[] tv1 = tmApp.getTimeVec(tm1);
        int[] tv2 = tmApp.getTimeVec(tm2);
        
        
        String info1=String.format("%d --> %4d-%02d-%02d %02d:%02d:%02d", tm1, tv1[0], tv1[1], tv1[2], tv1[3], tv1[4], tv1[5]);
        String info2=String.format("%d --> %4d-%02d-%02d %02d:%02d:%02d", tm2, tv2[0], tv2[1], tv2[2], tv2[3], tv2[4], tv2[5]);
        
        System.out.println("Granuale beam start: " + info1);
        System.out.println("Granuale beam end : " + info2);
        info1 = String.format("FileStart=%d, Delta start= %d", this.fileTimeStart , tm1 - this.fileTimeStart );
        info2 = String.format("FileEnd  =%d, Delta end  = %d", this.fileTimeEnd, tm2 - this.fileTimeEnd );
        System.out.println( info1);
        System.out.println( info2);
        
        System.out.println("");
        
        tm1= Math.round(((double) this.startTime[0])*1.0e-6);
        tm2= Math.round(((double) this.startTime[dim1-1])*1.0e-6);
        tv1 = tmApp.getTimeVec(tm1);
        tv2 = tmApp.getTimeVec(tm2);        
        info1=String.format("%d --> %4d-%02d-%02d %02d:%02d:%02d", tm1, tv1[0], tv1[1], tv1[2], tv1[3], tv1[4], tv1[5]);
        info2=String.format("%d --> %4d-%02d-%02d %02d:%02d:%02d", tm2, tv2[0], tv2[1], tv2[2], tv2[3], tv2[4], tv2[5]);
        
        System.out.println("Granuale beam start: " + info1);
        System.out.println("Granuale beam end  : " + info2);

        info1 = String.format("FileStart=%d, Delta start= %d", this.fileTimeStart , tm1 - this.fileTimeStart );
        info2 = String.format("FileEnd  =%d, Delta end  = %d", this.fileTimeEnd, tm2 - this.fileTimeEnd );
        System.out.println( info1);
        System.out.println( info2);
        
    }
    
    /**
     * Get orbit type (ascending/descending)
     * @return orbit type
     */
    public int getOrbitType(){
        return this.orbitType;
    }
    
    /**
     * Set ATMS TDRdata files and load data.
     *
     * @param filenameAtmsGeo Name of ATMS geolocation file
     * @param filenameAtmsTdr Name of ATMS TDRdata file
     */
    public void setAtmsTDR(String filenameAtmsGeo, String filenameAtmsTdr) {
        this.filenameAtmsGeo = filenameAtmsGeo;
        this.filenameAtmsTdr = filenameAtmsTdr;
        // re-load data
        this.init();
    }

}

