/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L3.AtmsGriddingTask.java
*
* NAME
*   atms.L3.AtmsGriddingTask.java
*
* PURPOSE
*   Generate gridded ATMS TDR data 
* 
* DESCRIPTION
*	This class is used to generate monthly gridded ATMS TDR measurements.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package atms.L3;


import java.lang.Runnable;
import java.io.File;
import java.util.ArrayList;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import static atms.etc.AtmsGridConstant.GRID_NUM_LAT;
import static atms.etc.AtmsGridConstant.GRID_NUM_LON;
import static atms.etc.AtmsGridConstant.NUMBER_SELECTED_CHANNELS;
import static atms.etc.AtmsGridConstant.INDEX_SCAN_POSITION_START;
import static atms.etc.AtmsGridConstant.INDEX_SCAN_POSITION_END;

import atms.etc.*;
import atms.L1.*;
import atms.config.AtmsTaskConfig;
import atms.util.*;

/**
 * This class is used to generate monthly gridded ATMS TDR measurements.
 * @author Xianjun Hao, George Mason University
 */
public class AtmsGriddingTask implements  Runnable{
    
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    
    /**
     * Year of the data
     */
    private int year = 0;

    /**
     * Month of the data
     */
    private int month = 0;
    
    /**
     * first day of the month
     */
    private int day1 = 0;
    
    /**
     * Last day of the month
     */
    private int day2 = 0;
        
    /**
     * Start TAI time 
     */
    private long startTAI = 0;  // start TAI time in seconds

    /**
     * End TAI time
     */
    private long endTAI  = 0;   // end TAI time in seconds
    
    /**
     * Task configuration
     */
    private AtmsTaskConfig taskConfig = null;
    
    /**
     * Gridded data for ascending orbits
     */
    private GriddedAtmsData gdataAscending = null;

    /**
     * Gridded data for descending orbits
     */
    private GriddedAtmsData gdataDescending = null;

    /**
     * Gridded data for both ascending and descending orbits
     */
    private GriddedAtmsData gdataAll = null;
    
    /**
     * Satellite name
     */
    private String satName;
    
    /**
     * Constructor.
     * @param year, year of the task
     * @param month, month of the task
     * @param taskConfig, task configuration information 
     */
    public AtmsGriddingTask(int year, int month, AtmsTaskConfig taskConfig) {
        this.year = year;
        this.month = month;
        this.setupPeriod();
        
        this.taskConfig = taskConfig;
        this.gdataAscending = new GriddedAtmsData();
        this.gdataDescending = new GriddedAtmsData();
        this.gdataAll = new GriddedAtmsData();
        this.satName = this.taskConfig.getTaskSatellite();
        
    }
    
    /**
     * Setup date period for the selected month
     */
    private void setupPeriod(){
        long jday2 = 0;
        int kyear = this.year;
        int kmonth = this.month;

        this.day1 = 1;
        
        if (kmonth<12) {
           jday2 =  TimeUtility.getDateNumber(kyear, kmonth+1,1) - 1;
        } else {
           jday2 =  TimeUtility.getDateNumber(kyear+1, 1,1) - 1;            
        } 
        TimeUtility tmApp = new TimeUtility();        
        int[] tm2 = tmApp.getDateVec(jday2);
        this.day2 = tm2[2];
        int[] te = tmApp.getDateVec(jday2+1);
        this.startTAI = TimeUtility.getTaiSeconds(this.year, this.month, this.day1, 0,0,0);
        this.endTAI = TimeUtility.getTaiSeconds(te[0], te[1], te[2], 0,0,0);        
    }
    
    /**
     * Show task information
     */
    public void showTaskInformation(){
        String msg = String.format("Gridding ATMS data, %4d-%02d-%02d->%02d", this.year, this.month, this.day1, this.day2);
        System.out.println(msg);
        msg = String.format("        TAI: %d -> %d", this.startTAI, this.endTAI);
        System.out.println(msg);
        msg = TimeUtility.getCalDayofTai(this.startTAI);
        System.out.println(msg);
        msg = TimeUtility.getCalDayofTai(this.endTAI);
        System.out.println(msg);        
    }
    
    /**
     * Override method run to produce daily gridded ATMS.
     */    
    public void run(){        
        this.processMonthlyData();
        this.writeDataFiles();
    }
    

    /**
     * Generate monthly gridded data
     */
    private void processMonthlyData(){
        int day = 0;
        int numTdr = 0;
        int numMissed = 0;
        int k = 0;
        String filenameGeo;
        String filenameTdr;
        String filenameLut;
        String filenameMask;

        AtmsLimbCorrector limbCorrector = null;     
        filenameLut=String.format("%s%s", this.taskConfig.rootAtmsLut, AtmsLimbConstant.FILENAME_LIMB_LUT);
        filenameMask=String.format("%s%s", this.taskConfig.rootAtmsLut, AtmsLimbConstant.FILENAME_LIMB_LANDSEA_MASK);
        limbCorrector = new AtmsLimbCorrector(filenameLut, filenameMask);
        
        AtmsTDR atmsTdr = new AtmsTDR(limbCorrector);


        // gridding day by day and do average
        for(day=this.day1; day<=this.day2; day++) {
            File[] listFileGeo = AtmsNamingConvention.getAtmsDailyGeoFileList(this.taskConfig.rootAtmsGeo, this.year, this.month, day);
            int numGeo = listFileGeo.length;

            numTdr = 0;
            numMissed = 0;
            
            for(k=0; k<numGeo; k++) {
                filenameGeo = listFileGeo[k].getName();
                filenameTdr = AtmsNamingConvention.pairAtmsTdrFile(filenameGeo, this.taskConfig.rootAtmsTdr, this.satName);
                if(filenameTdr==null) {
                    numMissed ++;
                } else {
                    // read data
                    try{
                        atmsTdr.setAtmsTDR(listFileGeo[k].getAbsolutePath(), filenameTdr);                
                        numTdr++;                
                        // update data: ascending, descending, and all
                        this.updateGridDataT(atmsTdr);
                    } catch (Exception e){
                        e.printStackTrace();
                        numMissed++;
                    }
                }
            }        
            String msg = String.format("%4d-%02d-%02d: NumGeo=%d, numTdr=%d, NumMissed=%d", year, month, day,  numGeo, numTdr, numMissed);
            System.out.println(msg);
            infoLogger.info(msg);        
        }
    }
    
    
    
    /**
     * Read GEO and TDR files, and grid data
     */
    private void processGridding(){
        int day = 0;
        // gridding day by day and do average
        
        File[] listFileGeo = AtmsNamingConvention.getAtmsDailyGeoFileList(this.taskConfig.rootAtmsGeo, this.year, this.month, day);
        int numGeo = listFileGeo.length;
        int numTdr = 0;
        int numMissed = 0;
        int k = 0;
        String filenameGeo;
        String filenameTdr;
        String filenameLut;
        String filenameMask;
        
        AtmsLimbCorrector limbCorrector = null;     
        filenameLut=String.format("%s%s", this.taskConfig.rootAtmsLut, AtmsLimbConstant.FILENAME_LIMB_LUT);
        filenameMask=String.format("%s%s", this.taskConfig.rootAtmsLut, AtmsLimbConstant.FILENAME_LIMB_LANDSEA_MASK);

        limbCorrector = new AtmsLimbCorrector(filenameLut, filenameMask);
        
        AtmsTDR atmsTdr = new AtmsTDR(limbCorrector);
        
        for(k=0; k<numGeo; k++) {
            filenameGeo = listFileGeo[k].getName();
            filenameTdr = AtmsNamingConvention.pairAtmsTdrFile(filenameGeo, this.taskConfig.rootAtmsTdr, this.satName);
            if(filenameTdr==null) {
                numMissed ++;
            } else {
                // read data
                try{
                    atmsTdr.setAtmsTDR(listFileGeo[k].getAbsolutePath(), filenameTdr);                
                    numTdr++;                
                    // update data: ascending, descending, and all
                    this.updateGridDataT(atmsTdr);
                    
                } catch (Exception e){
                    e.printStackTrace();
                    numMissed++;
                }
            }
        }        
        String msg = String.format("%4d-%02d-%02d: NumGeo=%d, numTdr=%d, NumMissed=%d", year, month, day,  numGeo, numTdr, numMissed);
        System.out.println(msg);
        infoLogger.info(msg);
        
    }
    
    /**
     * Update gridded data of ascending, descending, and all orbits
     * @param atmsTdr ATMS TDR record
     */
    private void updateGridDataT(AtmsTDR atmsTdr){
        int i = 0;
        int j = 0;
        int k = 0;
        int k1 = 0;
        int k2 = 0;
        int grow = 0;
        int gcol = 0;
        double npixel = 0.0d;
        double tmpBt = 0.0d;
        double tmpVza = 0.0d;
        double tmpMeanOld = 0.0d;
        double tmpStdOld = 0.0d;
        double tmpLat1 = 0.0d;
        double tmpLat2 = 0.0d;
        //double tmpLon1 = 0.0d;
        //double tmpLon2 = 0.0d;
        
        // set debug mode for testing
        atmsTdr.setDebugMode(true);
        
        
        // update this.gdataAll 
        double[][] plat = atmsTdr.getLatitude();  // YTrack, XTrack
        double[][] plon = atmsTdr.getLongitude(); // YTrack, XTrack
//        double[][][] pbt = atmsTdr.getTdrBT();    // YTrck, XTrack, Channel
        double[][][] pbt = atmsTdr.getLimbcorrTdrBT();    //  get limb corrected data YTrck, XTrack, Channel
        
        double[][] pvza =atmsTdr.getSatelliteZenithAngle();  // XTrack, XTrack
        
        int[] shapeGeo  = atmsTdr.getShapeGeo(); 
        int[] shapeBt = atmsTdr.getShapeBt(); 
                
        int numScan = shapeGeo[0]; // number of scans
        int[] orbitFlag = new int[numScan]; // orbit glag: -1: descendiing, 1: ascendding; 0, undetermined
        double[] z = new double[numScan];
        double dz = 0.0d;
        double z1 = 0.0d;
        double z2 = 0.0d;

        double zmin =  Math.abs(AtmsGridConstant.FILLED_VALUE_INVALID_LATITUDE);
        double zmax = AtmsGridConstant.FILLED_VALUE_INVALID_LATITUDE;
        int indexMin = -1;
        int indexMax = -1;
        int numValidScan = 0;
        int kOrbitFlag = 0;
                
        for(i=0; i<numScan; i++){
            
            //testing only
             //System.out.println("pbt limb corr ch6:" +pbt[i][47][5]   );
            
            orbitFlag[i] = 0;  // undetermined            
            z[i] = AtmsGridConstant.FILLED_VALUE_INVALID_LATITUDE;
            tmpLat1 = plat[i][AtmsGridConstant.INDEX_SCAN_NADIR_LEFT];
            //tmpLon1 = plon[i][AtmsGridConstant.INDEX_SCAN_NADIR_LEFT];
            tmpLat2 = plat[i][AtmsGridConstant.INDEX_SCAN_NADIR_RIGHT];
            //tmpLon2 = plon[i][AtmsGridConstant.INDEX_SCAN_NADIR_RIGHT];
            
            if( ( Math.abs(tmpLat1)>90.0d)|| ( Math.abs(tmpLat2)>90.0d) ) {
                continue; // skip invalid                
            }

            numValidScan ++;
            z[i] = (tmpLat1 + tmpLat2)*0.5;            
            if (z[i] > zmax) {
                zmax = z[i];
                indexMax = i;
            }
            if (z[i] < zmin){
                zmin = z[i];
                indexMin = i;
            }    
        }
        
        //System.out.println("........... # scan="+ numScan +  ",    #Valid scan=" + numValidScan);
        // If all geolocations 
        if (numValidScan < 2) {
            return; // no enough valid scan, skip this granule
        }
        
        // not in polar area, easy to determine
        if ((Math.abs(zmax)< AtmsGridConstant.ORBIT_POLAR_LATITUDE)&& (Math.abs(zmin)< AtmsGridConstant.ORBIT_POLAR_LATITUDE)) {
            // check direction
            if (indexMin < indexMax) {
                // ascending
                for(i=0; i<numScan; i++) {
                    orbitFlag[i] = 1;
                }
            } else {
                // descending
                for(i=0; i<numScan; i++) {
                    orbitFlag[i] = -1;
                }
            }
        } else {
            // Maybe mixed orbit, check scan by scan
            // find first valid scan
            k1 = 0;
            while( z[k1] < AtmsGridConstant.THRESHOLD_DELTA_LATITUDE) {
                k1 ++;
            }

            // find last valid scan
            k2 = numScan - 1;
            while( z[k2] < AtmsGridConstant.THRESHOLD_DELTA_LATITUDE) {
                k2 --;
            }

            for(k=k1; k<=k2; k++) {
                if(z[k] <AtmsGridConstant.THRESHOLD_DELTA_LATITUDE) {
                    orbitFlag[k] = 0;
                }  else {                                           
                    if (k<k2) {
                        if (z[k+1] < AtmsGridConstant.THRESHOLD_DELTA_LATITUDE) {
                           // try k-1
                           if (k==k1) {
                               // failed to check
                               orbitFlag[k] = 0;
                               dz = AtmsGridConstant.FILLED_VALUE_INVALID_LATITUDE;
                           } else {
                               dz = z[k] - z[k-1];
                           }
                        } else {
                            dz = z[k+1] - z[k];                   
                        }
                    } else {
                        dz = z[k] - z[k-1];
                    }
                    if (dz < AtmsGridConstant.THRESHOLD_DELTA_LATITUDE) {
                          orbitFlag[k] = 0;
                    } else {
                        if(dz>0) {
                              orbitFlag[k] = 1;
                        } else {
                              orbitFlag[k] = -1;                            
                        }
                    }
                }                
            }
            
        }
        
        for(i=0; i<numScan; i++) {
            kOrbitFlag = orbitFlag[i];
                // only keep scan positions with limited scan angle ( 44 scan positions: 27-70 (26-69)) 
            for(j=INDEX_SCAN_POSITION_START; j<=INDEX_SCAN_POSITION_END; j++){
                // calculate grid index
                grow = (int) Math.floor((plat[i][j]-AtmsGridConstant.GRID_LATITUDE_START) / AtmsGridConstant.GRID_STEP_LATITUDE);
                gcol = (int) Math.floor((plon[i][j]-AtmsGridConstant.GRID_LONGITUDE_START) / AtmsGridConstant.GRID_STEP_LONGITUDE);
                
                if (gcol==AtmsGridConstant.GRID_NUM_LON) {
                    gcol = 0;
                }
                if ((grow<0) || (grow>=GRID_NUM_LAT) || (gcol<0) || (gcol>=GRID_NUM_LON)) {
                    continue; // skip invalid geolocation
                }
                
                // VZA
                tmpVza = pvza[i][j];
                
                // check channels, update channel by channel
                for(k=0; k<NUMBER_SELECTED_CHANNELS; k++) {
                    tmpBt = pbt[i][j][k];
                    if ( (tmpBt<=AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE) || (tmpBt>= AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE)) {
                        continue; // skip invalid value
                    }
                    // increase pixel count 
                    this.gdataAll.pc_gridded[k][grow][gcol] = this.gdataAll.pc_gridded[k][grow][gcol] + 1;
                    if (kOrbitFlag==1) {
                        this.gdataAscending.pc_gridded[k][grow][gcol] = this.gdataAscending.pc_gridded[k][grow][gcol] + 1;                        
                    } else {
                        if(kOrbitFlag == -1) {
                            this.gdataDescending.pc_gridded[k][grow][gcol] = this.gdataDescending.pc_gridded[k][grow][gcol] + 1;                                                    
                        }
                    }
                    
                    // update gridded data for all orbits
                    npixel = (double) this.gdataAll.pc_gridded[k][grow][gcol];
                    if (npixel <2) {
                        // first pixel
                        this.gdataAll.bt_gridded_mean[k][grow][gcol] = (float)tmpBt;
                        this.gdataAll.bt_gridded_std[k][grow][gcol] = 0.0f;
                        // minvza
                        this.gdataAll.minvza_gridded[k][grow][gcol] = (float)tmpVza;
                        this.gdataAll.bt_gridded_minvza[k][grow][gcol] = (float)tmpBt;                        
                        
                    } else {
                        // update mean and std
                        tmpMeanOld = this.gdataAll.bt_gridded_mean[k][grow][gcol];
                        tmpStdOld = this.gdataAll.bt_gridded_std[k][grow][gcol];
                       
                        this.gdataAll.bt_gridded_mean[k][grow][gcol] = (float) (( tmpBt + tmpMeanOld*(npixel-1.0d) ) / npixel);
                        this.gdataAll.bt_gridded_std[k][grow][gcol] = (float) ( ((npixel-2.0d)/(npixel-1.0d))* tmpStdOld +(tmpBt-tmpMeanOld)*(tmpBt-tmpMeanOld)/npixel );                            
                        
                        // update minvza
                        if(this.gdataAll.minvza_gridded[k][grow][gcol] > tmpVza) {
                            this.gdataAll.minvza_gridded[k][grow][gcol] = (float)tmpVza;
                            this.gdataAll.bt_gridded_minvza[k][grow][gcol] = (float)tmpBt; 
                        }

                    }                    
                    
                    //  update gridded data for ascending orbits
                    if (kOrbitFlag==1) {
                        npixel = (double) this.gdataAscending.pc_gridded[k][grow][gcol];
                        if (npixel <2) {
                         // first pixel
                            this.gdataAscending.bt_gridded_mean[k][grow][gcol] = (float)tmpBt;
                            this.gdataAscending.bt_gridded_std[k][grow][gcol] = 0.0f;
                            // minvza
                            this.gdataAscending.minvza_gridded[k][grow][gcol] = (float)tmpVza;
                            this.gdataAscending.bt_gridded_minvza[k][grow][gcol] = (float)tmpBt;                                                        
                        } else {
                         // update mean and std
                            tmpMeanOld = this.gdataAscending.bt_gridded_mean[k][grow][gcol];
                            tmpStdOld = this.gdataAscending.bt_gridded_std[k][grow][gcol];                       
                            this.gdataAscending.bt_gridded_mean[k][grow][gcol] = (float) (( tmpBt + tmpMeanOld*(npixel-1.0d) ) / npixel);
                            this.gdataAscending.bt_gridded_std[k][grow][gcol] = (float) ( ((npixel-2.0d)/(npixel-1.0d))* tmpStdOld +(tmpBt-tmpMeanOld)*(tmpBt-tmpMeanOld)/npixel );                            
                        // update minvza
                            if(this.gdataAll.minvza_gridded[k][grow][gcol] > tmpVza) {
                                this.gdataAscending.minvza_gridded[k][grow][gcol] = (float)tmpVza;
                                this.gdataAscending.bt_gridded_minvza[k][grow][gcol] = (float)tmpBt;                                 
                            }
                        }                    
                    }    
                    
                    //  update gridded data for descending orbits
                    if (kOrbitFlag==-1){
                        npixel = (double) this.gdataDescending.pc_gridded[k][grow][gcol];
                        if (npixel <2) {
                            // first pixel
                            this.gdataDescending.bt_gridded_mean[k][grow][gcol] = (float)tmpBt;
                            this.gdataDescending.bt_gridded_std[k][grow][gcol] = 0.0f;
                            // update minvza
                            this.gdataDescending.minvza_gridded[k][grow][gcol] = (float)tmpVza;
                            this.gdataDescending.bt_gridded_minvza[k][grow][gcol] = (float)tmpBt;
                        } else {
                            // update mean and std
                            tmpMeanOld = this.gdataDescending.bt_gridded_mean[k][grow][gcol];
                            tmpStdOld = this.gdataDescending.bt_gridded_std[k][grow][gcol];                       
                            this.gdataDescending.bt_gridded_mean[k][grow][gcol] = (float) (( tmpBt + tmpMeanOld*(npixel-1.0d) ) / npixel);
                            this.gdataDescending.bt_gridded_std[k][grow][gcol] = (float) ( ((npixel-2.0d)/(npixel-1.0d))* tmpStdOld +(tmpBt-tmpMeanOld)*(tmpBt-tmpMeanOld)/npixel );                            
                            if(this.gdataAll.minvza_gridded[k][grow][gcol] > tmpVza) {
                                this.gdataDescending.minvza_gridded[k][grow][gcol] = (float)tmpVza;
                                this.gdataDescending.bt_gridded_minvza[k][grow][gcol] = (float)tmpBt;                                
                            }
                        }                    
                    }
                }
            }
        }
        
    }
    
    /**
     * Write gridded data to netCDF files. Three files per day corresponding to all orbits, ascending orbits, and descending orbits
     * respectively.
     */
    private void writeDataFiles(){
        // 3 files per day: all orbits; ascending orbits; descending orbits
        // write gridded data for all orbits
        int orbit = AtmsGridConstant.ORBIT_FLAG_ALL;
        FileGriddedAtms app = new FileGriddedAtms(year, month, orbit, taskConfig.rootAtmsGridded, this.gdataAll, this.taskConfig);
        //System.out.println(app.getFileName());
        //System.out.println(app.getFileAbsolutePath());

        // write gridded data for ascending orbits
        orbit = AtmsGridConstant.ORBIT_FLAG_ASCENDING;
        app = new FileGriddedAtms(year, month,  orbit, taskConfig.rootAtmsGridded, this.gdataAscending, this.taskConfig);
        //System.out.println(app.getFileName());
        //System.out.println(app.getFileAbsolutePath());

        // write gridded data for descending orbits
        orbit = AtmsGridConstant.ORBIT_FLAG_DESCENDING;
        app = new FileGriddedAtms(year, month,  orbit, taskConfig.rootAtmsGridded, this.gdataDescending, this.taskConfig);
        //System.out.println(app.getFileName());
        //System.out.println(app.getFileAbsolutePath());        
    }    
    

}
