/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L3.AtmsTdrL3.java
*
* NAME
*   atms.L3.AtmsTdrL3.java
*
* PURPOSE
*   Generate monthly gridded ATMS TDR data 
* 
* DESCRIPTION
*	This is the main class to generate monthly gridded ATMS TDR measurements.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package atms.L3;

import java.io.File;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import atms.etc.*;
import atms.L1.*;
import atms.config.AtmsTaskConfig;
import atms.util.*;

/**
 * Main class for producing ATMS L3 data. 
 * @author <Xianjun Hao <xhao1@gmu.edu>
 * 
 */
public class AtmsTdrL3 {

    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");

    /**
     * Task configuration
     */
    private AtmsTaskConfig taskConfig;
    
    /**
     * Time utility tool
     */
    private TimeUtility timeApp = new TimeUtility();
    
    /**
     * Number of months to be processed
     */
    private int numMonth = 0;
    
    /**
     * Years for the selected task
     */
    private int[] fyear;  // year for the selected task

    /**
     * Months for the selected task
     */
    private int[] fmonth; // month for the selected task
    
    /**
     * Constructor.
     * @param taskConfig, task config.
     */
    public AtmsTdrL3(AtmsTaskConfig taskConfig) {
        this.taskConfig = taskConfig;
        this.numMonth = taskConfig.numMonth;
        this.fyear = taskConfig.fyear;
        this.fmonth = taskConfig.fmonth;
    }


    
    /**
     * Run the gridding tasks using multiple threads
     */
    public void processAtmsTask(){
        
        // use ExecutorService to run threads
        ExecutorService executor = Executors.newFixedThreadPool(this.taskConfig.NumberOfThreads);
        for(int k=0; k<this.numMonth; k++) {
            // get calendar day (year, month, day)
            AtmsGriddingTask taskProcessor = new AtmsGriddingTask(this.fyear[k], this.fmonth[k],  this.taskConfig);
            executor.execute(taskProcessor);
        }
        
        // shutdown ExecutorService
        try {
            executor.shutdown();
            executor.awaitTermination(60, TimeUnit.SECONDS);                
        } catch (Exception e) {
            e.printStackTrace();
        }         
    }

    /**
     * Display task information
     */
    public void showTaskInfo(){
        String msg = null;
        msg="******************************************************************";
        infoLogger.info(msg);
        infoLogger.info("Satellite:" + this.taskConfig.getTaskSatellite());

        infoLogger.info("DIR_ROOT_GEO:" + this.taskConfig.rootAtmsGeo);
        infoLogger.info("DIR_ROOT_TDR:" + this.taskConfig.rootAtmsTdr);
        infoLogger.info("DIR_ROOT_GRIDDED:" + this.taskConfig.rootAtmsGridded);
        infoLogger.info("DIR_ROOT_LUT:" + this.taskConfig.rootAtmsLut);
        infoLogger.info("Number of Threads:" + this.taskConfig.NumberOfThreads);
        infoLogger.info("Precipitation Screening:" + this.taskConfig.precipitationScreening);
        msg = String.format("Start Time: %04d-%02d", this.taskConfig.yearStart, this.taskConfig.monthStart);
        infoLogger.info(msg);
        msg = String.format("End   Time: %04d-%02d", this.taskConfig.yearEnd, this.taskConfig.monthEnd);
        infoLogger.info(msg);
    }
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        String dirRootGeo;
        String dirRootTdr;
        String dirRootGridded; 
        String dirLut;
        String satName="j01";
        String str;
        int k = 0;

        System.out.println("Number of args="+ args.length);
        
        for (k=0; k<args.length; k++) {
            System.out.println(args[k]);
        } 
        
        if (OSValidator.isWindows()) {
            dirRootGeo = "D:\\data\\ATMS\\TDR\\n20\\ATMS-GEO\\";
            dirRootTdr = "D:\\data\\ATMS\\TDR\\n20\\ATMS-TDR\\";
            dirRootGridded = "D:\\data\\ATMS\\TDR\\Gridded\\";
            dirLut="D:\\data\\ATMS\\LUT\\";
        } else {
            dirRootGeo = "/home/xhao/data/ATMS/TDR/n20/ATMS-GEO/";
            dirRootTdr = "/home/xhao/data/ATMS/TDR/n20/ATMS-TDR/";
            dirRootGridded = "/home/xhao/data/ATMS/TDR/testing/L3/n20/";
            dirLut="/home/xhao/data/ATMS/lut/";
        }
        
        
        AtmsTaskConfig taskConfig = new AtmsTaskConfig();

        k=0;
        if (args.length>=11){
            satName=args[k++];
            taskConfig.setTaskSatellite(satName.toLowerCase());

            str = args[k++];
            if (!str.endsWith(File.separator)) {
                str = str+File.separator;
            }
            taskConfig.rootAtmsGeo = str;

            str = args[k++];
            if (!str.endsWith(File.separator)) {
                str = str+File.separator;
            }            
            taskConfig.rootAtmsTdr = str;

            str = args[k++];
            if (!str.endsWith(File.separator)) {
                str = str+File.separator;
            }            
            taskConfig.rootAtmsGridded = str;
 
            str = args[k++];
            if (!str.endsWith(File.separator)) {
                str = str+File.separator;
            }            
            taskConfig.rootAtmsLut = str;
            
            taskConfig.NumberOfThreads = Integer.parseInt(args[k++]);
            taskConfig.yearStart = Integer.parseInt(args[k++]);;
            taskConfig.monthStart = Integer.parseInt(args[k++]);;
            taskConfig.yearEnd = Integer.parseInt(args[k++]);;
            taskConfig.monthEnd = Integer.parseInt(args[k++]);;
            taskConfig.precipitationScreening = Boolean.parseBoolean(args[k++]);
        } else {        
            // default Task configuration
            taskConfig.setTaskSatellite(satName.toLowerCase());
            taskConfig.rootAtmsGeo = dirRootGeo;
            taskConfig.rootAtmsTdr = dirRootTdr;
            taskConfig.rootAtmsGridded = dirRootGridded;
            taskConfig.rootAtmsLut = dirLut;
            taskConfig.NumberOfThreads = 1;
            taskConfig.yearStart = 2020;
            taskConfig.monthStart = 10;
            taskConfig.yearEnd = 2020;
            taskConfig.monthEnd = 10;
            taskConfig.precipitationScreening=false;
        }

        taskConfig.setupMonthList();
        AtmsTdrL3 taskApp = new AtmsTdrL3(taskConfig);
        
        infoLogger.info("start processing ATMS L3 task");
        
        taskApp.showTaskInfo();
        
        taskApp.processAtmsTask();
        
    }

    
}
