/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L3.FileGriddedAtms.java
*
* NAME
*   atms.L3.FileGriddedAtms.java
*
* PURPOSE
*   Define file structure of the monthly gridded ATMS TDR data
* 
* DESCRIPTION
*   This class is used to define file structure of the monthly gridded L3 file.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package atms.L3;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Files;
import java.nio.file.Paths;

import java.util.ArrayList;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.TimeZone;
import java.util.Date;

import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;
import ucar.ma2.*;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriter;
import java.util.List;

import static atms.etc.AtmsGridConstant.GRID_NUM_LAT;
import static atms.etc.AtmsGridConstant.GRID_NUM_LON;
import static atms.etc.AtmsGridConstant.NUMBER_SELECTED_CHANNELS;

import atms.etc.*;
import atms.L1.*;
import atms.config.AtmsTaskConfig;
import atms.util.*;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 * NetCDF file for the gridded ATMS data
 * @author Xianjun Hao, George Mason University
 */
public class FileGriddedAtms {

    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    /**
     * Satellite ID
     */
    private String satId = "npp"; // default satId
    
    /**
     * Year
     */
    private int year = 0;

    /**
     * Month
     */
    private int month = 0;
    
    /**
     * Day
     */
    private int day = 0;
    
    /**
     * Orbit flag to indicate ascending or descending
     */
    private int orbitFlag = 0;
    
    /**
     * File name
     */
    private String fileName = null;
    
    /**
     * File directory
     */
    private String fileDir = null;
    
    /**
     * Full path of the data file.
     */
    private String fileFullPath = null;
    
    /**
     *  Flag to indicate if the gridded data is ready or not.
     */
    private boolean flagReady = false;
    
    /**
     * Flag for precipitation screening. 
     */
    private boolean precipitationScreening = false;
    
    /**
     * Object for gridded data
     */
    private GriddedAtmsData gdata = null;
    
    /**
     * NetCDF writer for output file - ascending.
     */
    private NetcdfFileWriter ncFileOut = null;

    /**
     * Constructor.
     * @param year, year of the data.
     * @param month, month of the data.
     * @param orbitFlag, orbit flag.
     * @param gdata, object of gridded data.
     * @param taskConfig, task configuration.
     */
    public FileGriddedAtms(int year, int month, int orbitFlag, GriddedAtmsData gdata, AtmsTaskConfig taskConfig) {
        this.year  = year;
        this.month = month;
        this.orbitFlag = orbitFlag;
        //this.fileDir = String.format("%s%s%04d", fileDir, File.separator, year);
        this.fileDir = taskConfig.rootAtmsGridded;
        this.flagReady = false;
        this.gdata = gdata;
        this.precipitationScreening = taskConfig.precipitationScreening;
        this.satId=(taskConfig.getTaskSatellite()).toLowerCase();
        this.init();
        //create NetCDF file     
        this.createNetcdfFile();
        // close file
        this.close();        
    }
    

    /**
     * Constructor.
     * @param year, year of the data.
     * @param month, month of the data.
     * @param orbitFlag, orbit flag.
     * @param fileDir, file directory
     * @param gdata, object of gridded data.
     * @param taskConfig, task configuration.
     */    
    public FileGriddedAtms(int year, int month, int orbitFlag, String fileDir, GriddedAtmsData gdata, AtmsTaskConfig taskConfig) {
        this.year  = year;
        this.month = month;
        this.orbitFlag = orbitFlag;
        this.fileDir = fileDir;
        this.flagReady = false;
        this.gdata = gdata;
        this.satId=(taskConfig.getTaskSatellite()).toLowerCase();        
        this.init();
        //create NetCDF file     
        this.createNetcdfFile();
        // close file
        this.close();        
    }
   
    /**
     * Initialize path and data file
     */
    private void init(){
        Path path = Paths.get(this.fileDir);
        if (!Files.exists(path)) {
            try{
                Files.createDirectories(path);
            } catch (IOException ioe){
                ioe.printStackTrace();
            }
        }
        this.setGriddedFilename();
    }
    
    /**
     * Initialize NetCDF file for descending orbit
     */
    private void createNetcdfFile() {
        int i = 0;
        int j = 0;
        int k = 0;
        float tmp = 0.0f;
        float tmpStd = 0.0f;
        int tmpInt = 0;
        try {
            this.ncFileOut = NetcdfFileWriter.createNew(NetcdfFileWriter.Version.netcdf3, this.fileFullPath);
            
            // create Dimensions
            Dimension dimChannel = this.ncFileOut.addDimension(null, "channel", NUMBER_SELECTED_CHANNELS);
            Dimension dimLon = this.ncFileOut.addDimension(null, "longitude",GRID_NUM_LON);
            Dimension dimLat = this.ncFileOut.addDimension(null, "latitude", GRID_NUM_LAT);

            List<Dimension> dimChD = new ArrayList<Dimension>();
            dimChD.add(dimChannel);

            List<Dimension> dimLatD = new ArrayList<Dimension>();
            dimLatD.add(dimLat);

            List<Dimension> dimLonD = new ArrayList<Dimension>();
            dimLonD.add(dimLon);
            
            
            List<Dimension> dimGridD = new ArrayList<Dimension>();
            dimGridD.add(dimChannel);
            dimGridD.add(dimLat);
            dimGridD.add(dimLon);

            List<Dimension> dimGrid2 = new ArrayList<Dimension>();
            dimGrid2.add(dimLat);
            dimGrid2.add(dimLon);
            
            // define global attributes
            this.defineGlobalAttributes();
            
            // define corridinates variable variables
            Variable vlon = this.ncFileOut.addVariable(null, "longitude", DataType.FLOAT, dimLonD);
            Variable vlat = this.ncFileOut.addVariable(null, "latitude", DataType.FLOAT, dimLatD);
            Variable vchn = this.ncFileOut.addVariable(null, "channel", DataType.INT, dimChD);
            vlat.addAttribute(new Attribute("units", "degree"));
            vlon.addAttribute(new Attribute("units", "degree"));
            
            // define data variables
            // add BT_MEAN variable & attributes         
            String varName = AtmsGridConstant.VAR_NAME_BT_BLOCKMEAN;
            Variable varBtMean = this.ncFileOut.addVariable(null, varName , DataType.FLOAT, dimGridD);
            varBtMean.addAttribute(new Attribute("units","Kelvin" ));
            varBtMean.addAttribute(new Attribute("valid_max", AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE));
            varBtMean.addAttribute(new Attribute("valid_min", AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE));
            varBtMean.addAttribute(new Attribute("_FillValue", AtmsGridConstant.FILLED_VALUE_INVALID_BRIGHTNESS_TEMPERATURE));          
            varBtMean.addAttribute(new Attribute("coordinates","channel latitude longitude" ));
            varBtMean.addAttribute(new Attribute("stand_name",varName));          
            
            // add BT_STD variable & attributes         
            varName = AtmsGridConstant.VAR_NAME_BT_STD;
            Variable varBtStd = this.ncFileOut.addVariable(null, varName , DataType.FLOAT, dimGridD);
            varBtStd.addAttribute(new Attribute("units","Kelvin" ));
            varBtStd.addAttribute(new Attribute("valid_max", AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE));
            varBtStd.addAttribute(new Attribute("valid_min", 0.0f));
            varBtStd.addAttribute(new Attribute("_FillValue", AtmsGridConstant.FILLED_VALUE_INVALID_BRIGHTNESS_TEMPERATURE));          
            varBtStd.addAttribute(new Attribute("coordinates","channel latitude longitude" ));
            varBtStd.addAttribute(new Attribute("stand_name",varName));          

            // add BT_minvza variable & attributes         
            varName = AtmsGridConstant.VAR_NAME_BT_MINVZA;
            Variable varBtMinvza = this.ncFileOut.addVariable(null, varName , DataType.FLOAT, dimGridD);
            varBtMinvza.addAttribute(new Attribute("units","Kelvin" ));
            varBtMinvza.addAttribute(new Attribute("valid_max", AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE));
            varBtMinvza.addAttribute(new Attribute("valid_min", AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE));
            varBtMinvza.addAttribute(new Attribute("_FillValue", AtmsGridConstant.FILLED_VALUE_INVALID_BRIGHTNESS_TEMPERATURE));          
            varBtMinvza.addAttribute(new Attribute("coordinates","channel latitude longitude" ));
            varBtMinvza.addAttribute(new Attribute("stand_name",varName));          
            
            // add MINVZA variable & attributes
            varName = AtmsGridConstant.VAR_NAME_MINVZA;
            Variable varVza = this.ncFileOut.addVariable(null, varName , DataType.FLOAT, dimGridD);
            varVza.addAttribute(new Attribute("units","degress" ));
            varVza.addAttribute(new Attribute("valid_max", 90.0f));
            varVza.addAttribute(new Attribute("valid_min", 0.0f));
            varVza.addAttribute(new Attribute("_FillValue", AtmsGridConstant.FILLED_VALUE_INVALID_VZA));          
            varVza.addAttribute(new Attribute("coordinates","latitude longitude" ));
            varVza.addAttribute(new Attribute("stand_name",varName));          

            // add Pix_count variable & attributes         
            varName = AtmsGridConstant.VAR_NAME_PIXELCOUNT;
            Variable varPixcount = this.ncFileOut.addVariable(null, varName , DataType.INT, dimGridD);
            varPixcount.addAttribute(new Attribute("valid_min", 0));
            varPixcount.addAttribute(new Attribute("coordinates","channel latitude longitude" ));
            varPixcount.addAttribute(new Attribute("stand_name",varName));          
            
            // create file 
            this.ncFileOut.create();
            
            //System.out.println("File created, start to write data .....");
            
            // write data
            ArrayFloat.D1 dataLat = new ArrayFloat.D1(GRID_NUM_LAT);
            ArrayFloat.D1 dataLon = new ArrayFloat.D1(GRID_NUM_LON);      
            ArrayInt.D1 dataChn = new ArrayInt.D1(NUMBER_SELECTED_CHANNELS);
            
            // write latitude                        
            for (i=0; i<GRID_NUM_LAT; i++) {
                tmp = AtmsGridConstant.GRID_MAXIMUM_LATITUDE + (i+0.5f)*AtmsGridConstant.GRID_STEP_LATITUDE;
                dataLat.set(i, tmp );
            }
            this.ncFileOut.write(vlat, dataLat);
            
            // write longitude                    
            for (i=0; i<GRID_NUM_LON; i++) {
                tmp = AtmsGridConstant.GRID_MINIMUM_LONGITUDE + (i+0.5f)*AtmsGridConstant.GRID_STEP_LONGITUDE;
                dataLon.set(i, tmp );
            }
            this.ncFileOut.write(vlon, dataLon);

            // write channel                   
            for (i=0; i<NUMBER_SELECTED_CHANNELS; i++) {
                tmpInt = i+1 ;
                dataChn.set(i, tmpInt );
            }
            this.ncFileOut.write(vchn, dataChn);
            
            // write BT_mean
            ArrayFloat.D3 vdata = new ArrayFloat.D3(NUMBER_SELECTED_CHANNELS, GRID_NUM_LAT, GRID_NUM_LON);
            for (i=0; i<NUMBER_SELECTED_CHANNELS; i++) {
                for(j=0; j<GRID_NUM_LAT; j++) {
                    for(k=0; k<GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.gdata.bt_gridded_mean[i][j][k]);
                    }
                }
            }
            this.ncFileOut.write(varBtMean, vdata);

            // write BT_std
            vdata = new ArrayFloat.D3(NUMBER_SELECTED_CHANNELS, GRID_NUM_LAT, GRID_NUM_LON);
            for (i=0; i<NUMBER_SELECTED_CHANNELS; i++) {
                for(j=0; j<GRID_NUM_LAT; j++) {
                    for(k=0; k<GRID_NUM_LON; k++) {
                        tmpStd = (float)Math.sqrt(this.gdata.bt_gridded_std[i][j][k]);
                        vdata.set(i,j,k, tmpStd);
                    }
                }
            }
            this.ncFileOut.write(varBtStd, vdata);

            // write BT_MINVZA
            vdata = new ArrayFloat.D3(NUMBER_SELECTED_CHANNELS, GRID_NUM_LAT, GRID_NUM_LON);
            for (i=0; i<NUMBER_SELECTED_CHANNELS; i++) {
                for(j=0; j<GRID_NUM_LAT; j++) {
                    for(k=0; k<GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.gdata.bt_gridded_minvza[i][j][k]);
                    }
                }
            }
            this.ncFileOut.write(varBtMinvza, vdata);

            // write MINVZA
            vdata = new ArrayFloat.D3(NUMBER_SELECTED_CHANNELS, GRID_NUM_LAT, GRID_NUM_LON);
            for (i=0; i<NUMBER_SELECTED_CHANNELS; i++) {
                for(j=0; j<GRID_NUM_LAT; j++) {
                    for(k=0; k<GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.gdata.minvza_gridded[i][j][k]);
                    }
                }
            }
            this.ncFileOut.write(varVza, vdata);

            // write Pix_count
             ArrayInt.D3 pdata = new ArrayInt.D3(NUMBER_SELECTED_CHANNELS, GRID_NUM_LAT, GRID_NUM_LON);
            for (i=0; i<NUMBER_SELECTED_CHANNELS; i++) {
                for(j=0; j<GRID_NUM_LAT; j++) {
                    for(k=0; k<GRID_NUM_LON; k++) {
                        pdata.set(i,j,k, this.gdata.pc_gridded[i][j][k]);
                    }
                }
            }
            this.ncFileOut.write(varPixcount, pdata);
                        
        } catch (Exception nce) {
            nce.printStackTrace();
            errLogger.error("Failed to wtite netcdf file" + this.fileFullPath);
        }
    }


    /**
     * Close NetCDF file
     */
    public void close(){
        if (this.ncFileOut != null) {
            try {
                this.ncFileOut.close();                
            } catch (Exception e) {
                e.printStackTrace();
            }
        }        
    }
    
    
    /**
     * Define the global attributes of NetCDF file
     */
    private void defineGlobalAttributes(){
        // define global attributes        
        this.ncFileOut.addGroupAttribute(null, new Attribute("Conventions", "CF-1.6"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("Metadata_Conventions", "CF-1.6,Unidata Dataset Discovery v1.0, NOAA CDR v1.0, GDS v2.0"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("_nc3_strict", 1d));
        this.ncFileOut.addGroupAttribute(null, new Attribute("acknowledgment", "This project was supported by a grant from the NOAA Climate Data Record (CDR) Program"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("cdr_data_type", "Grid"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("cdr_program", ""));
        this.ncFileOut.addGroupAttribute(null, new Attribute("contributor_name", "Xianjun Hao"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("contributor_role", "Visiting Scientist"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("creator_email", "Xianjun.Hao@noaa.gov"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("creator_name", "Xianjun Hao"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("PI_email", "Cheng-Zhi.Zou@noaa.gov"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("PI_name", "Cheng-Zhi Zou"));

        // date and time
        Date date = new Date();
        String strDateFormat = "yyyy-MM-dd hh:mm:ss";
        DateFormat dateFormat = new SimpleDateFormat(strDateFormat);
        dateFormat.setTimeZone(TimeZone.getTimeZone( "UTC" ) );
        String formattedDate= String.format("%s Z", dateFormat.format(date));
        this.ncFileOut.addGroupAttribute(null, new Attribute("date_created", formattedDate));
        this.ncFileOut.addGroupAttribute(null, new Attribute("geospatial_lat_max", 90.0f));
        this.ncFileOut.addGroupAttribute(null, new Attribute("geospatial_lat_min", -90.0f));
        this.ncFileOut.addGroupAttribute(null, new Attribute("geospatial_lon_max", 180.0f));
        this.ncFileOut.addGroupAttribute(null, new Attribute("geospatial_lon_min", -180.0f));       

        if (this.precipitationScreening) {
            this.ncFileOut.addGroupAttribute(null, new Attribute("precipitationScreening", "Yes"));
        } else {
            this.ncFileOut.addGroupAttribute(null, new Attribute("precipitationScreening", "No"));            
        }

        this.ncFileOut.addGroupAttribute(null, new Attribute("institution", "NOAA/NESDIS/STAR>"
                + "Center for Satellite Applications and Research, NESDIS, NOAA, U.S. Department of Commerce"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("keywords", "Atmosphere > Atmospheric Temperature > "
                + "Air Temperature, Vertical Location > Troposphere"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("keywords_vocabulary", 
                "NASA Global Change Master Directory (GCMD) Earth Science Keywords, Version 8.0"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("license", "No constraints on data access or use."));
        this.ncFileOut.addGroupAttribute(null, new Attribute("metadata_link", "gov.noaa.ncdc:C00757"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("naming_authority", "gov.noaa.ncdc"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("platform", "SNPP"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("processing_level", "NOAA Level 3"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("product_version", "2.0"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("sensor", "ATMS> Advanced Technology Microwave Sounder"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("source", "NOAA/NESDIS_MSU_Radiance_FCDR"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("spatial_resolution", "2.5 latitude by 2.5 longitude"));
        this.ncFileOut.addGroupAttribute(null, new Attribute("standard_name_vocabulary",
                "CF Standard Name Table (v26, 08 Novemeber 2013)"));        

        this.ncFileOut.addGroupAttribute(null, new Attribute("Satellite Name", this.satId));
        this.ncFileOut.addGroupAttribute(null, new Attribute("Year", (int) this.year));
        this.ncFileOut.addGroupAttribute(null, new Attribute("Month", (int) this.month));
        switch(this.orbitFlag){
            case AtmsGridConstant.ORBIT_FLAG_ALL:
                this.ncFileOut.addGroupAttribute(null, new Attribute("Orbit", (String) "Ascending & Descensing"));
                break;
            case AtmsGridConstant.ORBIT_FLAG_ASCENDING:
                this.ncFileOut.addGroupAttribute(null, new Attribute("Orbit", (String) "Ascending"));
               break;
            case AtmsGridConstant.ORBIT_FLAG_DESCENDING:
                this.ncFileOut.addGroupAttribute(null, new Attribute("Orbit", (String) "Descending"));            
                break;
            default:
                break;                    
        }
        
    }
    
    /** 
     * check if the ATMS file is ready for data I/O.
     * @return flag to indicate if the netcdf4 file fr Gridded ATMS data is ready
     */
    public boolean isReady(){
        return this.flagReady;
    }

    /**
     * Define file name of daily gridded ATMS brightness temperature FCDR.
     *
     * @param satId, ID of satellite.
     * @param orbitId, ID of orbit.
     * @param year, year.
     * @param month, month.
     * @return short filename (without path) of  gridded AMSU-A brightness
     * temperature FCDR for specified satellite, orbit, and date.
     */    
    public static String getAtmsL3Filename(String satId,int orbitId, int year, int month) {
        String filename ="";
        int index = JPSSConstant.getJpssSatelliteIndex(satId);
        String fileTag = JPSSConstant.getJpssFileTag(index);
       switch(orbitId) {
           case AtmsGridConstant.ORBIT_FLAG_ALL:
               filename = String.format("NESDIS-STAR_%s_Monthly_Gridded_ATMS_V5_T%04d%02d_all.nc", fileTag, year, month);               
               break;
           case AtmsGridConstant.ORBIT_FLAG_ASCENDING:
               filename = String.format("NESDIS-STAR_%s_Monthly_Gridded_ATMS_V5_T%04d%02d_asc.nc", fileTag, year, month);
               break;
           case AtmsGridConstant.ORBIT_FLAG_DESCENDING:
               filename = String.format("NESDIS-STAR_%s_Monthly_Gridded_ATMS_V5_T%04d%02d_dsc.nc", fileTag, year, month);
               break;
           default:
               filename = null;
               break;
       }                       
        
        return filename;
    }

    /**
     * Define L3 file name
     */
    private void setGriddedFilename() {
       this.fileName = getAtmsL3Filename(this.satId, this.orbitFlag, this.year, this.month);
       
       if(this.fileDir !=null) {
           if(this.fileDir.endsWith(File.separator)) {
               this.fileFullPath = String.format("%s%s", this.fileDir, this.fileName);
           } else {
               this.fileFullPath = String.format("%s%s%s", this.fileDir, File.separator, this.fileName);               
           }
           
       }
    }
    
    /**
     * Get name of data file.
     * @return 
     */
    public String getFileName(){
        return this.fileName;
    }
    
    /**
     * Get absolute path of the ATMS data file.
     * @return 
     */
    public String getFileAbsolutePath(){
        return this.fileFullPath;
    }
    

}

