/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L4.AtmsTdrL4Data.java
*
* NAME
*   atms.L4.AtmsTdrL4Data.java
*
* PURPOSE
*   Define time series of the monthly gridded ATMS TDR data 
* 
* DESCRIPTION
*	This class is used to generate time series of the monthly gridded ATMS TDR data.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package atms.L4;

import java.io.File;

import ucar.ma2.ArrayFloat;
import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import static atms.etc.AtmsGridConstant.GRID_NUM_LAT;
import static atms.etc.AtmsGridConstant.GRID_NUM_LON;
import atms.L3.FileGriddedAtms;
import atms.etc.AtmsGridConstant;
import atms.etc.JPSSConstant;
import geogrid.Geo2dTimeSeries;
import geogrid.GeoGridConstant;

/**
 * This class is used to merge gridded monthly ATMS TDR data into time series 
  * @author Xianjun Hao, George Mason University 
 */
public class AtmsTdrL4Data {

    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    
    /**
     * Start year of the L4 data
     */
    private int yearStart = 0;

    /**
     * Start month of the L4 data
     */
    private int monthStart = 0;
    
    /**
     * End year of the L4 data
     */
    private int yearEnd = 0;
    
    /**
     * End year of the L4 data
     */
    private int monthEnd = 0;
    
    /**
     * ATMS channel index (1-22)
     */
    private int channel = 0;
    
    /**
     * Satellite name
     */
    private String satName = "";
    
    /**
     * Satellite tag
     */
    private String satTag = "";
    
    /**
     * Number of months of the L4 time series
     */
    private int numMonths = 0; 
    
    /**
     *  brightness temperature time series of ascending orbits 
     */
    private float[][][] bt_asc; 
    
    /**
     *  brightness temperature time series of descending orbits
     */
    private float[][][] bt_dsc; 

    /**
     *  Year for the time series
     */
    private int[] fyear;
    
    /**
     *  Month for the time series
     */
    private int[] fmonth;
    
    /**
     *  time for the time seris
     */
    private float[] ft; 
    
    /**
     * Land/Sea Mask
     */
    private int[][] landMask = null;
    
    /**
     * Flag to mark if the data is valid
     */
    private boolean isValid = false;
    
    /**
     * Constructor.
     * @param satName, satellite name
     * @param channel, ATMS channel
     * @param year1, year of the start time
     * @param month1, month of the start time 
     * @param year2,  year of the end time
     * @param month2, month of the end time 
     */
    public AtmsTdrL4Data(String satName, int channel, int year1, int month1, int year2, int month2){
        this.satName = satName;
        this.channel = channel;
        this.yearStart = year1;
        this.monthStart = month1;
        this.yearEnd = year2;
        this.monthEnd = month2;
        int index = JPSSConstant.getJpssSatelliteIndex(satName);
        this.satTag = JPSSConstant.getJpssFileTag(index);
        this.initData();
    }
    
    /**
     * Initialize the L4 data
     */
    private void initData(){
        int km, kyear, kmonth;
        
        this.numMonths = (this.yearEnd - this.yearStart)*12 + (this.monthEnd - this.monthStart) + 1;
        if (this.numMonths < 1) {
            this.isValid = false;
        } else {
            this.isValid = true;
            this.bt_asc = new float[this.numMonths][GRID_NUM_LAT][GRID_NUM_LON];
            this.bt_dsc = new float[this.numMonths][GRID_NUM_LAT][GRID_NUM_LON];
            this.fyear = new int[this.numMonths];
            this.fmonth = new int[this.numMonths];
            this.ft = new float[this.numMonths];
            for(int k=0; k<this.numMonths; k++) {
                km = k + this.monthStart;
                kmonth = km%12;
                if (kmonth == 0) {
                        kmonth = 12;
                        kyear = this.yearStart + ((km-1)/12);                        
                } else {
                    kyear = this.yearStart + (km/12);
                }
                this.fyear[k] = kyear;
                this.fmonth[k] = kmonth;
                this.ft[k] = (float)kyear + ((float)kmonth - 0.5f)/12;
            }
        }
    }
    
    /**
     * check if data is valid
     * @return true if data is valid, otherwise false.
     */
    public boolean isValid() {
        return this.isValid;
    }
    
    /**
     * Get start year
     * @return start year
     */
    public int getYearStart() {
        return this.yearStart;
    }
    
    /**
     * Get end year
     * @return end year
     */
    public int getYearEnd(){
        return this.yearEnd;
    }

    /**
     * Get start month
     * @return start month
     */
    public int getMonthStart() {
        return this.monthStart;
    }
    
    /**
     * Get end month
     * @return end month
     */
    public int getMonthEnd(){
        return this.monthEnd;
    }
    
    /**
     * Get year list of L4 data
     * @return year list of L4 data
     */
    public int[] getFyear() {
        return this.fyear;
    }

    /**
     * Get month list of L4 data
     * @return month list of L4 data
     */
    public int[] getFmonth() {
        return this.fmonth;
    }
    
    /**
     * Get time for the time series
     * @return time for the L4 data time series
     */
    public float[] getFt(){
        return this.ft;
    }
    /**
     * Get satellite name
     * @return  satellite name 
     */
    public String getSatName() {
        return this.satName;
    }
    
    /**
     * Get ATMS channel
     * @return ATMS channel 
     */
    public int getChannel() {
        return this.channel;
    }
    
    /**
     * Get number of months of the L4 data time series
     * @return number of months
     */
    public int getNumMonths(){
        return this.numMonths;
    }
    /**
     * Show data information for testing.
     */
    public void showInfo(){
        System.out.println("L4 data, satellite = " + this.satName);
        System.out.println("L4 data, channel = " + this.channel);
        System.out.println("L4 data, year1 = " + this.yearStart);
        System.out.println("L4 data, month1 = " + this.monthStart);
        System.out.println("L4 data, year2 = " + this.yearEnd);
        System.out.println("L4 data, month2 = " + this.monthEnd);        
        System.out.println("L4 data, numMonths = " + this.numMonths);        
    }
    
    /**
     * Set data for a specific month
     * @param kyear, year of the data
     * @param kmonth, month of the data
     * @param dataL3, gridded data for the specified month 
     * @return true if no error, false if there are problems
     */
    public boolean setMonthlyData(int kyear, int kmonth, float[][] dataL3Asc, float[][] dataL3Dsc) {
        boolean ret = false;
        int i, j, i2;
        int km = (kyear - this.yearStart) * 12 + kmonth - this.monthStart;
        if (km<0) {
            ret = false;
        } else {
            for(i=0; i<GRID_NUM_LAT; i++){
               i2 =  GRID_NUM_LAT -i - 1; // L3 grid data in reverse order 
               for (j=0; j<GRID_NUM_LON; j++) {                   
                   this.bt_asc[km][i][j]= dataL3Asc[i2][j];
                   this.bt_dsc[km][i][j]= dataL3Dsc[i2][j];
               }   
            } 
            ret = true;
        }
        return ret;
    }
    
    /**
     * Get L4 data of ascending orbits
     * @return L4 data, monthly time series
     */
    public float[][][] getBtDataAsc(){
        return this.bt_asc;
    }

    /**
     * Get L4 data of descending orbits
     * @return L4 data, monthly time series
     */
    public float[][][] getBtDataDsc(){
        return this.bt_dsc;
    }

    /**
     * Get global mean time series of the ascending L4 data
     * @return 
     */
    public float[] getAscGlobalMean() {
        float[] btm = new float[this.numMonths];
        double[] vm;        
        Geo2dTimeSeries g2t = new Geo2dTimeSeries( this.yearStart, this.monthStart, this.yearEnd, this.monthEnd, this.bt_asc);
        int k1 = 3;
        int k2 = 68;
        g2t.setValidLatIndex(k1,k2);
        g2t.setValidRange(AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE, AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE);
        vm = g2t.getGlobalMean();
        for(int i=0; i<this.numMonths; i++) {
            btm[i] = (float)vm[i];
        }
        return btm;
    }    

    /**
     * Get global mean time series of the descending L4 data
     * @return 
     */
    public float[] getDscGlobalMean() {
        float[] btm = new float[this.numMonths];
        double[] vm;        
        Geo2dTimeSeries g2t = new Geo2dTimeSeries( this.yearStart, this.monthStart, this.yearEnd, this.monthEnd, this.bt_dsc);
        int k1 = 3;
        int k2 = 68;
        g2t.setValidLatIndex(k1,k2);
        g2t.setValidRange(AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE, AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE);
        vm = g2t.getGlobalMean();
        for(int i=0; i<this.numMonths; i++) {
            btm[i] = (float)vm[i];
        }
        return btm;
    }    
    /**
     * Get global means (ocean, land, & global) of ascending orbits
     * @return global mean time series of ascending orbits
     */
    public float[][] getAscGlobalMeanAll() {
        float[][] btm = new float[3][this.numMonths];
        double[][] vm;        
        Geo2dTimeSeries g2t = new Geo2dTimeSeries( this.yearStart, this.monthStart, this.yearEnd, this.monthEnd, this.bt_asc);
        int k1 = 3;
        int k2 = 68;
        if (this.landMask!= null){
            g2t.setLandseaMask(this.landMask);
        }
        g2t.setValidLatIndex(k1,k2);
        g2t.setValidRange(AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE, AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE);
        vm = g2t.getGlobalMeanAll();
        
        for(int i=0; i<this.numMonths; i++) {            
            for(int k=0; k<3; k++){
                btm[k][i] = (float)vm[k][i];
            }
        }
        return btm;
    }    

    
    /**
     * Get global means (ocean, land, & global)
     * @return global mean time series
     */
    public float[][] getDscGlobalMeanAll() {
        float[][] btm = new float[3][this.numMonths];
        double[][] vm;        
        Geo2dTimeSeries g2t = new Geo2dTimeSeries( this.yearStart, this.monthStart, this.yearEnd, this.monthEnd, this.bt_dsc);
        int k1 = 3;
        int k2 = 68;
        if (this.landMask!= null){
            g2t.setLandseaMask(this.landMask);
        }
        g2t.setValidLatIndex(k1,k2);
        g2t.setValidRange(AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE, AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE);
        vm = g2t.getGlobalMeanAll();
        
        for(int i=0; i<this.numMonths; i++) {            
            for(int k=0; k<3; k++){
                btm[k][i] = (float)vm[k][i];
            }
        }
        return btm;
    }    
    
    /**
     * Set Land/Sea Mask
     * @param mask 
     */
    public void setLandseaMask(int[][] mask) {
        this.landMask = new int[GRID_NUM_LAT][GRID_NUM_LON];
        for(int i=0; i<GRID_NUM_LAT; i++) {
            for(int j=0; j<GRID_NUM_LON; j++) {
                this.landMask[i][j] = mask[i][j];
            }
        }
    }
    
    /**
     * Generate L4 data from L3 data files
     * @param dirL3, directory of L3 data files 
     * @return number of valid files
     */
    public int[] generateL4Data(String dirL3) {
        int numValidAsc = 0;
        int numValidDsc = 0;
        int[] numValid = new int[2];
        //getAtmsL3Filename(String satId,int orbitId, int year, int month);     
        int i, j, k;
        int kyear, kmonth;
        String filenameAsc, filenameDsc, fullpathAsc, fullpathDsc;
        float[][] Ta, Td;

        if (dirL3.endsWith(File.separator)) {
            dirL3=dirL3+this.satTag+File.separator;
        } else {
            dirL3=dirL3+File.separator+this.satTag+File.separator;            
        }
        
        for (k=0; k<this.numMonths; k++) {
            kyear = this.fyear[k];
            kmonth = this.fmonth[k];
            filenameAsc = FileGriddedAtms.getAtmsL3Filename(this.satName, AtmsGridConstant.ORBIT_FLAG_ASCENDING, kyear, kmonth);
            filenameDsc = FileGriddedAtms.getAtmsL3Filename(this.satName, AtmsGridConstant.ORBIT_FLAG_DESCENDING, kyear, kmonth);
            fullpathAsc=dirL3+filenameAsc;
            fullpathDsc=dirL3+filenameDsc;
            Ta = loadL3Data(fullpathAsc);
            if (Ta != null){
                numValidAsc++;
                for (i = 0; i < GRID_NUM_LAT; i++) {
                    for (j = 0; j < GRID_NUM_LON; j++) {
                        // L3 grid in different oder
                        this.bt_asc[k][i][j] = Ta[GRID_NUM_LAT-1-i][j];
                    }
                }
            }
            Td = loadL3Data(fullpathDsc);
            if (Td != null){
                numValidDsc++;
                for (i = 0; i < GRID_NUM_LAT; i++) {
                    for (j = 0; j < GRID_NUM_LON; j++) {
                        // L3 grid in different oder
                        this.bt_dsc[k][i][j] = Td[GRID_NUM_LAT-1-i][j];
                    }
                }
            }
            
        }
                
        numValid[0] = numValidAsc;
        numValid[1] = numValidDsc;
        return numValid;
    }
    
    /**
     * Quality control to remove data for specified time.
     * @param qyear, year of the data to be removed.
     * @param qmonth, month of the data to be removed. 
     */
    public void qaControl(int[] qyear, int[] qmonth) {
        int nq = qyear.length;
        int kyear, kmonth, index;
        for (int k=0; k<nq; k++) {
            kyear = qyear[k];
            kmonth = qmonth[k];
            // find index for the corresponding time
            index = -1;
            for(int j=0; j<this.numMonths; j++) {
                if ((this.fyear[j]==kyear) & (this.fmonth[j]==kmonth)) {
                    index = j;
                    break;
                }
            }
            if (index>=0){
                for(int i=0; i<GRID_NUM_LAT; i++) {
                    for(int j=0; j<GRID_NUM_LON; j++) {
                        this.bt_asc[index][i][j] = AtmsGridConstant.FILLED_VALUE_INVALID_BRIGHTNESS_TEMPERATURE;
                        this.bt_dsc[index][i][j] = AtmsGridConstant.FILLED_VALUE_INVALID_BRIGHTNESS_TEMPERATURE;
                    }
                }
            }
        }
    }
    
    
    /**
     * Read data from L3 data file
     * @param filenameL3
     * @return 
     */
    private float[][] loadL3Data(String filenameL3) {
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayFloat.D3 darr3 = null;
        int[] shape = new int[3];
        int[] origin = new int[3];
        int dim1, dim2, dim3;
        int i, j, k;
        int nt = 0;
        String varName = "BT_mean";
        float[][] T = new float[GRID_NUM_LAT][GRID_NUM_LON];

        for (i = 0; i < 3; i++) {
            origin[i] = 0;
        }
        
        for(i=0; i<GRID_NUM_LAT; i++) {
            for(j=0; j<GRID_NUM_LON; j++) {
                T[i][j] = AtmsGridConstant.FILLED_VALUE_INVALID_BRIGHTNESS_TEMPERATURE;
            }
        }
        
        try {
            fileNc = NetcdfFile.open(filenameL3);
            dataVar = fileNc.findVariable(varName);
            if (dataVar == null) {
                errLogger.error("failed to find variable BT_mean from file " + filenameL3);
                T = null;
            } else {
                shape = dataVar.getShape();
                darr3 = (ArrayFloat.D3) dataVar.read(origin, shape);
                nt = shape[0];
                k = this.channel - 1;
                for (i = 0; i < GRID_NUM_LAT; i++) {
                    for (j = 0; j < GRID_NUM_LON; j++) {
                        T[i][j] = darr3.get(k, i, j);
                    }
                }
            } 
        } catch (Exception e) {
            e.printStackTrace();
            errLogger.error("failed to read BT_mean from " + filenameL3 + ", filled with invalid values");
            T = null;
        }
                    
        return T;
    }
    
}
