/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.L4.FileGriddedAtmsL4.java
*
* NAME
*   atms.L4.FileGriddedAtmsL4.java
*
* PURPOSE
*   Define files for the time series of the monthly gridded ATMS TDR data 
* 
* DESCRIPTION
*	This class is used to generate netcdf files for the time series of the monthly gridded ATMS TDR data.
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package atms.L4;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;
import ucar.ma2.*;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriter;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import atms.etc.*;
import atms.util.*;
import atms.config.AtmsTaskConfig;
import static atms.etc.AtmsGridConstant.GRID_NUM_LAT;
import static atms.etc.AtmsGridConstant.GRID_NUM_LON;
import static atms.etc.AtmsGridConstant.NUMBER_SELECTED_CHANNELS;
import cdr.MltCdrConstant;
import geogrid.GeoGrid2D;
import geogrid.GeoGridConstant;

/**
 * This class is used to generate netcdf files for the time series of the monthly gridded ATMS TDR data.
 * @author Xianjun Hao, George Mason University
 */
public class FileGriddedAtmsL4 {
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");

    /**
     * Satellite name
     */
    private String satName;
    
    /**
     * file tag
     */
    private String fileTag;
    
    /**
     * ATMS channel
     */
    private int channel;    
    
    /** 
     * year of the start time
     */
    private int yearStart;
    
    /**
     * month of the start time
     */
    private int monthStart;
    
    /**
     * year of the end time
     */
    private int yearEnd;
    
    /**
     * month of the end time
     */
    private int monthEnd;
    
    /**
     * Number of months in the time series
     */
    private int numMonths;
    
    /**
     * Directory of the L3 monthly gridded data
     */
    private String dirL3; 
    
    /**
     * Directory of the L4 data
     */
    private String dirL4; 
    
    /**
     * Gridded L4 data
     */
    private AtmsTdrL4Data dataL4;
    
    /**
     * Name of L4 file
     */
    private String filenameL4;
    
    /**
     * Constructor
     * @param satName, satellite name
     * @param channel, ATMS channel
     * @param year1, year of the start time
     * @param month1, month of the start time
     * @param year2, year of the end time
     * @param month2, month of the end time
     * @param dirL3, Directory of the L3 monthly gridded data
     * @param dirL4, Directory of the L4  data 
     */
    public FileGriddedAtmsL4(String satName, int channel, int year1, int month1, int year2, int month2, String dirL3, String dirL4) {
        this.satName = satName;
        this.channel = channel;
        this.yearStart = year1;
        this.monthStart = month1;
        this.yearEnd = year2;
        this.monthEnd = month2;
        int index = JPSSConstant.getJpssSatelliteIndex(satName);
        this.fileTag = JPSSConstant.getJpssFileTag(index);
        this.numMonths = (year2-year1)*12 + (month2-month1+1);
        this.dirL3 = dirL3;
        this.dirL4 = dirL4;
        String sfilename = getAtmsL4Filename(satName, channel);
        if (dirL4.endsWith(File.separator)) {
            this.filenameL4 = dirL4 + sfilename;
        } else {
            this.filenameL4 = dirL4 + File.separator + sfilename;            
        }
        this.generateL4Data();
    }
         
    /**
     * Generate L4 data from L3 data
     */
    private void generateL4Data(){
        this.dataL4 = new AtmsTdrL4Data( this.satName,  this.channel,  this.yearStart, this.monthStart, this.yearEnd, this.monthEnd);
        this.dataL4.generateL4Data(dirL3);
    }
    
    /**
     * Write ATMS L4 data to netcdf file.
     * @return true if success, false if there are any errors.
     */
    public boolean writeL4DataToFile(){
        boolean ret = false;
        NetcdfFileWriter ncFileOut = null;
        
        try{
            ncFileOut = NetcdfFileWriter.createNew(NetcdfFileWriter.Version.netcdf3, this.filenameL4);
            
            // create Dimensions
            Dimension dimTime = ncFileOut.addDimension(null, "time", this.numMonths);
            Dimension dimLon = ncFileOut.addDimension(null, "longitude",GRID_NUM_LON);
            Dimension dimLat = ncFileOut.addDimension(null, "latitude", GRID_NUM_LAT);

            List<Dimension> dimTimeD = new ArrayList<Dimension>();
            dimTimeD.add(dimTime);

            List<Dimension> dimLatD = new ArrayList<Dimension>();
            dimLatD.add(dimLat);

            List<Dimension> dimLonD = new ArrayList<Dimension>();
            dimLonD.add(dimLon);
            
            
            List<Dimension> dimGridD = new ArrayList<Dimension>();
            dimGridD.add(dimTime);
            dimGridD.add(dimLat);
            dimGridD.add(dimLon);

            List<Dimension> dimGrid2 = new ArrayList<Dimension>();
            dimGrid2.add(dimLat);
            dimGrid2.add(dimLon);
                        
            // define corridinates variable variables
            Variable vlon = ncFileOut.addVariable(null, "longitude", DataType.FLOAT, dimLonD);
            Variable vlat = ncFileOut.addVariable(null, "latitude", DataType.FLOAT, dimLatD);
            Variable vyear = ncFileOut.addVariable(null, "fyear", DataType.INT, dimTimeD);
            Variable vmonth = ncFileOut.addVariable(null, "fmonth", DataType.INT, dimTimeD);
            Variable vft = ncFileOut.addVariable(null, "ft", DataType.FLOAT, dimTimeD);
            vlat.addAttribute(new Attribute("units", "degree"));
            vlon.addAttribute(new Attribute("units", "degree"));
            
            // define data variables
            // add BT_ASC & BT_DSC variable & attributes    
            String varName = MltCdrConstant.VARNAME_BRIGHTNESS_TEMPERATURE_ASCENDING;
            Variable varBtAsc = ncFileOut.addVariable(null, varName , DataType.FLOAT, dimGridD);
            varBtAsc.addAttribute(new Attribute("units","Kelvin" ));
            varBtAsc.addAttribute(new Attribute("valid_max", AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE));
            varBtAsc.addAttribute(new Attribute("valid_min", AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE));
            varBtAsc.addAttribute(new Attribute("_FillValue", AtmsGridConstant.FILLED_VALUE_INVALID_BRIGHTNESS_TEMPERATURE));          
            varBtAsc.addAttribute(new Attribute("coordinates","time latitude longitude" ));

            varName = MltCdrConstant.VARNAME_BRIGHTNESS_TEMPERATURE_DESCENDING;
            Variable varBtDsc = ncFileOut.addVariable(null, varName , DataType.FLOAT, dimGridD);
            varBtDsc.addAttribute(new Attribute("units","Kelvin" ));
            varBtDsc.addAttribute(new Attribute("valid_max", AtmsGridConstant.VALID_MAX_BRIGHTNESS_TEMPERATURE));
            varBtDsc.addAttribute(new Attribute("valid_min", AtmsGridConstant.VALID_MIN_BRIGHTNESS_TEMPERATURE));
            varBtDsc.addAttribute(new Attribute("_FillValue", AtmsGridConstant.FILLED_VALUE_INVALID_BRIGHTNESS_TEMPERATURE));          
            varBtDsc.addAttribute(new Attribute("coordinates","time latitude longitude" ));

            ncFileOut.create();
            
            // write data
            ArrayFloat.D1 dataLat = new ArrayFloat.D1(GRID_NUM_LAT);
            ArrayFloat.D1 dataLon = new ArrayFloat.D1(GRID_NUM_LON);      
            ArrayInt.D1 dataFyear = new ArrayInt.D1(this.numMonths);
            ArrayInt.D1 dataFmonth = new ArrayInt.D1(this.numMonths);
            ArrayFloat.D1 dataFt = new ArrayFloat.D1(this.numMonths);
            
            // write latitude   
            float[] plat = GeoGrid2D.getGridLatitudeFloat();
            for (int i=0; i<GRID_NUM_LAT; i++) {
                dataLat.set(i, plat[i]);
            }
            ncFileOut.write(vlat, dataLat);
            
            // write longitude
            float[] plon = GeoGrid2D.getGridLongitudeFloat();
            for(int i=0; i<GRID_NUM_LON; i++){
                dataLon.set(i, plon[i]);
            }
            ncFileOut.write(vlon, dataLon);
            
            // write year 
            int[] fyear = this.dataL4.getFyear();
            for(int i=0; i<this.numMonths; i++) {
                dataFyear.set(i, fyear[i]);
            }
            ncFileOut.write(vyear, dataFyear);
            
            // write month
            int[] fmonth = this.dataL4.getFmonth();
            for(int i=0; i<this.numMonths; i++) {
                dataFmonth.set(i, fmonth[i]);
            }
            ncFileOut.write(vmonth, dataFmonth);
            
            // write time
            float[] ft = this.dataL4.getFt();
            for(int i=0; i<this.numMonths; i++) {
                dataFt.set(i, ft[i]);
            }
            ncFileOut.write(vft, dataFt);
            
            // write bt_asc
            float[][][]bt_asc = this.dataL4.getBtDataAsc();
            ArrayFloat.D3 vdata = new ArrayFloat.D3(this.numMonths, GRID_NUM_LAT, GRID_NUM_LON);
            for (int i=0; i<this.numMonths; i++) {
                for(int j=0; j<GRID_NUM_LAT; j++) {
                    for(int k=0; k<GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, bt_asc[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varBtAsc, vdata);
            
            // write bt_dsc
            float[][][]bt_dsc = this.dataL4.getBtDataDsc();
            vdata = new ArrayFloat.D3(this.numMonths, GRID_NUM_LAT, GRID_NUM_LON);
            for (int i=0; i<this.numMonths; i++) {
                for(int j=0; j<GRID_NUM_LAT; j++) {
                    for(int k=0; k<GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, bt_dsc[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varBtDsc, vdata);
            
            if (ncFileOut != null) {
                try {
                    ncFileOut.close();
                    ret = true;
                } catch (Exception exp) {
                    errLogger.error("Failed to close netcdf file" + this.filenameL4);
                    exp.printStackTrace();
                    ret = true;
                }
            }
        }catch(Exception e) {
            e.printStackTrace();
            String errMsg = String.format("failed to write %s ATMS channel %d L4 data to file", this.satName, this.channel);
            errLogger.error(errMsg);
            if (ncFileOut != null) {
                try {
                    ncFileOut.close();
                } catch (Exception exp) {
                    exp.printStackTrace();
                }  
            }
            ret = false;
        }
        
        return ret;
    }
    
    
    /**
     * Get short name of the ATMS L4 data file
     * @return short name of the ATMS L4 data file
     */
    final public static String getAtmsL4Filename(String satName, int channel){
        int index = JPSSConstant.getJpssSatelliteIndex(satName);
        String tag = JPSSConstant.getJpssFileTag(index);
        String filename = String.format("ATMS_L4_%s_ch%02d.nc", tag, channel);        
        return filename;
    }
    
}
