/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.util.TimeUtility.java
*
* NAME
*   atms.util.TimeUtility.java
*
* PURPOSE
*   To process time related operations
* 
* DESCRIPTION
*	This class is used to process time informaton and format conversion
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package atms.util;

import java.util.Date;
import java.time.Instant;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoField;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.text.SimpleDateFormat;

import java.util.Calendar;
import java.util.GregorianCalendar;

import static java.time.temporal.ChronoUnit.DAYS;

import org.threeten.extra.scale.TaiInstant;
import org.threeten.extra.scale.UtcInstant;

/**
 * Utility class to process date and time.
 *   This class provides tools to 
 *     convert time in format of seconds since 1958 to a vector of year, month, day, hour, minute, and seconds;
 *     get previous day of the specified year, month and day;
 *     get next day of the specified year, month and day;
 *     convert calendar day to Julian day;
 *     convert Julian day to calendar day.
 * 
 * @author Xianjun Hao, George Mason University < 
 *
*/
public class TimeUtility {
    /**
     * An static instance of Date.
     */	 
    final private static Date date=new Date();
	
	/**
	 * An Instant instance for the time  1958-01-01 T00:00:00Z.
	 */
    private Instant dateIET58 = null;         

	/**
	 * Seconds of "1958-01-01 00:00:00Z" since 1970-01-01 00:00:00Z. 
	 */
    private long offset_in_seconds = 0;    // seconds of "1958-01-01 00:00:00Z" since 1970.
    
    /**
     * Construct an instance of TimeUnitity.
     */
    public TimeUtility(){
        DateTimeFormatter formatter = DateTimeFormatter.ISO_INSTANT;
        try {
            this.dateIET58 = Instant.parse("1958-01-01T00:00:00Z");
            this.offset_in_seconds=this.dateIET58.getEpochSecond();
        } catch (Exception e){
            e.printStackTrace();
            this.dateIET58=null;
            this.offset_in_seconds=0;
        }
    }
                
    /**
     * Convert time in format of seconds since 1958 to a vector of year, month, day, hour, minute, and seconds.
     * @param secondsSince58, time in seconds since 1958-01-01 T00:00:00Z.
     * @return Time vector of year, month, day, hour, minute, and seconds.
     */
    public  int[] getTimeVec(long secondsSince58){
        Instant ins = this.dateIET58.plusSeconds(secondsSince58);
        LocalDateTime tm = LocalDateTime.ofInstant(ins, ZoneId.of("UTC"));
        int[] tmv = new int[6];
        tmv[0]=tm.get(ChronoField.YEAR);
        tmv[1]=tm.get(ChronoField.MONTH_OF_YEAR);
        tmv[2]=tm.get(ChronoField.DAY_OF_MONTH);
        tmv[3]=tm.get(ChronoField.HOUR_OF_DAY);
        tmv[4]=tm.get(ChronoField.MINUTE_OF_HOUR);
        tmv[5]=tm.get(ChronoField.SECOND_OF_MINUTE);
        return tmv;
    }

    /**
     * Convert time in format of days since 0000-01-01 to a vector of year, month, day.
     * @param daysSince, time in days since 1900-01-01 T00:00:00Z.
     * @return Time vector of year, month, day.
     */
    public int[] getDateVec(long daysSince){
        long num = getDateNumber(1958, 1,1);
        long secondsSince58 = (daysSince- num)*24*3600;
        Instant ins = this.dateIET58.plusSeconds(secondsSince58);
        LocalDateTime tm = LocalDateTime.ofInstant(ins, ZoneId.of("UTC"));
        int[] tmv = new int[3];
        tmv[0]=tm.get(ChronoField.YEAR);
        tmv[1]=tm.get(ChronoField.MONTH_OF_YEAR);
        tmv[2]=tm.get(ChronoField.DAY_OF_MONTH);
        return tmv;
    }
    

    
/**
 * Get the previous day of a calendar day specified with year, month and day.
 * @param year, year.
 * @param month, month.
 * @param day, day.
 * @return previous day specified with a vector of year, month and day.
 */
    public static int[] getPreviousDay(int year, int month, int day){
        int[] tmv = new int[3];
        //GregorianCalendar, month starts with 0
        Calendar cal = new GregorianCalendar(year, month-1, day);
        //System.out.println("DOY:"+cal.get(Calendar.DAY_OF_YEAR));
        cal.add(Calendar.DAY_OF_YEAR,-1);
        tmv[0]=cal.get(Calendar.YEAR);
        tmv[1]=cal.get(Calendar.MONTH)+1;      // Calendar, month starts with 0, need to add 1 to correct it
        tmv[2]=cal.get(Calendar.DAY_OF_MONTH);        
        return tmv;
    }

/**
 * Get the next day of a calendar day specified with year, month and day.
 * @param year, year.
 * @param month, month.
 * @param day, day.
 * @return next day specified with a vector of year, month and day.
 */
    public static int[] getNextDay(int year, int month, int day){
        int[] tmv = new int[3];
        //GregorianCalendar, be careful that month starts with 0 in GregorianCalendar.
        Calendar cal = new GregorianCalendar(year, month-1, day);
        cal.add(Calendar.DAY_OF_YEAR,1);
        tmv[0]=cal.get(Calendar.YEAR);
		//  Month starts with 0 GregorianCalendar, need to add 1 to correct it for general use.
        tmv[1]=cal.get(Calendar.MONTH)+1;      
        tmv[2]=cal.get(Calendar.DAY_OF_MONTH);        
        return tmv;
    }
    
    /**
     * Convert calendar day to Julian day (starting with 1).
     * @param year
     * @param month
     * @param day
     * @return Julian day
     */
    public static int getDayofYear(int year, int month, int day){
        int jday =0;
        Calendar cal = new GregorianCalendar(year, month-1, day);
        jday = cal.get(Calendar.DAY_OF_YEAR);        
        return jday;
    }
    
    /**
     * Convert Julian day to calendar day.
     * @param year
     * @param jday
     * @return a vector with year, month, day
     */
    public static int[] getCalDay(int year, int jday){
        int[] tmv=new int[3];
        Calendar cal = new GregorianCalendar(year, 0, 1);
        cal.add(Calendar.DAY_OF_YEAR,jday-1);
        tmv[0]=cal.get(Calendar.YEAR);
		// Month starts with 0 GregorianCalendar, need to add 1 to correct it for general use.
        tmv[1]=cal.get(Calendar.MONTH)+1;      
        tmv[2]=cal.get(Calendar.DAY_OF_MONTH);        
       
        return tmv;
    }
    
    /**
     * Get the start time of a day in seconds since 1958-01-01 T00:00:00Z.
     * @param year
     * @param jday
     * @return start time of a day in seconds since 1958-01-01 T00:00:00Z.
     */
    public static long getStartSecondSince58(int year, int month, int day){
        long ret = 0;
        ZoneId zone;
        zone = ZoneId.of("UTC");
        ZonedDateTime zdt  = ZonedDateTime.of(year, month, day, 0, 0, 0, 0, zone);
        ZonedDateTime zdt0 = ZonedDateTime.of(1958, 1, 1, 0, 0, 0, 0, zone);
        ret = zdt.toEpochSecond() - zdt0.toEpochSecond();
        return ret;
    }

    /**
     * Get the start time of a day in seconds since 1958-01-01 T00:00:00Z.
     * @param year
     * @param jday
     * @return start time of a day in seconds since 1958-01-01 T00:00:00Z.
     */
    public static long getStartSecondSince58(int year, int month, int day, int hour, int minute, int sec, int tsec){
        long ret = 0;
        
        ZoneId zone;
        zone = ZoneId.of("UTC");
        int nanosec = tsec*100000000;
        ZonedDateTime zdt  = ZonedDateTime.of(year, month, day, hour, minute, sec, nanosec, zone);
        ZonedDateTime zdt0 = ZonedDateTime.of(1958, 1, 1, 0, 0, 0, 0, zone);
        ret = zdt.toEpochSecond() - zdt0.toEpochSecond();
        return ret;
    }
    
    
    /**
     * Get number of days since 0000-01-01.  0000-01-01 returns 1.
     * 
     * @param year
     * @param month
     * @param day
     * @return number of days since 0000-01-01 00:00:00 
     */
    public static long getDateNumber(int year, int month, int day){
        long num = 0;
        LocalDate dateBefore = LocalDate.of(0, 1, 1);
        LocalDate dateAfter = LocalDate.of(year, month, day);
        num = DAYS.between(dateBefore, dateAfter) + 1;
        return num;
    }

    /**
     *  Get TAI time
     */
    public static long getTaiSeconds(int kyear, int kmonth, int kday, int khour, int kminute, int ksec){
        long ts = -1;
        try{
            String str=String.format("%04d-%02d-%02dT%02d:%02d:%02dZ", kyear, kmonth, kday, khour, kminute, ksec);
            Instant te = Instant.parse(str);
            TaiInstant tt = TaiInstant.of(te);
            ts = tt.getTaiSeconds();
        }catch (Exception e) {
            e.printStackTrace();
        }
        return ts;
    }

    public static String  getCalDayofTai(long ts) {
        String str = null;
        
        try{
            TaiInstant tt = TaiInstant.ofTaiSeconds(ts, 0);
            UtcInstant tu = tt.toUtcInstant();
            str= tu.toString();
        }catch (Exception e) {
            e.printStackTrace();
        }
        return str;
    }

// testing 1    
    public static void test1() {
        long jday0 = getDateNumber(0,1,1);
        System.out.println("jday0=" + jday0);
        
        TimeUtility tmApp = new TimeUtility();
        int year1 = 2005;
        int month1 = 1;
        int day1  = 1;
        long jday1 =  TimeUtility.getDateNumber(year1, month1, day1);
        int[] tm1 = tmApp.getDateVec(jday1);
        String msg  = String.format("%4d-%02d-%02d -> %d ->%4d-%02d-%02d", year1, month1, day1, jday1, tm1[0], tm1[1], tm1[2]); 
        System.out.println(msg);

        int year2 = 2016;
        int month2 = 12;
        int day2  = 31;
        long jday2 =  TimeUtility.getDateNumber(year2, month2, day2);
        int[] tm2 = tmApp.getDateVec(jday2);
        msg  = String.format("%4d-%02d-%02d -> %d ->%4d-%02d-%02d", year2, month2, day2, jday2, tm2[0], tm2[1], tm2[2]); 
        System.out.println(msg);        
    }
    
    // testing starting and ending day of selected month
    public static void test2(int kyear, int kmonth){
        long jday1 =  TimeUtility.getDateNumber(kyear, kmonth,1);
        long jday2 = 0;
        
        if (kmonth<12) {
           jday2 =  TimeUtility.getDateNumber(kyear, kmonth+1,1) - 1;
        } else {
           jday2 =  TimeUtility.getDateNumber(kyear+1, 1,1) - 1;            
        } 
        TimeUtility tmApp = new TimeUtility();
        
        int[] tm1 = tmApp.getDateVec(jday1);
        int[] tm2 = tmApp.getDateVec(jday2);
        String msg  = String.format("%4d-%02d: %4d-%02d-%02d -> %4d-%02d-%02d", kyear, kmonth,  tm1[0], tm1[1], tm1[2], tm2[0], tm2[1], tm2[2]); 
        System.out.println(msg);
    }
    
    // testing starting and ending seconds of selected month
    public static void test3(int kyear, int kmonth){
        long jday1 =  TimeUtility.getDateNumber(kyear, kmonth,1);
        long jday2 = 0;
        long jsec1 = 0; 
        long jsec2 = 0;
        
        if (kmonth<12) {
           jday2 =  TimeUtility.getDateNumber(kyear, kmonth+1,1) ;
        } else {
           jday2 =  TimeUtility.getDateNumber(kyear+1, 1,1) ;            
        } 
        TimeUtility tmApp = new TimeUtility();
        
        int[] tm1 = tmApp.getDateVec(jday1);
        int[] tm2 = tmApp.getDateVec(jday2);
        
        jsec1 = TimeUtility.getStartSecondSince58(tm1[0], tm1[1], tm1[2]);
        jsec2 = TimeUtility.getStartSecondSince58(tm2[0], tm2[1], tm2[2]);
        
        String msg  = String.format("%4d-%02d: %4d-%02d-%02d -> %4d-%02d-%02d", kyear, kmonth,  tm1[0], tm1[1], tm1[2], tm2[0], tm2[1], tm2[2]); 
        System.out.println(msg);
        msg  = String.format("%4d-%02d: %d ->  %d", kyear, kmonth,  jsec1, jsec2); 
        System.out.println(msg);
    }
    
    // test seconds since 1958 for ATMS data
    public static long test4(int kyear, int kmonth, int kday, int khour, int kminute, int ksec, int tsec) {
        long jsec = TimeUtility.getStartSecondSince58(kyear, kmonth, kday, khour, kminute, ksec, tsec);
        System.out.println("UTC seconds: "+ jsec);
        return jsec;
    }
    
    
    /**
     *  Test TAI time
     */
    public static void testTai(int kyear, int kmonth, int kday, int khour, int kminute, int ksec, int tsec){
        try{
            String str=String.format("%04d-%02d-%02dT%02d:%02d:%02dZ", kyear, kmonth, kday, khour, kminute, ksec);
            System.out.println(str);
            Instant te = Instant.parse(str);
            TaiInstant tt = TaiInstant.of(te);
            long ts = tt.getTaiSeconds();
            System.out.println(ts);
        }catch (Exception e) {
            e.printStackTrace();
        }
    }
        
    
}

