/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.cdr.MltCdrConstant.java
*
* NAME
*   atms.cdr.MltCdrConstant.java
*
* PURPOSE
*   To define constants in MLT CDR processing
* 
* DESCRIPTION
*	This class is used to process time informaton and format conversion
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package cdr;

import java.time.LocalDate;
import java.time.Month;
/**
 * Constant definitions for CDR data products.
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class MltCdrConstant {
    /**
     * ATMS channel for TMT
     */
    final public static int TMT_CHANNEL_INDEX_ATMS = 6; 

    /**
     * ATMS channel for TUT
     */
    final public static int TUT_CHANNEL_INDEX_ATMS = 8; 

    /**
     * ATMS channel for TLS
     */    
    final public static int TLS_CHANNEL_INDEX_ATMS = 10;

    /**
     * AMSU-A channel for TMT
     */    
    final public static int TMT_CHANNEL_INDEX_AMSUA = 5; 

    /**
     * AMSU-A channel for TUT
     */        
    final public static int TUT_CHANNEL_INDEX_AMSUA = 7; 

    /**
     * AMSU-A channel for TLS
     */    
    final public static int TLS_CHANNEL_INDEX_AMSUA = 9;
    
    /**
     * MSU channel for TMT
     */        
    final public static int TMT_CHANNEL_INDEX_MSU = 2; 

    /**
     * MSU channel for TUT
     */        
    final public static int TUT_CHANNEL_INDEX_MSU = 3; 

    /**
     * MSU channel for TLS
     */        
    final public static int TLS_CHANNEL_INDEX_MSU = 4;

    /**
     * Variable name of brightness temperature for ascending orbits
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_ASCENDING = "bt_asc";

    /**
     * Variable name of brightness temperature for descending orbits
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_DESCENDING = "bt_dsc";

    /**
     * Variable name of RTMT
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTMT = "RTMT";

    /**
     * Variable name of RTMT climatology
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTMT_CLIMATOLOGY = "RTMT_Climatology";
    
    /**
     * Variable name of RTMT anomaly
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTMT_ANOMALY = "RTMT_Anomaly";

     /**
     * Variable name of RTUT
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTUT = "RTUT";

     /**
     * Variable name of RTUT climatology
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTUT_CLIMATOLOGY = "RTUT_Climatology";

     /**
     * Variable name of RTUT anomaly
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTUT_ANOMALY = "RTUT_Anomaly";

     /**
     * Variable name of RTLS
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTLS = "RTLS";

     /**
     * Variable name of RTLS climatology
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTLS_CLIMATOLOGY = "RTLS_Climatology";

     /**
     * Variable name of RTLS anomaly
     */
    final public static String VARNAME_BRIGHTNESS_TEMPERATURE_RTLS_ANOMALY = "RTLS_Anomaly";
 
     /**
     * Start year of CDR base period for climatology
     */
    final public static int MLT_CDR_CLIMATOLOGY_YEAR_START = 2000;

     /**
     * End year of CDR base period for climatology
     */
     final public static int MLT_CDR_CLIMATOLOGY_YEAR_END= 2019;
            
    /**
     * Get corresponding ATMS channel number from channel name.
     * 
     * @param name, channel name (TMT, TUT, TLS)
     * @return channel index
     */
    final public static int getAtmsChannelIndex(String name) {
        int index = -1;
        if (name.equalsIgnoreCase("TMT")) {
            index = TMT_CHANNEL_INDEX_ATMS;
        } else if (name.equalsIgnoreCase("TUT")){
            index = TUT_CHANNEL_INDEX_ATMS;            
        } else if (name.equalsIgnoreCase("TLS")){
            index = TLS_CHANNEL_INDEX_ATMS;            
        } else {
            index = -1;
        }        
        return index;
    }

    /**
     * Get corresponding AMSUA channel number from channel name.
     * 
     * @param name, channel name (TMT, TUT, TLS)
     * @return channel index
     */
    final public static int getAmsuaChannelIndex(String name) {
        int index = -1;
        if (name.equalsIgnoreCase("TMT")) {
            index = TMT_CHANNEL_INDEX_AMSUA;
        } else if (name.equalsIgnoreCase("TUT")){
            index = TUT_CHANNEL_INDEX_AMSUA;            
        } else if (name.equalsIgnoreCase("TLS")){
            index = TLS_CHANNEL_INDEX_AMSUA;            
        } else {
            index = -1;
        }        
        return index;
    }

   /**
     * Get corresponding MSU channel number from channel name.
     * 
     * @param name, channel name (TMT, TUT, TLS)
     * @return channel index
     */
    final public static int getMsuChannelIndex(String name) {
        int index = -1;
        if (name.equalsIgnoreCase("TMT")) {
            index = TMT_CHANNEL_INDEX_MSU;
        } else if (name.equalsIgnoreCase("TUT")){
            index = TUT_CHANNEL_INDEX_MSU;            
        } else if (name.equalsIgnoreCase("TLS")){
            index = TLS_CHANNEL_INDEX_MSU;            
        } else {
            index = -1;
        }        
        return index;
    }

    /**
     * Get channel name from ATMS channel index ( for 6:TMT, 8:TUT, 10:TLS only).
     * @param index, channel number
     * @return channel name (TMT, TUT or TLS)
     */
    final public static String getNameFromAtmsChannelIndex (int index) {
        String name;        
        switch(index) {
            case TMT_CHANNEL_INDEX_ATMS:
                name="TMT";
                break;
            case TUT_CHANNEL_INDEX_ATMS:
                name="TUT";
                break;
            case TLS_CHANNEL_INDEX_ATMS:
                name="TLS";
                break;
            default:
                name=null;
        }               
        return name;
    }

    /**
     * Get channel name from AMSU-A channel index ( for 6:TMT, 8:TUT, 10:TLS only).
     * @param index, channel number
     * @return channel name (TMT, TUT or TLS)
     */
    final public static String getNameFromAmsuaChannelIndex (int index) {
        String name;        
        switch(index) {
            case TMT_CHANNEL_INDEX_AMSUA:
                name="TMT";
                break;
            case TUT_CHANNEL_INDEX_AMSUA:
                name="TUT";
                break;
            case TLS_CHANNEL_INDEX_AMSUA:
                name="TLS";
                break;
            default:
                name=null;
        }               
        return name;
    }

    /**
     * Get channel name from MSU channel index ( for 6:TMT, 8:TUT, 10:TLS only).
     * @param index, channel number
     * @return channel name (TMT, TUT or TLS)
     */
    final public static String getNameFromMsuChannelIndex (int index) {
        String name;        
        switch(index) {
            case TMT_CHANNEL_INDEX_MSU:
                name="TMT";
                break;
            case TUT_CHANNEL_INDEX_MSU:
                name="TUT";
                break;
            case TLS_CHANNEL_INDEX_MSU:
                name="TLS";
                break;
            default:
                name=null;
        }               
        return name;
    }

/**
 * Get official filename for the MLT CDR data products
 * @param varname, data name (TLT, TMT, TUT, TLS)
 * @param year1, start year
 * @param month1, start month
 * @param year2, end year
 * @param month2, end month
 * @param cyear, creation year
 * @param cmonth, creation month
 * @param cday, creation day
 * @return filename
 */    
    public static String getMltDataFilename(String varname, int year1, int month1, int year2, int month2, int cyear, int cmonth, int cday) {
        String filename = String.format("Mean-Layer-Temperature-NOAA_v05r00_%s_S%4d%02d_E%4d%02d_C%4d%02d%02d.nc", varname, year1, month1,  year2, month2, cyear, cmonth, cday);
        return filename;
    }
    
/**
 * Get official filename for the MLT CDR data products
 * @param varname, data name (TLT, TMT, TUT, TLS)
 * @param year1, start year
 * @param month1, start month
 * @param year2, end year
 * @param month2, end month
 * @param cyear, creation year
 * @param cmonth, creation month
 * @param cday, creation day
 * @return filename
 */    
    public static String getMltDataFilename(String varname, int year1, int month1, int year2, int month2) {
        int cyear, cmonth, cday;
        LocalDate currentdate = LocalDate.now();
        cyear = currentdate.getYear();
        cmonth = currentdate.getMonthValue();
        cday = currentdate.getDayOfMonth();
        String filename = String.format("Mean-Layer-Temperature-NOAA_v05r00_%s_S%4d%02d_E%4d%02d_C%4d%02d%02d.nc", varname, year1, month1, year2, month2, cyear, cmonth, cday);
        return filename;
    }
    
   /**
    *  Get short name of the RefCDR file
    * @param varname, variable name (TMT, TUT, TLS)
    * @param year1, start year of the data
    * @param month1, start month of the data
    * @param year2, end year of the data
    * @param month2, end month of the data
    * @return 
    */
    public static String  getRefCDRFileShortname(String varname, int year1, int month1,  int year2, int month2){
        String   filename = String.format("NESDIS-STAR_R%s_Anomaly_AMSUA_ATMS_V01_S%4d%02d_E%4d%02d.nc", varname, year1, month1, year2, month2);                    
        
        return filename;                
    }
    
    
}
