/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.cdr.MltCdrV5.java
*
* NAME
*   atms.cdr.MltCdrV5.java
*
* PURPOSE
*   To generate the final MLT CDR product
* 
* DESCRIPTION
*	This class is the main class for MLT CDR generation
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package cdr;

import java.io.File;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 * Main class to geenarte MLT CDR V5
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class MltCdrV5 {
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    
    /**
     *  Year of the end time
     */
    private int yearEnd;
    
    /**
     *  Month of the end time
     */
    private int monthEnd;
    
    /**
     * Full path of the TMT data file
     */
    private String fileTMT;
    
    /**
     * Full path of the TUT data file
     */
    private String fileTUT;

    /**
     * Full path of the TLS data file
     */
    private String fileTLS;
    
    /**
     * Full path of the TLT data file
     */
    private String fileTLT;
    
    /**
     * Task configuration
     */
    private MltCdrTaskConfig taskConf;
    
    /**
     * Constructor. 
     * @param taskConf, task configuration. 
     */
    public MltCdrV5(MltCdrTaskConfig taskConf ){
        this.taskConf = taskConf;
        this.yearEnd = taskConf.yearEnd;
        this.monthEnd = taskConf.monthEnd;
    }
        
    
    /**
     * Produce CDR Product
     */
    public void produceCDR(){
        MltTMT objTmt  = new MltTMT(taskConf);
        objTmt.mergeCDR();
        this.fileTMT = objTmt.getCdrFileFullPath();
        infoLogger.info("Generated MLT TMT V5 data file " + this.fileTMT);

        MltTUT objTut  = new MltTUT(taskConf);
        objTut.mergeCDR();
        this.fileTUT = objTut.getCdrFileFullPath();
        infoLogger.info("Generated MLT TUT V5 data file " + this.fileTUT);
        
        MltTLS objTls  = new MltTLS(taskConf);
        objTls.mergeCDR();
        this.fileTLS = objTls.getCdrFileFullPath();
        infoLogger.info("Generated MLT TLS V5 data file " + this.fileTLS);
        
        MltTLT objTlt = new MltTLT(taskConf, objTmt, objTut, objTls);
        objTlt.mergeCDR();
        this.fileTLT = objTlt.getCdrFileFullPath();
         infoLogger.info("Generated MLT TLT V5 data file " + this.fileTLT);
       
    }
    
    /**
     * Main method to generate MLT CDR V5 products
     * @param args 
     */
    public static void main(String[] args){
        int year2, month2;
        String dirLut, dirL4, dirRefCDR, dirCDR;
                
        if (args.length <6){
            year2 = 2023;
            month2 = 12;
            dirLut = "/home/xhao/data/mltcdrv5/lut/";
            dirL4 = "/home/xhao/data/mltcdrv5/data/merg/";
            dirRefCDR = "/home/xhao/data/mltcdrv5/data/rfcdr/";            
            dirCDR = "/home/xhao/data/mltcdrv5/data/cdr/";            
        }  else {
            int k = 0;
            year2 = Integer.parseInt(args[k++]);
            month2 = Integer.parseInt(args[k++]);
            dirLut = args[k++];
            dirL4 = args[k++];
            dirRefCDR = args[k++];
            dirCDR = args[k++];
        }
        MltCdrTaskConfig taskConf = new MltCdrTaskConfig(year2, month2,dirLut,dirL4, dirRefCDR, dirCDR);
        MltCdrV5 app = new MltCdrV5(taskConf);
        app.produceCDR();
    }        
    
}
