/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.cdr.MltFileNetCDF.java
*
* NAME
*   atms.cdr.MltFileNetCDF.java
*
* PURPOSE
*   To write the MLT CDR data to netcdf file
* 
* DESCRIPTION
*	This class writes MLT CDR data to netcdf files 
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */

package cdr;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.TimeZone;
import java.util.Date;

import ucar.nc2.Variable;
import ucar.ma2.*;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriter;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import geogrid.Geo2dTimeSeries;
import geogrid.GeoGrid2D;
import geogrid.GeoGridConstant;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;

import atms.util.TimeUtility;

/**
 * Class to control the file format and attributes of MLT CDR, and write data to netcdf file.
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class MltFileNetCDF {
    
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    
    /**
     * CDR variable name
     */
    private String cdrVarName = null;
    
    /**
     * Name of temperature in netcdf file
     */
    private String cdrVariableBt = null;

   /**
     * Name of temperature climatology in netcdf file
     */
    private String cdrVariableClim = null;

   /**
     * Name of temperature anomaly in netcdf file
    */
    private String cdrVariableAnom = null;

   /**
     * variable ID
     */
    private int cdrVarId = -1;
   
    /**
     * netcdf file writer
     */
    private NetcdfFileWriter ncWriter = null;
    
    /**
     * Data time series (CDR)
     */
    Geo2dTimeSeries geoData = null;

    /**
     *  Directory of CDR data
     */
    private String dirCDR;
    
    /**
     *  Full path of the CDR data file
     */    
    private String fileFullpath;

    /**
     * Short name of the CDR data file
     */    
    private String fileShortname;

    /**
     *  Number of months
     */    
    private int numberMonths;
    
    /**
     * Year of the start time
     */
    private int year1;

    /**
     * Month of the start time
     */
    private int month1;

    /**
     * Year of the end time
     */
    private int year2;
    
    /**
     * Month of the end time
     */
     private int month2;
   
    /**
     * Constructor
     * @param cdrVarName, name of CDR variable
     * @param dirCDR, directory of CDR data
     * @param geoData, data time series 
     */
    MltFileNetCDF(String cdrVarName, String dirCDR, Geo2dTimeSeries geoData) {
        this.dirCDR = dirCDR;
        this.geoData = geoData;
        if (cdrVarName.equalsIgnoreCase("TLT")) {
            this.cdrVarName = "TLT";
            this.cdrVarId = 0;
        } else if  (cdrVarName.equalsIgnoreCase("TMT")) {
            this.cdrVarName = "TMT";
            this.cdrVarId = 1;            
        } else if  (cdrVarName.equalsIgnoreCase("TUT")) {
            this.cdrVarName = "TUT";
            this.cdrVarId = 2;            
        } else if  (cdrVarName.equalsIgnoreCase("TLS")) {
            this.cdrVarName = "TLS";
            this.cdrVarId = 3;            
        }  else {
            this.cdrVarName = null;
            this.cdrVarId = -1;
        }
        
        int[] tyear = geoData.getFyear();
        int[] tmonth = geoData.getFmonth();
        this.numberMonths = geoData.getTimeSeriesLength();
        this.year1 = tyear[0];
        this.month1 = tmonth[0];
        this.year2 = tyear[this.numberMonths-1];
        this.month2 = tmonth[this.numberMonths-1];
        if (this.cdrVarId >= 0){
            this.cdrVariableBt = String.format("tcdr_MSU_AMSUA_ATMS_%s", this.cdrVarName);
            this.cdrVariableAnom = String.format("tcdr_MSU_AMSUA_ATMS_%s_anomaly", this.cdrVarName);
            this.cdrVariableClim = String.format("tcdr_MSU_AMSUA_ATMS_%s_climatology", this.cdrVarName);
            this.fileShortname = MltCdrConstant.getMltDataFilename(this.cdrVarName, year1, month1, year2, month2);
            this.fileFullpath = this.dirCDR + this.fileShortname; 
            infoLogger.info("start generating CDR file " + this.fileFullpath);
        } else {
            this.fileShortname = null;
            this.fileFullpath = null;
            errLogger.error("Failed to start processing CDR product " + cdrVarName);
        }
    }
    
    
    /**
     * Get full path of the CDR file
     * @return full path of the CDR file 
     */
    public String getCdrFilename(){
        return this.fileFullpath;
    }
    
    /**
     * Write data to NetCDF file.
     */
    public void writeDataToNetcdf(){
        int i, j, k, kmonth, kyear;
        double tm0, tm1, tm2;
        if (this.cdrVarId < 0){
            errLogger.error("Invalid CDR data, can't write to netcdf file");
            return;
        }
        this.numberMonths = this.geoData.getTimeSeriesLength();
        try{
            ncWriter = NetcdfFileWriter.createNew(NetcdfFileWriter.Version.netcdf3, this.fileFullpath);
            // create Dimensions
            Dimension dimTime = ncWriter.addDimension(null, "time", this.numberMonths);
            Dimension dimLon = ncWriter.addDimension(null, "longitude",GEO_GRID_NUM_LON);
            Dimension dimLat = ncWriter.addDimension(null, "latitude", GEO_GRID_NUM_LAT);
            Dimension dimBound = ncWriter.addDimension(null, "bounds", 2);
            Dimension dimMonth = ncWriter.addDimension(null, "month_of_year",12);

            List<Dimension> dimTimeD = new ArrayList<Dimension>();
            dimTimeD.add(dimTime);

            List<Dimension> dimLatD = new ArrayList<Dimension>();
            dimLatD.add(dimLat);

            List<Dimension> dimLonD = new ArrayList<Dimension>();
            dimLonD.add(dimLon);
            
            List<Dimension> dimMonD = new ArrayList<Dimension>();
            dimMonD.add(dimMonth);
            
            List<Dimension> dimBoundD = new ArrayList<Dimension>();
            dimBoundD.add(dimBound);

            List<Dimension> dimGridD = new ArrayList<Dimension>();
            dimGridD.add(dimTime);
            dimGridD.add(dimLat);
            dimGridD.add(dimLon);
            
            List<Dimension> dimGridClim = new ArrayList<Dimension>();
            dimGridClim.add(dimMonth);
            dimGridClim.add(dimLat);
            dimGridClim.add(dimLon);

            List<Dimension> dimBoundTime = new ArrayList<Dimension>();
            dimBoundTime.add(dimTime);
            dimBoundTime.add(dimBound);

            List<Dimension> dimBoundLat = new ArrayList<Dimension>();
            dimBoundLat.add(dimLat);
            dimBoundLat.add(dimBound);
            
            List<Dimension> dimBoundLon = new ArrayList<Dimension>();
            dimBoundLon.add(dimLon);
            dimBoundLon.add(dimBound);
        // create variables
            Variable vlon = this.ncWriter.addVariable(null, "longitude", DataType.FLOAT, dimLonD);
            vlon.addAttribute(new Attribute("units", "degrees_east"));
            vlon.addAttribute(new Attribute("standard_name", "longitude"));
            vlon.addAttribute(new Attribute("valid_min", -180.0f));
            vlon.addAttribute(new Attribute("valid_max", 180.0f));
            vlon.addAttribute(new Attribute("bounds", "longitude_bounds"));
            

            Variable vlat = this.ncWriter.addVariable(null, "latitude", DataType.FLOAT, dimLatD);
            vlat.addAttribute(new Attribute("units", "degrees_north"));
            vlat.addAttribute(new Attribute("standard_name", "latitude"));
            vlat.addAttribute(new Attribute("valid_min", -90.0f));
            vlat.addAttribute(new Attribute("valid_max", 90.0f));
            vlat.addAttribute(new Attribute("bounds", "latitude_bounds"));
            
            Variable vtime = this.ncWriter.addVariable(null, "time", DataType.DOUBLE, dimTimeD);
            vtime.addAttribute(new Attribute("units", "days since 1978-01-01 00:00:00"));
            vtime.addAttribute(new Attribute("standard_name", "time"));
            vtime.addAttribute(new Attribute("valid_min", 0d));
            vtime.addAttribute(new Attribute("valid_max", 60000d));
            vtime.addAttribute(new Attribute("calendar", "gregorian"));
            vtime.addAttribute(new Attribute("bounds", "time_bounds"));
            
            Variable vlonBound = this.ncWriter.addVariable(null, "longitude_bounds", DataType.FLOAT, dimBoundLon);
            vlonBound.addAttribute(new Attribute("units", "degrees_east"));
            vlonBound.addAttribute(new Attribute("standard_name", "longitude"));
            vlonBound.addAttribute(new Attribute("valid_min", -180.0f));
            vlonBound.addAttribute(new Attribute("valid_max", 180.0f));
            vlonBound.addAttribute(new Attribute("long_name", "longitude_bounds"));                   
            
            Variable vlatBound = this.ncWriter.addVariable(null, "latitude_bounds", DataType.FLOAT, dimBoundLat);
            vlatBound.addAttribute(new Attribute("units", "degrees_north"));
            vlatBound.addAttribute(new Attribute("standard_name", "latitude"));
            vlatBound.addAttribute(new Attribute("valid_min", -90.0f));
            vlatBound.addAttribute(new Attribute("valid_max", 90.0f));
            vlatBound.addAttribute(new Attribute("long_name", "latitude_bounds"));
                        
            Variable vtimeBound = this.ncWriter.addVariable(null, "time_bounds", DataType.DOUBLE, dimBoundTime);
            vtimeBound.addAttribute(new Attribute("units", "days since 1978-01-01 00:00:00"));
            vtimeBound.addAttribute(new Attribute("standard_name", "time bounds"));
            vtimeBound.addAttribute(new Attribute("valid_min", 0d));
            vtimeBound.addAttribute(new Attribute("valid_max", 60000d));
            vtimeBound.addAttribute(new Attribute("calendar", "gregorian"));
            vtimeBound.addAttribute(new Attribute("long_name", "time_bounds"));

            // Brightness temperature 
            Variable varBt= this.ncWriter.addVariable(null, this.cdrVariableBt , DataType.FLOAT, dimGridD);
            varBt.addAttribute(new Attribute("units","Kelvin" ));
            varBt.addAttribute(new Attribute("standard_name","brightness_temperature" ));
            switch(this.cdrVarId){
                case 0:
                    varBt.addAttribute(new Attribute("long_name","NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of lower-troposphere" ));
                    break;
                case 1:
                    varBt.addAttribute(new Attribute("long_name","NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of mid-troposphere" ));
                    break;
                case 2:
                    varBt.addAttribute(new Attribute("long_name","NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of upper-troposphere" ));
                    break;
                case 3:
                    varBt.addAttribute(new Attribute("long_name","NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of lower-stratosphere" ));
                    break;                                
            }
            varBt.addAttribute(new Attribute("_FillValue", (float)GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varBt.addAttribute(new Attribute("coordinates","time latitude longitude" ));
            varBt.addAttribute(new Attribute("valid_max", (float)GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE));
            varBt.addAttribute(new Attribute("valid_min", (float)GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE));

            // Anomaly
            Variable varAnom= this.ncWriter.addVariable(null, this.cdrVariableAnom , DataType.FLOAT, dimGridD);
            varAnom.addAttribute(new Attribute("units","Kelvin" ));
            varAnom.addAttribute(new Attribute("standard_name","brightness_temperature_anomaly" ));
            switch(this.cdrVarId){
                case 0:
                    varAnom.addAttribute(new Attribute("long_name","Anomaly of NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of lower-troposphere" ));
                    varAnom.addAttribute(new Attribute("description","Anomaly of monthly mean layer temperature of lower-troposphere, the base period is from January 2000 to December 2019"));
                    break;
                case 1:
                    varAnom.addAttribute(new Attribute("long_name","Anomaly of NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of mid-troposphere" ));
                    varAnom.addAttribute(new Attribute("description","Anomaly of monthly mean layer temperature of mid-troposphere, the base period is from January 2000 to December 2019"));
                    break;
                case 2:
                    varAnom.addAttribute(new Attribute("long_name","Anomaly of NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of upper-troposphere" ));
                    varAnom.addAttribute(new Attribute("description","Anomaly of monthly mean layer temperature of upper-troposphere, the base period is from January 2000 to December 2019"));
                    break;
                case 3:
                    varAnom.addAttribute(new Attribute("long_name","Anomaly of NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of lower-stratosphere" ));
                    varAnom.addAttribute(new Attribute("description","Anomaly of monthly mean layer temperature of lower-stratosphere, the base period is from January 2000 to December 2019"));
                    break;                                
            }
            varAnom.addAttribute(new Attribute("_FillValue", (float)GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varAnom.addAttribute(new Attribute("coordinates","time latitude longitude" ));
            varAnom.addAttribute(new Attribute("valid_max", (float)GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE));
            varAnom.addAttribute(new Attribute("valid_min", (float)GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE));
            
            //climatology
            Variable varClim= this.ncWriter.addVariable(null, this.cdrVariableClim , DataType.FLOAT, dimGridClim);
            varClim.addAttribute(new Attribute("units","Kelvin" ));
            varClim.addAttribute(new Attribute("standard_name","brightness_temperature" ));
            switch(this.cdrVarId){
                case 0:
                    varClim.addAttribute(new Attribute("long_name","Climatology of NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of lower-troposphere" ));
                    varClim.addAttribute(new Attribute("description","Climatology of monthly mean layer temperature of lower-troposphere, the base period is from January 2000 to December 2019"));
                    break;
                case 1:
                    varClim.addAttribute(new Attribute("long_name","Climatology of NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of mid-troposphere" ));
                    varClim.addAttribute(new Attribute("description","Climatology of monthly mean layer temperature of mid-troposphere, the base period is from January 2000 to December 2019"));
                    break;
                case 2:
                    varClim.addAttribute(new Attribute("long_name","Climatology of NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of upper-troposphere" ));
                    varClim.addAttribute(new Attribute("description","Climatology of monthly mean layer temperature of upper-troposphere, the base period is from January 2000 to December 2019"));
                    break;
                case 3:
                    varClim.addAttribute(new Attribute("long_name","Climatology of NOAA Climate Data Record of MSU/AMSU-A/ATMS monthly mean layer temperature of lower-stratosphere" ));
                    varClim.addAttribute(new Attribute("description","Climatology of monthly mean layer temperature of lower-stratosphere, the base period is from January 2000 to December 2019"));
                    break;                                
            }
            varClim.addAttribute(new Attribute("_FillValue", (float)GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varClim.addAttribute(new Attribute("coordinates","month_of_year latitude longitude" ));
            varClim.addAttribute(new Attribute("valid_max", (float)GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE));
            varClim.addAttribute(new Attribute("valid_min", (float)GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE));
            
            // global attributes
            this.defineGlobalAttributes();
            
            // create file
            this.ncWriter.create();
            
            // write data 
            ArrayFloat.D1 dataLat = new ArrayFloat.D1(GEO_GRID_NUM_LAT);
            ArrayFloat.D1 dataLon = new ArrayFloat.D1(GEO_GRID_NUM_LON);      
            ArrayFloat.D2 dataLatBound = new ArrayFloat.D2(GEO_GRID_NUM_LAT, 2);
            ArrayFloat.D2 dataLonBound = new ArrayFloat.D2(GEO_GRID_NUM_LON, 2);      
            // write latitude   
            float deltaGrid = (float) (GeoGrid2D.GEO_GRID_STEP_LATITUDE/2.0);
            float[] plat = GeoGrid2D.getGridLatitudeFloat();
            for (i=0; i<GEO_GRID_NUM_LAT; i++) {
                dataLat.set(i, plat[i]);
                dataLatBound.set(i, 0, plat[i] - deltaGrid);
                dataLatBound.set(i, 1, plat[i] + deltaGrid);
            }
            this.ncWriter.write(vlat, dataLat);
            this.ncWriter.write(vlatBound, dataLatBound);
            
            // write longitude
            deltaGrid = (float) (GeoGrid2D.GEO_GRID_STEP_LONGITUDE/2.0);
            float[] plon = GeoGrid2D.getGridLongitudeFloat();
            for(i=0; i<GEO_GRID_NUM_LON; i++){
                dataLon.set(i, plon[i]);
                dataLonBound.set(i,0, plon[i]-deltaGrid);
                dataLonBound.set(i,1, plon[i]+deltaGrid);
            }
            this.ncWriter.write(vlon, dataLon);
            this.ncWriter.write(vlonBound, dataLonBound);

            // write time
            ArrayDouble.D1 dataTime = new ArrayDouble.D1(this.numberMonths);
            ArrayDouble.D2 dataTimeBound = new ArrayDouble.D2(this.numberMonths, 2);
            double t0 = TimeUtility.getDateNumber(1978,1, 1);
            double[] t = new double[this.numberMonths];
            double[][] tb = new double[this.numberMonths][2];
            int[] tyear = this.geoData.getFyear();
            int[] tmonth = this.geoData.getFmonth();
            for(i=0; i<this.numberMonths; i++) {
                kyear = tyear[i];
                kmonth = tmonth[i];
                tm0 = TimeUtility.getDateNumber(kyear, kmonth, 16) - t0;
                tm1 = TimeUtility.getDateNumber(kyear, kmonth, 1) - t0;
                if (kmonth<12) {
                    tm2 = TimeUtility.getDateNumber(kyear, kmonth+1, 1) - t0 -1;
                } else {
                    tm2 = TimeUtility.getDateNumber(kyear, 12, 31) - t0;                    
                }
                dataTime.set(i, tm0);
                dataTimeBound.set(i, 0, tm1);
                dataTimeBound.set(i, 1, tm2);
            }
            this.ncWriter.write(vtime, dataTime);
            this.ncWriter.write(vtimeBound, dataTimeBound);
            
            // write BT
            double[][][] T = this.geoData.getData();
            ArrayDouble.D3 vdata = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<this.numberMonths; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, T[i][j][k]);
                    }
                }
            }
            this.ncWriter.write(varBt, vdata);

            // write anomaly and climatology
            int cyear1 = MltCdrConstant.MLT_CDR_CLIMATOLOGY_YEAR_START;
            int cmonth1 = 1;
            int cyear2 = MltCdrConstant.MLT_CDR_CLIMATOLOGY_YEAR_END;
            int cmonth2 =12;
            
            double[][][] vanom = this.geoData.getAnomaly(cyear1, cmonth1, cyear2, cmonth2);
            double[][][] vclim = this.geoData.getClimatology(cyear1, cmonth1, cyear2, cmonth2);
            
            vdata = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<this.numberMonths; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, vanom[i][j][k]);
                    }
                }
            }
            this.ncWriter.write(varAnom, vdata);
            
            vdata = new ArrayDouble.D3(12, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<12; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, vclim[i][j][k]);
                    }
                }
            }
            this.ncWriter.write(varClim, vdata);            
            infoLogger.info("generated CDR data file " + this.fileFullpath);
        }catch (Exception e) {
            e.printStackTrace();
        }
        
        if (this.ncWriter != null) {
            try {
                    this.ncWriter.close();
                } catch (Exception exp) {
                    errLogger.error("Failed to close netcdf file" + this.fileFullpath);
                    exp.printStackTrace();
                }
        }
        
    }
    
    
    /**
     * Define NetCDF Variables
     */
    public void defineGlobalAttributes(){
        ncWriter.addGroupAttribute(null, new Attribute("Conventions", "CF-1.6, ACDD-1.3"));
            switch(this.cdrVarId){
                case 0:
                    ncWriter.addGroupAttribute(null, new Attribute("title", "NOAA Climate Data Record (CDR) for Mean Layer Temperature, Version 5: Lower-Troposphere (TLT)"));
                    break;
                case 1:
                    ncWriter.addGroupAttribute(null, new Attribute("title", "NOAA Climate Data Record (CDR) for Mean Layer Temperature, Version 5: Mid-Troposphere (TMT)"));
                    break;
                case 2:
                    ncWriter.addGroupAttribute(null, new Attribute("title", "NOAA Climate Data Record (CDR) for Mean Layer Temperature, Version 5: Upper-Troposphere (TUT)"));
                    break;
                case 3:
                    ncWriter.addGroupAttribute(null, new Attribute("title", "NOAA Climate Data Record (CDR) for Mean Layer Temperature, Version 5: Lower-Stratospher (TLS)"));
                    break;                                
            }

        ncWriter.addGroupAttribute(null, new Attribute("source", "NOAA Fundamental Climate Data Record (FCDR) of MSU and AMSU-A Level 1c Brightness Temperature, Version 1.0; and NOAA ATMS Reprocessed SDR Products (RPATMSTDR), Version 2.0.  http://doi.org/10.7289/V51Z429F, http://doi:10.7289/V5X63JT2, and https://www.avl.class.noaa.gov/saa/products/search?datatype_family=RPATMSTDR"));
        ncWriter.addGroupAttribute(null, new Attribute("references",  "Zou, C.-Z., H. Xu, X. Hao, and Q. Liu, 2023: Mid-Tropospheric Layer Temperature Record Derived from Satellite Microwave Sounder Observations with Backward Merging Approach, J. Geophys. Res.-Atmosphere. Published online: 03 March 2023. DOI: 10.1029/2022JD037472"));
        ncWriter.addGroupAttribute(null, new Attribute("id", this.fileShortname));        
        ncWriter.addGroupAttribute(null, new Attribute("naming_authority", "gov.noaa.ncei"));
        // date and time
        Date date = new Date();
        String strDateFormat = "yyyy-MM-dd hh:mm:ss";
        DateFormat dateFormat = new SimpleDateFormat(strDateFormat);
        dateFormat.setTimeZone(TimeZone.getTimeZone( "UTC" ) );
        String formattedDate= String.format("%s Z", dateFormat.format(date));
        ncWriter.addGroupAttribute(null, new Attribute("date_created", formattedDate));       
        ncWriter.addGroupAttribute(null, new Attribute("license", "These data were produced by NOAA and are not subject to copyright protection in the United States. NOAA waives any potential copyright and related rights in these data worldwide through the Creative Commons Zero 1.0 Universal Public Domain Dedication (CC0-1.0)."));
        ncWriter.addGroupAttribute(null, new Attribute("summary", "The Mean Layer Temperature – NOAA Climate Data Record (CDR) provides long-term temperature measurements for a thick layer of the upper atmosphere. "
                + "Input data comes from Mean Layer Temperature (MLT) level-1c swath radiances taken by the Microwave Sounding Unit (MSU),"
                + "and the Advanced Microwave Sounder Unit-A (AMSU-A) and the Advanced Technology Microwave Sounder "
                + "(ATMS) aboard NOAA, NASA, and EUMETSAT satellites. The CDR was calibrated using the NOAA STAR "  
                + "Integrated Microwave Inter-Calibration Approach (IMICA) to remove calibration drifting errors. The final " 
                + "dataset incorporates inter-satellite bias corrections to ensure consistency, and adjustments that account " 
                + "for diurnal drift effect, differences between viewing angles, and channel frequency differences between "  
                + "sensors. The final record is a monthly global gridded dataset with 2.5°x2.5° Latitude/Longitude" 
                + "resolution, and extends from 1978–present."));

        if ( this.cdrVarId>2 ) {
            ncWriter.addGroupAttribute(null, new Attribute("keywords", "Atmosphere > Atmospheric Temperature > Air Temperature, Vertical Location > Stratosphere"));
        } else {
            ncWriter.addGroupAttribute(null, new Attribute("keywords", "Atmosphere > Atmospheric Temperature > Air Temperature, Vertical Location > Troposphere"));            
        }        
        ncWriter.addGroupAttribute(null, new Attribute("keywords_vocabulary", "NASA Global Change Master Directory (GCMD) Earth Science Keywords, Version 8.0"));
        
        ncWriter.addGroupAttribute(null, new Attribute("institution", "NOAA/NESDIS/STAR>Center for Satellite Applications and Research, NESDIS, NOAA, U.S. Department of Commerce"));
        String tmstr1 = String.format("%%4d-%02d-01 00:00:00Z", this.year1, this.month1);
        ncWriter.addGroupAttribute(null, new Attribute("time_coverage_start", tmstr1));
        int[] tv = TimeUtility.getPreviousDay(this.year2, this.month2+1,1);
        String tmstr2 = String.format("%%4d-%02d-%02d 23:59:59Z", tv[0], tv[1], tv[2]);
        ncWriter.addGroupAttribute(null, new Attribute("time_coverage_end", tmstr2));
        String tmStr3 = String.format("P%3dM", this.numberMonths);
        ncWriter.addGroupAttribute(null, new Attribute("time_coverage_duration", tmstr2));
        ncWriter.addGroupAttribute(null, new Attribute("time_coverage_resolution", "P1M"));
        ncWriter.addGroupAttribute(null, new Attribute("contributor_name", "Cheng-Zhi Zou, Xianjun Hao, Hui Xu, Qian Liu"));
        ncWriter.addGroupAttribute(null, new Attribute("contributor_role", "Principal Investigator and originator of NOAA/STAR MSU mean layer temperature product, software and algorithm development, algorithm development and testing."));
        ncWriter.addGroupAttribute(null, new Attribute("acknowledgment", "This project was supported by a grant from the NOAA Climate Data Record (CDR) Program for Satellites"));
        ncWriter.addGroupAttribute(null, new Attribute("cdr_program", "NOAA Climate Data Record Program for Satellites"));        
        ncWriter.addGroupAttribute(null, new Attribute("cdr_variable", this.cdrVariableBt));
        ncWriter.addGroupAttribute(null, new Attribute("metadata_link", "gov.noaa.ncei:C00935"));
        ncWriter.addGroupAttribute(null, new Attribute("uncertainty_estimate", "0.04 K/decade"));
        ncWriter.addGroupAttribute(null, new Attribute("cdr_data_type", "Grid"));
        ncWriter.addGroupAttribute(null, new Attribute("creator_email", "Cheng-Zhi.Zou@noaa.gov"));
        ncWriter.addGroupAttribute(null, new Attribute("creator_name", "Cheng-Zhi Zou"));        
        ncWriter.addGroupAttribute(null, new Attribute("geospatial_lat_max", 90f));
        ncWriter.addGroupAttribute(null, new Attribute("geospatial_lat_min", -90f));
        ncWriter.addGroupAttribute(null, new Attribute("geospatial_lon_max", 180f));
        ncWriter.addGroupAttribute(null, new Attribute("geospatial_lon_min", -180f));
        ncWriter.addGroupAttribute(null, new Attribute("processing_level", "NOAA Level 3"));
        ncWriter.addGroupAttribute(null, new Attribute("product_version", "5.0"));
        ncWriter.addGroupAttribute(null, new Attribute("platform", "TIROS-N,NOAA-06,NOAA-07,NOAA-08,NOAA-09,NOAA-10,NOAA-11,NOAA-12,NOAA-14,NOAA-15,NOAA-18,NOAA-19,NOAA-20, MetOp-A, AQUA, SNPP"));
        ncWriter.addGroupAttribute(null, new Attribute("sensor", "MSU>Microwave Sounding Unit; AMSU-A->Advanced Microwave Sounding Unit-A, ATMS->Advanced Technology Microwave Sounder"));
        ncWriter.addGroupAttribute(null, new Attribute("spatial_resolution", "2.5 latitude by 2.5 longitude"));
        String timeStamp = new SimpleDateFormat("dd-MMM-yyyy HH:mm:ss").format(new java.util.Date());
        ncWriter.addGroupAttribute(null, new Attribute("creation_date", timeStamp));        
    }
    
}
