/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.cdr.MltTLS.java
*
* NAME
*   atms.cdr.MltTLS.java
*
* PURPOSE
*   To generate the final MLT TLS product
* 
* DESCRIPTION
*	This class is the class for MLT TLS generation
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package cdr;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import ucar.nc2.Variable;
import ucar.ma2.*;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriter;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import geogrid.Geo2dTimeSeries;
import geogrid.GeoGrid2D;
import geogrid.GeoGridConstant;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;
import geogrid.GeoTimeseriesMerger;
import ucar.nc2.NetcdfFile;

/**
 * Class to generate TLS data.
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class MltTLS {
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");

    /**
     * Start year of TLS
     */
    final private int yearStart = 1978;  
    
    /**
     * Start month of TLS
     */
    final private int monthStart = 12;
    
    /**
     * CDR name
     */
    final private String cdrVarname = "TLS";

    /**
     * Year of the end time
     */
    private int yearEnd;

    /**
     * Month of the end time
     */
    private int monthEnd;
    
    /**
     * Number of months
     */
    private int numberMonths;
    
    /**
     * CDR file name
     */
    private String filenameCDR;
    
    /**
     * Full path of the CDR file
     */
    private String fileFullPathCDR = null;
    
    /**
     * CDR data directory
     */
    private String dirMltCDR;
    
    /**
     * Directory of reference CDR
     */
    private String dirRefCDR;
    
    /**
     * Directory of L4 data
     */
    private String dirL4;  
    
    /**
     * Root directory of LUTs
     */
    private String dirLut;    
    
    /**
     * List of years for data
     */
    private int[] fyear;

    /**
     * List of months for data
     */
    private int[] fmonth;
    
    /**
     *  Time for data
     */
    private double[] ft;

    /**
     * Anomaly data
     */
    private double[][][] Tanom;
    
    /**
     * climatology data
     */
    private double[][][] Tclim;
    
    /**
     * Brightness temperature data
     */
    private double[][][] Tb;
    
    /**
     *  Merged time series
     */
    private Geo2dTimeSeries g2tCdr;
    
    /**
     * Constructor with task configuration information
     * @param taskConf, task configuration 
     */
    public MltTLS(MltCdrTaskConfig taskConf) {
        this.yearEnd = taskConf.yearEnd;
        this.monthEnd = taskConf.monthEnd;
        this.numberMonths = (this.yearEnd - this.yearStart)*12 + (this.monthEnd - this.monthStart +1);
        
        this.initTime();
        
        this.Tanom = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tb = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tclim = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        this.dirL4 = taskConf.dirL4;
        if (!this.dirL4.endsWith(File.separator)) {
            this.dirL4=this.dirL4 + File.separator;
        }
        
        this.dirLut = taskConf.dirLut;
        if (!this.dirLut.endsWith(File.separator)) {
            this.dirLut=this.dirLut + File.separator;
        }
        
        this.dirRefCDR = taskConf.dirRefCDR;
        if (!this.dirRefCDR.endsWith(File.separator)) {
            this.dirRefCDR=this.dirRefCDR + File.separator;
        }               

        this.dirMltCDR = taskConf.dirMltCDR;
        if (!this.dirMltCDR.endsWith(File.separator)) {
            this.dirMltCDR=this.dirMltCDR + File.separator;
        }                       
    }    

    /**
     * Initialize time 
     */
    private void initTime(){
        int i, j , k;
        int  km, kyear, kmonth;
        int year1 = this.yearStart;
        int month1 = this.monthStart  ;
        int year2 = this.yearEnd;
        int month2 = this.monthEnd;
        this.numberMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numberMonths];
        this.fmonth = new int[numberMonths];
        this.ft = new double[numberMonths];
                
        for(i=0; i<numberMonths; i++){
            km = i + month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
    }
    
    /**
     * Get year list for data time series
     * @return year list for data time series
     */
    public int[] getFyear() {
        return this.fyear;
    }

    /**
     * Get month list for data time series
     * @return month list for data time series
     */
    public int[] getFmonth(){
        return this.fmonth;
    }
    
    /**
     * Get CDR temperature data
     * @return  CDR temperature data
     */
    public double[][][] getTbData(){
        return this.Tb;
    }
        
    /** 
     * Merge MSU/AMSU-A/ATMS data to get MLT CDR
     */
    public void mergeCDR() {
        double vbias, vtmp;
        int[] tyear;
        int[] tmonth;
        double vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE;
        double vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE;
        
        //create a GeoTimeseriesMerger to process data merging
        GeoTimeseriesMerger merger = new GeoTimeseriesMerger(this.yearStart, this.monthStart, this.yearEnd, this.monthEnd, vmin, vmax );     
        String fileMsuAmsuaL4 = this.dirL4 + "TLS_MSU_AMSUA_L4.nc";
        // RTLS starts from 2002/09
        String fileRefCDR = this.dirRefCDR + MltCdrConstant.getRefCDRFileShortname( this.cdrVarname, 2002,9, this.yearEnd, this.monthEnd);
        // add old data (MSU & AMSU-A)
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayInt.D1 iarr1 = null;
        ArrayDouble.D4 darr4 = null;
        int[] shape = new int[4];
        int[] origin = new int[4];
        int[] shape1 = new int[1];
        int[] origin1 = new int[1];
        int[] pyear = new int[500]; 
        int[] pmonth = new int[500];
        int i, j, k, m, mt, ms;        
        
        origin1[0] = 0;
        for (i = 0; i < 4; i++) {
            origin[i] = 0;
        }
        
        try{
            fileNc = NetcdfFile.open(fileMsuAmsuaL4);
            // read year
            dataVar = fileNc.findVariable("fyear");
            if (dataVar == null) {
                errLogger.error("failed to find variable fyear from file " + fileMsuAmsuaL4);
            } else {
                shape1 = dataVar.getShape();
                iarr1 = (ArrayInt.D1) dataVar.read(origin1, shape1);
                mt = shape1[0];
                pyear = new int[mt];
                for(i=0; i<mt;i++){
                    pyear[i] = iarr1.get(i);
                }
            }
            // read month
            dataVar = fileNc.findVariable("fmonth");
            if (dataVar == null) {
                errLogger.error("failed to find variable fmonth from file " + fileMsuAmsuaL4);
            } else {
                shape1 = dataVar.getShape();
                mt = shape1[0];
                pmonth = new int[mt];
                iarr1 = (ArrayInt.D1) dataVar.read(origin1, shape1);
                for(i=0; i<mt;i++){
                    pmonth[i] = iarr1.get(i);
                }
            }
            
            // read all data
            String varName = "TLS";
            dataVar = fileNc.findVariable(varName);
            if (dataVar == null) {
                //System.out.println("failed to find variable from file " + fileMsuAmsuaL4);
                errLogger.error("failed to find variable from file " + fileMsuAmsuaL4);
            } else {
                shape = dataVar.getShape();
                darr4 = (ArrayDouble.D4) dataVar.read(origin, shape);  
                ms = shape[0];
                mt = shape[1];
                for (m=0; m<ms; m++) {
                    double[][][] T1 = new double[mt][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
                    for(k=0; k<mt; k++) {
                        for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                            for(j=0; j<GEO_GRID_NUM_LON; j++) {
                               T1[k][i][j] = darr4.get(m,k, i, j);
                               if ( (T1[k][i][j]<vmin) || (T1[k][i][j]>vmax) ){
                                   T1[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE; 
                               } 
                            }
                        }
                    }
                    
                    merger.addData(pyear, pmonth, T1);                    
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            errLogger.error("Failed to read MSU/AMSU-A L4 data file " + fileMsuAmsuaL4);
        }

        // add new reference data 
        Geo2dTimeSeries g2tRef = new Geo2dTimeSeries(fileRefCDR, "RTLS");        
        tyear = g2tRef.getFyear();
        tmonth = g2tRef.getFmonth();
        double[][][] T2 = g2tRef.getData();
        merger.addData(tyear, tmonth, T2);
        
        //for testing only
        //this.saveMergedData(merger);
        
        this.Tb = merger.mergeData();
        
        this.g2tCdr = new Geo2dTimeSeries(this.fyear, this.fmonth, this.Tb);
        this.g2tCdr.setDeaultTemperatureRange();
//        System.out.println("DirCDR = "+ this.dirMltCDR);
        MltFileNetCDF netcdfWriter = new MltFileNetCDF("TLS", this.dirMltCDR, this.g2tCdr);
        netcdfWriter.writeDataToNetcdf();
        this.fileFullPathCDR = netcdfWriter.getCdrFilename();
    }
    
    /**
     * Get full path of the CDR file
     * @return full path of the CDR file
     */
    public String getCdrFileFullPath(){
        return this.fileFullPathCDR;
    }

    /**
     * show global mean for testing
     */
    private void showGlobalMean(){
        int cyear1 = MltCdrConstant.MLT_CDR_CLIMATOLOGY_YEAR_START;
        int cmonth1 = 1;
        int cyear2 = MltCdrConstant.MLT_CDR_CLIMATOLOGY_YEAR_END;
        int cmonth2 = 12;
        String msg;
        
        double[][][] anom = this.g2tCdr.getAnomaly(cyear1, cmonth1, cyear2, cmonth2);
        Geo2dTimeSeries g2t = new Geo2dTimeSeries(this.fyear, this.fmonth, anom);
        g2t.setDeaultTemperatureAnomalyRange();
        g2t.setDefaultValidLatIndex();
        double[] ym = g2t.getGlobalMean();
        
        for (int k=0; k<this.numberMonths; k++) {
            msg = String.format("%4d, %2d: %12.5f", this.fyear[k], this.fmonth[k], ym[k]);
            System.out.println(msg);
        }
    }


    /**
     * For testing only, save merged data to file before calculating mean
     */
    private void saveMergedData(GeoTimeseriesMerger merger){
        String fileTest = this.dirMltCDR + "Testing_tls_merged_.nc";
        NetcdfFileWriter ncFileOut = null;
        Variable dataVar = null;
        ArrayInt.D1 iarr1 = null;
        ArrayDouble.D3 darr3 = null;
        ArrayDouble.D1 darr1 = null;
        try{
            ncFileOut = NetcdfFileWriter.createNew(NetcdfFileWriter.Version.netcdf3, fileTest);
            
            // create Dimensions
            Dimension dimTime = ncFileOut.addDimension(null, "time", this.numberMonths);
            Dimension dimLon = ncFileOut.addDimension(null, "longitude",GEO_GRID_NUM_LON);
            Dimension dimLat = ncFileOut.addDimension(null, "latitude", GEO_GRID_NUM_LAT);
            Dimension dimMonth = ncFileOut.addDimension(null, "month",12);

            List<Dimension> dimTimeD = new ArrayList<Dimension>();
            dimTimeD.add(dimTime);

            List<Dimension> dimLatD = new ArrayList<Dimension>();
            dimLatD.add(dimLat);

            List<Dimension> dimLonD = new ArrayList<Dimension>();
            dimLonD.add(dimLon);
            
            List<Dimension> dimMonD = new ArrayList<Dimension>();
            dimMonD.add(dimMonth);
            
            List<Dimension> dimGridD = new ArrayList<Dimension>();
            dimGridD.add(dimTime);
            dimGridD.add(dimLat);
            dimGridD.add(dimLon);

            List<Dimension> dimGridClim = new ArrayList<Dimension>();
            dimGridClim.add(dimMonth);
            dimGridClim.add(dimLat);
            dimGridClim.add(dimLon);
            // define corridinates variable variables
            Variable vlon = ncFileOut.addVariable(null, "longitude", DataType.FLOAT, dimLonD);
            Variable vlat = ncFileOut.addVariable(null, "latitude", DataType.FLOAT, dimLatD);
            Variable vmon = ncFileOut.addVariable(null, "month", DataType.INT, dimTimeD);
            Variable vyear = ncFileOut.addVariable(null, "fyear", DataType.INT, dimTimeD);
            Variable vmonth = ncFileOut.addVariable(null, "fmonth", DataType.INT, dimTimeD);
            Variable vft = ncFileOut.addVariable(null, "ft", DataType.DOUBLE, dimTimeD);
            vlat.addAttribute(new Attribute("units", "degree"));
            vlon.addAttribute(new Attribute("units", "degree"));
            
            // define data variables and attributes
            String varName = "bt_sum";
            Variable varBt= ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridD);
            varBt.addAttribute(new Attribute("coordinates","time latitude longitude" ));
            // define data variables and attributes
            varName = "pct";
            Variable varPix= ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridD);
            varBt.addAttribute(new Attribute("coordinates","time latitude longitude" ));
            ncFileOut.create();
            // write data
            ArrayFloat.D1 dataLat = new ArrayFloat.D1(GEO_GRID_NUM_LAT);
            ArrayFloat.D1 dataLon = new ArrayFloat.D1(GEO_GRID_NUM_LON);      
            ArrayInt.D1 dataCmonth = new ArrayInt.D1(12);
            ArrayInt.D1 dataFyear = new ArrayInt.D1(this.numberMonths);
            ArrayInt.D1 dataFmonth = new ArrayInt.D1(this.numberMonths);
            ArrayDouble.D1 dataFt = new ArrayDouble.D1(this.numberMonths);
            // write latitude   
            float[] plat = GeoGrid2D.getGridLatitudeFloat();
            for (int i=0; i<GEO_GRID_NUM_LAT; i++) {
                dataLat.set(i, plat[i]);
            }
            ncFileOut.write(vlat, dataLat);
            
            // write longitude
            float[] plon = GeoGrid2D.getGridLongitudeFloat();
            for(int i=0; i<GEO_GRID_NUM_LON; i++){
                dataLon.set(i, plon[i]);
            }
            ncFileOut.write(vlon, dataLon);
            
            // write year 
            for(int i=0; i<this.numberMonths; i++) {
                dataFyear.set(i, this.fyear[i]);
            }
            ncFileOut.write(vyear, dataFyear);
            
            // write fmonth
            for(int i=0; i<this.numberMonths; i++) {
                dataFmonth.set(i, fmonth[i]);
            }
            ncFileOut.write(vmonth, dataFmonth);

            
            // write time
            for(int i=0; i<this.numberMonths; i++) {
                dataFt.set(i, ft[i]);
            }
            ncFileOut.write(vft, dataFt);
            
            // write month
            for(int i=0; i<12; i++) {
                dataCmonth.set(i, i+1);
            }
            ncFileOut.write(vmon, dataCmonth);

            // write vsum
            double[][][] vsumMerg = merger.getSum();
            ArrayDouble.D3 vdata = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (int i=0; i<this.numberMonths; i++) {
                for(int j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(int k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k,vsumMerg[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varBt, vdata);

            // write pixel number
            int[][][] pctMerg = merger.getPixelCount();
            vdata = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (int i=0; i<this.numberMonths; i++) {
                for(int j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(int k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, pctMerg[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varPix, vdata);      
        } catch(Exception exp) {
            exp.printStackTrace();
        }

        if (ncFileOut != null) {
            try {
                    ncFileOut.close();
                } catch (Exception exp) {
                    errLogger.error("Failed to close netcdf file for testing");
                    exp.printStackTrace();
                }
        }
        
    }
    
}
