/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.cdr.MltTLT.java
*
* NAME
*   atms.cdr.MltTLT.java
*
* PURPOSE
*   To generate the final MLT TLT product
* 
* DESCRIPTION
*	This class is the class for MLT TLT generation
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package cdr;

import java.io.File;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import geogrid.Geo2dTimeSeries;
import geogrid.GeoGrid2D;
import geogrid.GeoGridConstant;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;
import geogrid.GeoTimeseriesMerger;

/**
 * Class for TLT CDR
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class MltTLT {
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    
    /**
     * TMT coefficient for generating TLT
     */
    final static private double tmtCoefForTLT=1.430d;

    /**
     * TUT coefficient for generating TLT
     */
    final static private double tutCoefForTLT=-0.462;

    /**
     * TLS coefficient for generating TLT
     */
    final static private double tlsCoefForTLT=0.032d;

    /**
     * Start year for TLT
     */
    final private int yearStart = 1981;  
    
    /** 
     * Start month for TLT
     */
    final private int monthStart = 1;
    
    /**
     * CDR Variable name
     */
    final private String cdrVarname = "TLT";

    /**
     * End year for TLT
     */    
    private int yearEnd;

    /**
     * End month for TLT
     */    
    private int monthEnd;

    /**
     * Number of months
     */
    private int numberMonths;
    
    /**
     * CDR file name
     */
    private String filenameCDR;
    
    /**
     * Full path of the CDR file
     */
    private String fileFullPathCDR = null;
    
    /**
     * CDR data directory
     */
    private String dirMltCDR;
    
    /**
     * Full path of the TMT file
     */
    private String fileTMT;

    /**
     * Full path of the TUT file
     */
    private String fileTUT;

    /**
     * Full path of the TLS file
     */
    private String fileTLS;


    /**
     * TMT object 
     */
    private MltTMT objMltTmt;

    /**
     * TUT object 
     */
    private MltTUT objMltTut;

    /**
     * TLS object 
     */
    private MltTLS objMltTls;
    
    /**
     * List of years for data
     */
    private int[] fyear;

    /**
     * List of months for data
     */
    private int[] fmonth;
    
    /**
     *  Time for data
     */
    private double[] ft;
    
    /**
     * Anomaly data
     */
    private double[][][] Tanom;
    
    /**
     * climatology data
     */
    private double[][][] Tclim;
    
    /**
     * Brightness temperature data
     */
    private double[][][] Tb;

    /**
     *  Merged time series
     */
    private Geo2dTimeSeries g2tCdr;
    
    
    /**
     * Construct TLT with TMT, TUT and TLS
     * @param taskConf, task configuration
     * @param objMltTmt, MltTMT object
     * @param objMltTut, MltTMT object
     * @param objMltTls, MltTMT object 
     */
    public MltTLT (MltCdrTaskConfig taskConf, MltTMT objMltTmt, MltTUT objMltTut, MltTLS objMltTls){
        this.yearEnd = taskConf.yearEnd;
        this.monthEnd = taskConf.monthEnd;
        this.dirMltCDR = taskConf.dirMltCDR;
        this.objMltTmt = objMltTmt;
        this.objMltTut = objMltTut;
        this.objMltTls = objMltTls;
        this.numberMonths = (this.yearEnd - this.yearStart)*12 + (this.monthEnd - this.monthStart +1);
        
        this.initTime();
        
        this.Tanom = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tb = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tclim = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];        
    }
    
        
    /**
     * Initialize time 
     */
    private void initTime(){
        int i, j , k;
        int  km, kyear, kmonth;
        int year1 = this.yearStart;
        int month1 = this.monthStart  ;
        int year2 = this.yearEnd;
        int month2 = this.monthEnd;
        this.numberMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numberMonths];
        this.fmonth = new int[numberMonths];
        this.ft = new double[numberMonths];
                
        for(i=0; i<numberMonths; i++){
            km = i + month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
    }
    
    
    /**
     * Generate TLT data using TMT, TUT and TLS
     */
    public void mergeCDR(){
        int i, j, k;
        int kyear, kmonth;
        int year1Tmt, month1Tmt, year1Tut, month1Tut, year1Tls, month1Tls;
        int indexTmt, indexTut, indexTls;
        double vTmt, vTut, vTls;
        double vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE;
        double vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE;
        
        int[] fyearTmt = objMltTmt.getFyear();
        int[] fmonthTmt = objMltTmt.getFmonth();
        double[][][] BtTmt = objMltTmt.getTbData();
        year1Tmt = fyearTmt[0];
        month1Tmt = fmonthTmt[0];
        
        int[] fyearTut = objMltTut.getFyear();
        int[] fmonthTut = objMltTut.getFmonth();
        double[][][] BtTut = objMltTut.getTbData();
        year1Tut = fyearTut[0];
        month1Tut = fmonthTut[0];

        int[] fyearTls = objMltTls.getFyear();
        int[] fmonthTls = objMltTls.getFmonth();
        double[][][] BtTls = objMltTls.getTbData();
        year1Tls = fyearTls[0];
        month1Tls = fmonthTls[0];
        
        for(k=0; k<this.numberMonths; k++) {
            kyear = this.fyear[k];
            kmonth = this.fmonth[k];
            indexTmt = (kyear - year1Tmt)*12 + (kmonth - month1Tmt);
            indexTut = (kyear - year1Tut)*12 + (kmonth - month1Tut);
            indexTls = (kyear - year1Tls)*12 + (kmonth - month1Tls);
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vTmt = BtTmt[indexTmt][i][j];
                    vTut = BtTut[indexTut][i][j];
                    vTls = BtTls[indexTls][i][j];
                    if( (vTmt>=vmin) &(vTmt<=vmax) & (vTut>=vmin) & (vTut<=vmax) & (vTls>=vmin) &(vTls<=vmax) ) {
                        this.Tb[k][i][j]= tmtCoefForTLT*vTmt + tutCoefForTLT*vTut + tlsCoefForTLT*vTls;
                    } else {
                        this.Tb[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }
        
        this.g2tCdr = new Geo2dTimeSeries(this.fyear, this.fmonth, this.Tb);
        this.g2tCdr.setDeaultTemperatureRange();
        MltFileNetCDF netcdfWriter = new MltFileNetCDF("TLT", this.dirMltCDR, this.g2tCdr);
        netcdfWriter.writeDataToNetcdf();
        this.fileFullPathCDR = netcdfWriter.getCdrFilename();        
        
    }
    
    /**
     * Get full path of the CDR file
     * @return full path of the CDR file
     */
    public String getCdrFileFullPath(){
        return this.fileFullPathCDR;
    }

   /**
     * show global mean for testing
     */
    private void showGlobalMean(){
        int cyear1 = MltCdrConstant.MLT_CDR_CLIMATOLOGY_YEAR_START;
        int cmonth1 = 1;
        int cyear2 = MltCdrConstant.MLT_CDR_CLIMATOLOGY_YEAR_END;
        int cmonth2 = 12;
        String msg;
        
        double[][][] anom = this.g2tCdr.getAnomaly(cyear1, cmonth1, cyear2, cmonth2);
        Geo2dTimeSeries g2t = new Geo2dTimeSeries(this.fyear, this.fmonth, anom);
        g2t.setDeaultTemperatureAnomalyRange();
        g2t.setDefaultValidLatIndex();
        double[] ym = g2t.getGlobalMean();
        
        for (int k=0; k<this.numberMonths; k++) {
            msg = String.format("%4d, %2d: %12.5f", this.fyear[k], this.fmonth[k], ym[k]);
            System.out.println(msg);
        }
    }
        
}
