/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.cdr.MltTUT.java
*
* NAME
*   atms.cdr.MltTUT.java
*
* PURPOSE
*   To generate the final MLT TUT product
* 
* DESCRIPTION
*	This class is the class for MLT TUT generation
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package cdr;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import ucar.nc2.Variable;
import ucar.ma2.*;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriter;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import geogrid.Geo2dTimeSeries;
import geogrid.GeoGrid2D;
import geogrid.GeoGridConstant;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;
import geogrid.GeoTimeseriesMerger;

/**
 * Class for MLT TUT CDE.
 * @author Xianjun Hao <xhao1@gmu.edu>
 */
public class MltTUT {

    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    
    /**
     * Start year of TUT
     */
    final private int yearStart = 1981;  
    
    /**
     * Start month of TUT
     */
    final private int monthStart = 1;
    
    /** 
     * Variable name
     */
    final private String cdrVarname = "TUT";
    
    /**
     * End year for TUT
     */    
    private int yearEnd;

    /**
     * End month for TUT
     */    
    private int monthEnd;

    /**
     * Number of months
     */
    private int numberMonths;
    
    /**
     * CDR file name
     */
    private String filenameCDR;
    
    /**
     * Full path of the CDR file
     */
    private String fileFullPathCDR = null;
    
    /**
     * CDR data directory
     */
    private String dirMltCDR;
    
    /**
     * Directory of reference CDR
     */
    private String dirRefCDR;
    
    /**
     * Directory of L4 data
     */
    private String dirL4;  
    
    /**
     * Root directory of LUTs
     */
    private String dirLut;  
    
    /**
     * List of years for data
     */
    private int[] fyear;

    /**
     * List of months for data
     */
    private int[] fmonth;
    
    /**
     *  Time for data
     */
    private double[] ft;
    
    /**
     * Anomaly data
     */
    private double[][][] Tanom;
    
    /**
     * climatology data
     */
    private double[][][] Tclim;
    
    /**
     * Brightness temperature data
     */
    private double[][][] Tb;
    
    /**
     *  Merged time series
     */
    private Geo2dTimeSeries g2tCdr;
    
    /**
     * Constructor with task configuration information
     * @param taskConf, task configuration 
     */
    public MltTUT(MltCdrTaskConfig taskConf) {
        this.yearEnd = taskConf.yearEnd;
        this.monthEnd = taskConf.monthEnd;
        this.numberMonths = (this.yearEnd - this.yearStart)*12 + (this.monthEnd - this.monthStart +1);
        
        this.initTime();
        
        this.Tanom = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tb = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tclim = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        this.dirL4 = taskConf.dirL4;
        if (!this.dirL4.endsWith(File.separator)) {
            this.dirL4=this.dirL4 + File.separator;
        }
        
        this.dirLut = taskConf.dirLut;
        if (!this.dirLut.endsWith(File.separator)) {
            this.dirLut=this.dirLut + File.separator;
        }
        
        this.dirRefCDR = taskConf.dirRefCDR;
        if (!this.dirRefCDR.endsWith(File.separator)) {
            this.dirRefCDR=this.dirRefCDR + File.separator;
        }               

        this.dirMltCDR = taskConf.dirMltCDR;
        if (!this.dirMltCDR.endsWith(File.separator)) {
            this.dirMltCDR=this.dirMltCDR + File.separator;
        }                       
    }    

    /**
     * Initialize time 
     */
    private void initTime(){
        int i, j , k;
        int  km, kyear, kmonth;
        int year1 = this.yearStart;
        int month1 = this.monthStart  ;
        int year2 = this.yearEnd;
        int month2 = this.monthEnd;
        this.numberMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numberMonths];
        this.fmonth = new int[numberMonths];
        this.ft = new double[numberMonths];
                
        for(i=0; i<numberMonths; i++){
            km = i + month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
    }
    
    /**
     * Get year list for data time series
     * @return year list for data time series
     */
    public int[] getFyear() {
        return this.fyear;
    }

    /**
     * Get month list for data time series
     * @return month list for data time series
     */
    public int[] getFmonth(){
        return this.fmonth;
    }
    
    /**
     * Get CDR temperature data
     * @return  CDR temperature data
     */
    public double[][][] getTbData(){
        return this.Tb;
    }
        
    /** 
     * Merge MSU/AMSU-A/ATMS data to get MLT CDR
     */
    public void mergeCDR() {
        double vbias, vtmp;
        double vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE;
        double vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE;
        
        //create a GeoTimeseriesMerger to process data merging
        GeoTimeseriesMerger merger = new GeoTimeseriesMerger(this.yearStart, this.monthStart, this.yearEnd, this.monthEnd, vmin, vmax );     
        String fileMsuAmsuaL4 = this.dirL4 + "TUT_MSU_AMSUA_L4.nc";
        String fileRefCDR = this.dirRefCDR + MltCdrConstant.getRefCDRFileShortname( this.cdrVarname, 2002,8, this.yearEnd, this.monthEnd);
        
        Geo2dTimeSeries g2tOld = new Geo2dTimeSeries(fileMsuAmsuaL4, "TUT");
        Geo2dTimeSeries g2tRef = new Geo2dTimeSeries(fileRefCDR, "RTUT");
                
        int[] tyear = g2tOld.getFyear();
        int[] tmonth = g2tOld.getFmonth();
        double[][][] T1 = g2tOld.getData();
        merger.addData(tyear, tmonth, T1);

        tyear = g2tRef.getFyear();
        tmonth = g2tRef.getFmonth();
        double[][][] T2 = g2tRef.getData();
        merger.addData(tyear, tmonth, T2);

        this.Tb = merger.mergeData();
        
        this.g2tCdr = new Geo2dTimeSeries(this.fyear, this.fmonth, this.Tb);
        this.g2tCdr.setDeaultTemperatureRange();
        infoLogger.info("DirCDR = "+ this.dirMltCDR);
        MltFileNetCDF netcdfWriter = new MltFileNetCDF("TUT", this.dirMltCDR, this.g2tCdr);
        netcdfWriter.writeDataToNetcdf();
        this.fileFullPathCDR = netcdfWriter.getCdrFilename();
    }

    /**
     * Get full path of the CDR file
     * @return full path of the CDR file
     */
    public String getCdrFileFullPath(){
        return this.fileFullPathCDR;
    }

    /**
     * show global mean for testing
     */
    private void showGlobalMean(){
        int cyear1 = MltCdrConstant.MLT_CDR_CLIMATOLOGY_YEAR_START;
        int cmonth1 = 1;
        int cyear2 = MltCdrConstant.MLT_CDR_CLIMATOLOGY_YEAR_END;
        int cmonth2 = 12;
        String msg;
        
        double[][][] anom = this.g2tCdr.getAnomaly(cyear1, cmonth1, cyear2, cmonth2);
        Geo2dTimeSeries g2t = new Geo2dTimeSeries(this.fyear, this.fmonth, anom);
        g2t.setDeaultTemperatureAnomalyRange();
        g2t.setDefaultValidLatIndex();
        double[] ym = g2t.getGlobalMean();
        
        for (int k=0; k<this.numberMonths; k++) {
            msg = String.format("%4d, %2d: %12.5f", this.fyear[k], this.fmonth[k], ym[k]);
            System.out.println(msg);
        }
    }


}
