/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.cdr.ReferenceTMT.java
*
* NAME
*   atms.cdr.ReferenceTMT.java
*
* PURPOSE
*   Generate Reference TMT CDR 
* 
* DESCRIPTION
*	This class is used to generate Reference TMT using satellites with stable orbits.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package cdr;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import ucar.nc2.Variable;
import ucar.ma2.*;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriter;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import geogrid.Geo2dTimeSeries;
import geogrid.GeoGrid2D;
import geogrid.GeoGridConstant;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;

/**
 * Class for generation of the reference TMT CDR
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class ReferenceTMT {
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");

    /**
     * Start year to use NPP data
     */    
    final private int nppYearStart = 2012;

    /**
     * Start month to use NPP data
     */
    final private int nppMonthStart = 1;

    /**
     * Start year to use N20 data
     */
    final private int n20YearStart = 2018;
    
     /**
     * Start month to use NPP data
     */
    final private int n20MonthStart = 1;
    
    /**
     * Start year of RTMT
     */
    final private int yearStart = 2002;

    /**
     * Start month of RTMT
     */
    final private int monthStart = 8;
    

    /**
     * End year of RTMT
     */
    private int yearEnd;

    /**
     * End month of RTMT
     */
    private int monthEnd;

    /**
     * Number of months
     */
    private int numberMonths;
    
    /**
     * Input directory
     */
    private String dirIn;

    /**
     * Directory of LUTs
     */
    private String dirLut;
    
    /**
     * Directory of outputs 
     */
    private String dirOut;
       
    /**
     * List of years for data
     */
    private int[] fyear;

    /**
     * List of months for data
     */
    private int[] fmonth;
    
    /**
     *  Time for data
     */
    private double[] ft;

    
    /**
     * Anomaly data
     */
    private double[][][] Tanom;
    
    /**
     * climatology data
     */
    private double[][][] Tclim;
    
    
    /**
     * Brightness temperature data
     */
    private double[][][] Tb;
    
    /**
     * Climatology of MetOp-A anomaly
     */
    private double[][][] climAnomMetopa; 

    /**
     * Adjusted ATMS anomaly
     */
    private double[][][] climAtmsAnomAdj = null; 
    
    /**
     * NPP time series, ascending orbits
     */
    private Geo2dTimeSeries g2tNppAsc = null;

    /**
     * NPP time series, descending orbits
     */
    private Geo2dTimeSeries g2tNppDsc = null;

    /**
     * N20 time series, ascending orbits
     */
    private Geo2dTimeSeries g2tN20Asc = null;
    
    /**
     * NPP time series, descending orbits
     */
    private Geo2dTimeSeries g2tN20Dsc = null;

    /**
     * Adjusted NPP time series
     */
    private Geo2dTimeSeries g2tNppAdj = null;

    /**
     * Adjusted N20 time series
     */
    private Geo2dTimeSeries g2tN20Adj = null;

    /**
     * Aqua anomaly time series
     */
    private Geo2dTimeSeries g2tAquaAnom = null;

    /**
     * MetOp-A anomaly time series
     */
    private Geo2dTimeSeries g2tMetopaAnom = null;
    
    /**
     * Sum for merging
     */
    private double[][][] vsumMerg = null; 

    /**
     * Pixel number for merging
     */
    private double[][][] pctMerg = null; 
    
    /**
     * Constructor with task configurations.
     * @param taskConf 
     */
    public ReferenceTMT(RefCdrTaskConfig taskConf) {
        this.yearEnd = taskConf.yearEnd;
        this.monthEnd = taskConf.monthEnd;
        this.numberMonths = (this.yearEnd - this.yearStart)*12 + (this.monthEnd - this.monthStart +1);
        
        this.initTime();
        
        this.Tanom = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tb = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tclim = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        this.dirIn = taskConf.dirL4;
        if (!this.dirIn.endsWith(File.separator)) {
            this.dirIn=this.dirIn + File.separator;
        }
        
        this.dirLut = taskConf.dirLut;
        if (!this.dirLut.endsWith(File.separator)) {
            this.dirLut=this.dirLut + File.separator;
        }
        
        this.dirOut = taskConf.dirRefCDR;
        if (!this.dirOut.endsWith(File.separator)) {
            this.dirOut=this.dirOut + File.separator;
        }               
    }
    
    /**
     * workflow to generate RTMT
     */
    public void processCDR() {
        this.loadAllL4Data();
        this.adjustSnppL4Data();
        this.adjustN20L4Data();
        this.mergeToRTMT();
        this.outputToFile();
    }
    
    /**
     * Load all L4 data for bias correction and merging
     */
    private void loadAllL4Data(){
        this.loadAquaL4Data();
        this.loadMetopaL4Data();
        this.loadSnppL4Data();
        this.loadN20L4Data();
        this.loadRefAnomClimatology();
        this.loadRefClim();
    }
    
    /**
     * Initialize time 
     */
    private void initTime(){
        int i, j , k;
        int  km, kyear, kmonth;
        int year1 = this.yearStart;
        int month1 = this.monthStart  ;
        int year2 = this.yearEnd;
        int month2 = this.monthEnd;
        this.numberMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numberMonths];
        this.fmonth = new int[numberMonths];
        this.ft = new double[numberMonths];
                
        for(i=0; i<numberMonths; i++){
            km = i + month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
    }

    
    /**
     * Adjust SNPP data with reference
     */
    private void adjustSnppL4Data(){
        int i, j, k, km, kmonth;
        double tmp1, tmp2;
        double[][][] anomAsc = this.g2tNppAsc.getAnomaly(nppYearStart, nppMonthStart, this.yearEnd, this.monthEnd);
        double[][][] anomDsc = this.g2tNppDsc.getAnomaly(nppYearStart, nppMonthStart, this.yearEnd, this.monthEnd);
        int nt = (this.yearEnd -  nppYearStart)*12 + (this.monthEnd - nppMonthStart +1);
        double[][][] anomMerg = new double[nt][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        double[][][] anomClimAdj = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        //merge snpp atms anomaly
        for(k=0; k<nt; k++){
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    tmp1 = anomAsc[k][i][j];
                    tmp2 = anomDsc[k][i][j];
                    if( (tmp1>=GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE)&(tmp1<=GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE) & (tmp2>=GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE)&(tmp2<=GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE)) {
                        // adjustment for SNPP ATMS
                        anomMerg[k][i][j] = (tmp1 + tmp2)/2.0d;
                    } else {
                        anomMerg[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                } 
            }
        }
        
        int[] tyear = this.g2tNppAsc.getFyear();
        int[] tmonth = this.g2tNppAsc.getFmonth();
        double ct1 = 2008;
        double ct2 = 2018;
        double vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        double vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        double[][][] anomClim2 = Geo2dTimeSeries.getClimatology(anomMerg, tyear, tmonth, ct1, ct2, vmin,vmax);
        
        for(k=0; k<12; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    anomClimAdj[k][i][j] = anomClim2[k][i][j] - climAnomMetopa[k][i][j];
                }
            }
        }
                
        //adjust SNPP data
        for(k=0; k<nt; k++){
            kmonth = tmonth[k];
            km = kmonth - 1;
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    anomMerg[k][i][j] = anomMerg[k][i][j] - anomClimAdj[km][i][j];
                }
            }
        }        
        this.g2tNppAdj = new Geo2dTimeSeries(tyear,tmonth, anomMerg);
        this.g2tNppAdj.setValidRange(vmin, vmax);
        infoLogger.info("SNPP data adjusted");
                
    }

    /**
     * Adjust N20 data with SNPP as reference
     */
    private void adjustN20L4Data(){
        int i, j, k, km, kmonth;
        double tmp1, tmp2;
        
        double[][][] anomAsc = this.g2tN20Asc.getAnomaly(nppYearStart, nppMonthStart, this.yearEnd, this.monthEnd);
        double[][][] anomDsc = this.g2tN20Dsc.getAnomaly(nppYearStart, nppMonthStart, this.yearEnd, this.monthEnd);
        int nt = (this.yearEnd -  n20YearStart)*12 + (this.monthEnd - n20MonthStart +1);
        double[][][] anomMerg = new double[nt][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        double[][][] anomClimAdj = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        //merge snpp atms anomaly
        for(k=0; k<nt; k++){
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    tmp1 = anomAsc[k][i][j];
                    tmp2 = anomDsc[k][i][j];
                    if( (tmp1>=GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE)&(tmp1<=GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE) & (tmp2>=GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE)&(tmp2<=GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE)) {
                        // adjustment for SNPP ATMS
                        anomMerg[k][i][j] = (tmp1 + tmp2)/2.0d;
                    } else {
                        anomMerg[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                } 
            }
        }
        
        int[] tyear = this.g2tN20Asc.getFyear();
        int[] tmonth = this.g2tN20Asc.getFmonth();
        double ct1 = 2018;
        double ct2 = 2023;// use data from 2018 to 2022 as common period to correct n20 with snpp
        double vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        double vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        //N20 climatology for common period
        double[][][] anomClim2 = Geo2dTimeSeries.getClimatology(anomMerg, tyear, tmonth, ct1, ct2, vmin,vmax);
        double[][][] anomAdjustedNpp = this.g2tNppAdj.getData();
        int[] tyearNpp = this.g2tNppAdj.getFyear();
        int[] tmonthNpp = this.g2tNppAdj.getFmonth();
        double[][][] anomClimRefNpp = Geo2dTimeSeries.getClimatology(anomAdjustedNpp, tyearNpp, tmonthNpp, ct1, ct2, vmin,vmax);
        
        for(k=0; k<12; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    anomClimAdj[k][i][j] = anomClim2[k][i][j] - anomClimRefNpp[k][i][j];
                }
            }
        }
                
        //adjust N20 data
        for(k=0; k<nt; k++){
            kmonth = tmonth[k];
            km = kmonth - 1;
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    anomMerg[k][i][j] = anomMerg[k][i][j] - anomClimAdj[km][i][j];
                }
            }
        }        
        this.g2tN20Adj = new Geo2dTimeSeries(tyear,tmonth, anomMerg);
        this.g2tN20Adj.setValidRange(vmin, vmax);        
        infoLogger.info("N20 data adjusted");        
    }
    
    /**
     *  Prepare to merge data
     */
    private void prepareMerge(){
        int k, i, j;
        this.vsumMerg = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.pctMerg = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];      
        for(k=0; k<this.numberMonths; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    this.vsumMerg[k][i][j] = 0.0;
                    this.pctMerg[k][i][j] = 0.0;
                }
            }
        }        
    }
    
    
/**
 * add data for merging
 * @param tyear, year list of the data
 * @param tmonth, month list of the data
 * @param vdata, data time series 
 */
    private void AddData(int[] tyear, int[] tmonth, double[][][] vdata){
        double vmin, vmax, vtmp;
        int i,j,k, km, kmonth, kyear;
        vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        int mt = tyear.length;
        for(k=0; k<mt; k++){
            kyear = tyear[k];
            kmonth = tmonth[k];
            km = (kyear - this.yearStart)*12 + (kmonth - this.monthStart); // time index
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp = vdata[k][i][j];
                    if ((vtmp >=vmin) &(vtmp <= vmax)) {
                        this.vsumMerg[km][i][j] = this.vsumMerg[km][i][j] + vtmp; 
                        this.pctMerg[km][i][j] = this.pctMerg[km][i][j] + 1.0d; 
                    }
                }
            }
        }     
    }
    
    /**
     * Merge anomaly of Aqua, MetOp-A, SNPP and N20 together to RTMT anomaly
     */
    private void mergeToRTMT() {
        int i,j,k,km, kyear, kmonth;
        double vtmp, vmin, vmax;
        vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        
        infoLogger.info("Start merging RTMT ...");
        // start merging
        this.prepareMerge();
        
        // add Aqua data
        int[] tyear = this.g2tAquaAnom.getFyear();
        int[] tmonth = this.g2tAquaAnom.getFmonth();
        double[][][] vdata = this.g2tAquaAnom.getData();
        this.AddData(tyear, tmonth, vdata);
        
        // add MetOpa
        tyear = this.g2tMetopaAnom.getFyear();
        tmonth = this.g2tMetopaAnom.getFmonth();
        vdata = this.g2tMetopaAnom.getData();
        this.AddData(tyear, tmonth, vdata);
        
        // add SNPP
        tyear = this.g2tNppAdj.getFyear();
        tmonth = this.g2tNppAdj.getFmonth();
        vdata = this.g2tNppAdj.getData();
        this.AddData(tyear, tmonth, vdata);
        
        // add N20
        tyear = this.g2tN20Adj.getFyear();
        tmonth = this.g2tN20Adj.getFmonth();
        vdata = this.g2tN20Adj.getData();
        this.AddData(tyear, tmonth, vdata);
        
        // for testing
        //this.saveMergedData();
        
        // do average
        for (k=0; k<this.numberMonths; k++) {
            kmonth = this.fmonth[k];
            km = kmonth-1;
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    if (this.pctMerg[k][i][j] > 0) {
                        this.Tanom[k][i][j] = this.vsumMerg[k][i][j]/this.pctMerg[k][i][j];
                        this.Tb[k][i][j] = this.Tanom[k][i][j]  + this.Tclim[km][i][j];
                    } else {
                        this.Tanom[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                        this.Tb[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }       
        
//        System.out.println("anom(1, 40, 100)=" + this.Tanom[0][39][99]);
        // for testing
        //this.countValidPixels();
        
    }
    
    /**
     * For testing, count valid data numbers of each month
     */
    private void countValidPixels(){
        double[] zm = new double[this.numberMonths];
        double tmp =0;
        double tmp1  = 0;
        double tmp2=0;
        double ct1 = 0;
        double vtmp;
        for(int k=0; k<this.numberMonths; k++) {
            tmp = 0;
            tmp2 = 0;
            ct1=0;
            for(int i=0;i<GEO_GRID_NUM_LAT; i++) {
                for(int j=0; j<GEO_GRID_NUM_LON; j++) {
                    tmp = tmp + this.pctMerg[k][i][j];
                    tmp2 = tmp2 + this.vsumMerg[k][i][j];  
                    vtmp = this.Tanom[k][i][j];
                    if ( (vtmp>=GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE) & ( vtmp<=GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE)) {
                        tmp1 = tmp1 + vtmp;    
                        ct1++;
                    }
                }
            }
            zm[k] = tmp;
            if (ct1>0) {
                tmp1 = tmp1/ct1;
            } else {
                tmp1 = -9999;
            }
            if (tmp>0) {
                String msg = String.format("year=%4d, month=%2d, anom=%10.5f,  mean=%10.5f, total pixels=%9.1f", this.fyear[k], this.fmonth[k], tmp1, tmp2/tmp, tmp);
                System.out.println(msg);
            } else {
                String msg = String.format("year=%4d, month=%2d, anom=%10.5f,  mean=%10.5f, total pixels=%9.1f", this.fyear[k], this.fmonth[k], tmp1, 0.0d, tmp);
                System.out.println(msg);                
            }
        }
                
    }
    
    
    /**
     * For testing only, save merged data to file before calculating mean
     */
    private void saveMergedData(){
        String fileTest = this.dirOut + "Testing_merged_RTMT.nc";
        NetcdfFileWriter ncFileOut = null;
        Variable dataVar = null;
        ArrayInt.D1 iarr1 = null;
        ArrayDouble.D3 darr3 = null;
        ArrayDouble.D1 darr1 = null;
        try{
            ncFileOut = NetcdfFileWriter.createNew(NetcdfFileWriter.Version.netcdf3, fileTest);
            
            // create Dimensions
            Dimension dimTime = ncFileOut.addDimension(null, "time", this.numberMonths);
            Dimension dimLon = ncFileOut.addDimension(null, "longitude",GEO_GRID_NUM_LON);
            Dimension dimLat = ncFileOut.addDimension(null, "latitude", GEO_GRID_NUM_LAT);
            Dimension dimMonth = ncFileOut.addDimension(null, "month",12);

            List<Dimension> dimTimeD = new ArrayList<Dimension>();
            dimTimeD.add(dimTime);

            List<Dimension> dimLatD = new ArrayList<Dimension>();
            dimLatD.add(dimLat);

            List<Dimension> dimLonD = new ArrayList<Dimension>();
            dimLonD.add(dimLon);
            
            List<Dimension> dimMonD = new ArrayList<Dimension>();
            dimMonD.add(dimMonth);
            
            List<Dimension> dimGridD = new ArrayList<Dimension>();
            dimGridD.add(dimTime);
            dimGridD.add(dimLat);
            dimGridD.add(dimLon);

            List<Dimension> dimGridClim = new ArrayList<Dimension>();
            dimGridClim.add(dimMonth);
            dimGridClim.add(dimLat);
            dimGridClim.add(dimLon);
            // define corridinates variable variables
            Variable vlon = ncFileOut.addVariable(null, "longitude", DataType.FLOAT, dimLonD);
            Variable vlat = ncFileOut.addVariable(null, "latitude", DataType.FLOAT, dimLatD);
            Variable vmon = ncFileOut.addVariable(null, "month", DataType.INT, dimTimeD);
            Variable vyear = ncFileOut.addVariable(null, "fyear", DataType.INT, dimTimeD);
            Variable vmonth = ncFileOut.addVariable(null, "fmonth", DataType.INT, dimTimeD);
            Variable vft = ncFileOut.addVariable(null, "ft", DataType.DOUBLE, dimTimeD);
            vlat.addAttribute(new Attribute("units", "degree"));
            vlon.addAttribute(new Attribute("units", "degree"));
            
            // define data variables and attributes
            String varName = "bt_sum";
            Variable varBt= ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridD);
            varBt.addAttribute(new Attribute("coordinates","time latitude longitude" ));
            // define data variables and attributes
            varName = "pct";
            Variable varPix= ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridD);
            varBt.addAttribute(new Attribute("coordinates","time latitude longitude" ));
            ncFileOut.create();
            // write data
            ArrayFloat.D1 dataLat = new ArrayFloat.D1(GEO_GRID_NUM_LAT);
            ArrayFloat.D1 dataLon = new ArrayFloat.D1(GEO_GRID_NUM_LON);      
            ArrayInt.D1 dataCmonth = new ArrayInt.D1(12);
            ArrayInt.D1 dataFyear = new ArrayInt.D1(this.numberMonths);
            ArrayInt.D1 dataFmonth = new ArrayInt.D1(this.numberMonths);
            ArrayDouble.D1 dataFt = new ArrayDouble.D1(this.numberMonths);
            // write latitude   
            float[] plat = GeoGrid2D.getGridLatitudeFloat();
            for (int i=0; i<GEO_GRID_NUM_LAT; i++) {
                dataLat.set(i, plat[i]);
            }
            ncFileOut.write(vlat, dataLat);
            
            // write longitude
            float[] plon = GeoGrid2D.getGridLongitudeFloat();
            for(int i=0; i<GEO_GRID_NUM_LON; i++){
                dataLon.set(i, plon[i]);
            }
            ncFileOut.write(vlon, dataLon);
            
            // write year 
            for(int i=0; i<this.numberMonths; i++) {
                dataFyear.set(i, this.fyear[i]);
            }
            ncFileOut.write(vyear, dataFyear);
            
            // write fmonth
            for(int i=0; i<this.numberMonths; i++) {
                dataFmonth.set(i, fmonth[i]);
            }
            ncFileOut.write(vmonth, dataFmonth);

            
            // write time
            for(int i=0; i<this.numberMonths; i++) {
                dataFt.set(i, ft[i]);
            }
            ncFileOut.write(vft, dataFt);
            
            // write month
            for(int i=0; i<12; i++) {
                dataCmonth.set(i, i+1);
            }
            ncFileOut.write(vmon, dataCmonth);

            // write vsum
            ArrayDouble.D3 vdata = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (int i=0; i<this.numberMonths; i++) {
                for(int j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(int k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.vsumMerg[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varBt, vdata);

            // write pixel number
            vdata = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (int i=0; i<this.numberMonths; i++) {
                for(int j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(int k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.pctMerg[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varPix, vdata);      
        } catch(Exception exp) {
            exp.printStackTrace();
        }

        if (ncFileOut != null) {
            try {
                    ncFileOut.close();
                } catch (Exception exp) {
                    errLogger.error("Failed to close netcdf file for testing");
                    exp.printStackTrace();
                }
        }
        
    }
    
    
    /**
     * Out put the RTMT data to a netcdf file
     */
    private void outputToFile(){
        int i, j, k;
        String filename = this.dirOut + getCDRFilename(this.yearEnd, this.monthEnd);
        System.out.println("Writing RTMT data to file " + filename);
        NetcdfFileWriter ncFileOut = null;
        Variable dataVar = null;
        ArrayInt.D1 iarr1 = null;
        ArrayDouble.D3 darr3 = null;
        ArrayDouble.D1 darr1 = null;
        try{
            ncFileOut = NetcdfFileWriter.createNew(NetcdfFileWriter.Version.netcdf3, filename);
            
            // create Dimensions
            Dimension dimTime = ncFileOut.addDimension(null, "time", this.numberMonths);
            Dimension dimLon = ncFileOut.addDimension(null, "longitude",GEO_GRID_NUM_LON);
            Dimension dimLat = ncFileOut.addDimension(null, "latitude", GEO_GRID_NUM_LAT);
            Dimension dimMonth = ncFileOut.addDimension(null, "month",12);

            List<Dimension> dimTimeD = new ArrayList<Dimension>();
            dimTimeD.add(dimTime);

            List<Dimension> dimLatD = new ArrayList<Dimension>();
            dimLatD.add(dimLat);

            List<Dimension> dimLonD = new ArrayList<Dimension>();
            dimLonD.add(dimLon);
            
            List<Dimension> dimMonD = new ArrayList<Dimension>();
            dimMonD.add(dimMonth);
            
            List<Dimension> dimGridD = new ArrayList<Dimension>();
            dimGridD.add(dimTime);
            dimGridD.add(dimLat);
            dimGridD.add(dimLon);

            List<Dimension> dimGridClim = new ArrayList<Dimension>();
            dimGridClim.add(dimMonth);
            dimGridClim.add(dimLat);
            dimGridClim.add(dimLon);

            // define corridinates variable variables
            Variable vlon = ncFileOut.addVariable(null, "longitude", DataType.FLOAT, dimLonD);
            Variable vlat = ncFileOut.addVariable(null, "latitude", DataType.FLOAT, dimLatD);
            Variable vmon = ncFileOut.addVariable(null, "month", DataType.INT, dimTimeD);
            Variable vyear = ncFileOut.addVariable(null, "fyear", DataType.INT, dimTimeD);
            Variable vmonth = ncFileOut.addVariable(null, "fmonth", DataType.INT, dimTimeD);
            Variable vft = ncFileOut.addVariable(null, "ft", DataType.DOUBLE, dimTimeD);
            vlat.addAttribute(new Attribute("units", "degree"));
            vlon.addAttribute(new Attribute("units", "degree"));
            
            // define data variables and attributes
            String varName = MltCdrConstant.VARNAME_BRIGHTNESS_TEMPERATURE_RTMT;
            Variable varBt= ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridD);
            varBt.addAttribute(new Attribute("units","Kelvin" ));
            varBt.addAttribute(new Attribute("valid_max", GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE));
            varBt.addAttribute(new Attribute("valid_min", GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE));
            varBt.addAttribute(new Attribute("_FillValue", GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varBt.addAttribute(new Attribute("coordinates","time latitude longitude" ));

            varName = MltCdrConstant.VARNAME_BRIGHTNESS_TEMPERATURE_RTMT_ANOMALY;
            Variable varAnom = ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridD);
            varAnom.addAttribute(new Attribute("units","Kelvin" ));
            varAnom.addAttribute(new Attribute("valid_max", GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE));
            varAnom.addAttribute(new Attribute("valid_min", GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE));
            varAnom.addAttribute(new Attribute("_FillValue", GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varAnom.addAttribute(new Attribute("coordinates","time latitude longitude" ));

            // define data variables and attributes
            varName = MltCdrConstant.VARNAME_BRIGHTNESS_TEMPERATURE_RTMT_CLIMATOLOGY;
            Variable varClim= ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridClim);
            varClim.addAttribute(new Attribute("units","Kelvin" ));
            varClim.addAttribute(new Attribute("valid_max", GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE));
            varClim.addAttribute(new Attribute("valid_min", GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE));
            varClim.addAttribute(new Attribute("_FillValue", GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varClim.addAttribute(new Attribute("coordinates","month latitude longitude" ));
            
            ncFileOut.create();
            // write data
            ArrayFloat.D1 dataLat = new ArrayFloat.D1(GEO_GRID_NUM_LAT);
            ArrayFloat.D1 dataLon = new ArrayFloat.D1(GEO_GRID_NUM_LON);      
            ArrayInt.D1 dataCmonth = new ArrayInt.D1(12);
            ArrayInt.D1 dataFyear = new ArrayInt.D1(this.numberMonths);
            ArrayInt.D1 dataFmonth = new ArrayInt.D1(this.numberMonths);
            ArrayDouble.D1 dataFt = new ArrayDouble.D1(this.numberMonths);
            // write latitude   
            float[] plat = GeoGrid2D.getGridLatitudeFloat();
            for (i=0; i<GEO_GRID_NUM_LAT; i++) {
                dataLat.set(i, plat[i]);
            }
            ncFileOut.write(vlat, dataLat);
            
            // write longitude
            float[] plon = GeoGrid2D.getGridLongitudeFloat();
            for(i=0; i<GEO_GRID_NUM_LON; i++){
                dataLon.set(i, plon[i]);
            }
            ncFileOut.write(vlon, dataLon);
            
            // write year 
            for(i=0; i<this.numberMonths; i++) {
                dataFyear.set(i, this.fyear[i]);
            }
            ncFileOut.write(vyear, dataFyear);
            
            // write fmonth
            for(i=0; i<this.numberMonths; i++) {
                dataFmonth.set(i, fmonth[i]);
            }
            ncFileOut.write(vmonth, dataFmonth);

            
            // write time
            for(i=0; i<this.numberMonths; i++) {
                dataFt.set(i, ft[i]);
            }
            ncFileOut.write(vft, dataFt);
            
            // write month
            for(i=0; i<12; i++) {
                dataCmonth.set(i, i+1);
            }
            ncFileOut.write(vmon, dataCmonth);

            // write RTMT
            ArrayDouble.D3 vdata = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<this.numberMonths; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.Tb[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varBt, vdata);

            // write Anomaly
            ArrayDouble.D3 vdata2 = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<this.numberMonths; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata2.set(i,j,k, this.Tanom[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varAnom, vdata2);

            // write climatology
            vdata = new ArrayDouble.D3(12, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<12; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.Tclim[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varClim, vdata);
            
        }catch(Exception e) {
            e.printStackTrace();
            System.out.println("Failed to write RTMT data to netcdf file " + filename);
        }

        if (ncFileOut != null) {
            try {
                    ncFileOut.close();
                } catch (Exception exp) {
                    errLogger.error("Failed to close netcdf file" + filename);
                    exp.printStackTrace();
                }
        }

    }
    
    
    /**
     * Get short name of the RTMT file
     */
    public static String  getCDRFilename(int year2, int month2){
        String filename = String.format("NESDIS-STAR_RTMT_Anomaly_AMSUA_ATMS_V01_S200208_E%4d%02d.nc", year2, month2);        
        return filename;                
    }
    
    /**
     * Load SNPP ATMS L4 data
     */
    private void loadSnppL4Data() {
        String filename = this.dirIn + "ATMS_L4_npp_ch06.nc";
        this.g2tNppAsc = new Geo2dTimeSeries(filename, "bt_asc");
        this.g2tNppDsc = new Geo2dTimeSeries(filename, "bt_dsc");
        this.g2tNppAsc.setDeaultTemperatureRange();
        this.g2tNppDsc.setDeaultTemperatureRange();
        System.out.println("NPP data loaded for RTMT merging");

    }

    /**
     * Load N20 ATMS L4 data
     */
    private void loadN20L4Data() {
        String filename = this.dirIn + "ATMS_L4_j01_ch06.nc";
        this.g2tN20Asc = new Geo2dTimeSeries(filename, "bt_asc");
        this.g2tN20Dsc = new Geo2dTimeSeries(filename, "bt_dsc");
        this.g2tN20Asc.setDeaultTemperatureRange();
        this.g2tN20Dsc.setDeaultTemperatureRange();
        System.out.println("N20 data loaded for RTMT merging");
    }
    
    /**
     * Load climatology for RTMT adjustment
     */
    private void loadRefClim(){
        String fileRef = this.dirLut+"RTMT_Climatology.nc";
        this.Tclim = Geo2dTimeSeries.readGeo2dTimeSeriesDouble(fileRef, "RTMT_Climatology");
        System.out.println("climatology data loaded for RTMT merging");
    }
    
    /**
     * Load Aqua L4 data for bias correction and merging
     */
    private void loadAquaL4Data(){
        String filename = this.dirIn + "AMSUA_L4_aqua_ch05.nc";
        this.g2tAquaAnom = new Geo2dTimeSeries(filename, "bt_anom");
        this.g2tAquaAnom.setDefaultValidLatIndex();  
        this.g2tAquaAnom.setDeaultTemperatureAnomalyRange();
        System.out.println("Aqua data loaded for RTMT merging");
    }

    /**
     * Load MetOp-A L4 data for bias correction and merging
     */
    private void loadMetopaL4Data(){
        String filename = this.dirIn + "AMSUA_L4_metopa_ch05.nc";
        this.g2tMetopaAnom = new Geo2dTimeSeries(filename, "bt_anom");
        this.g2tMetopaAnom.setDefaultValidLatIndex();        
        this.g2tMetopaAnom.setDeaultTemperatureAnomalyRange();
        System.out.println("MetOp-A data loaded for RTMT merging");
    }
    
  
    /**
     *  Read reference climatology from MetOP-A 2008-2017 data.  This anom climatology was generated 
     */
    private void loadRefAnomClimatology() {
        String fileRef = this.dirLut+"lut_ref_metopa_ch05.nc";
        this.climAnomMetopa = Geo2dTimeSeries.readGeo2dTimeSeriesDouble(fileRef, "clim_metopa_anom");
        System.out.println("Loaded reference anomaly data loaded for RTMT merging");

    }
    
    /**
     *  For testing only
     * @param args 
     */
    public static void main(String[] args){
        int year2 =  2023;
        int month2 = 12;
        String dirLut = "/home/xhao/data/mltcdrv5/lut/ref/";
        String dirL4 = "/home/xhao/data/mltcdrv5/data/L4/";
        String dirRefCDR = "/home/xhao/data/mltcdrv5/data/rfcdr/";
        RefCdrTaskConfig taskConf = new RefCdrTaskConfig(year2, month2,dirLut,dirL4, dirRefCDR);
        ReferenceTMT rtmt = new ReferenceTMT(taskConf);
        rtmt.processCDR();
    }
}
