/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.cdr.ReferenceTUT.java
*
* NAME
*   atms.cdr.ReferenceTUT.java
*
* PURPOSE
*   Generate Reference TUT CDR 
* 
* DESCRIPTION
*	This class is used to generate Reference TUT using satellites with stable orbits.
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package cdr;


import java.io.File;
import java.util.ArrayList;
import java.util.List;

import ucar.nc2.Variable;
import ucar.ma2.*;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriter;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import geogrid.Geo2dTimeSeries;
import geogrid.GeoGrid2D;
import geogrid.GeoGridConstant;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;


/**
 * Class to generate Reference TUT generated from measurements of satellites with stable orbits.  
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class ReferenceTUT {
    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");

    /**
     *  Start year to use NPP as reference 
     */
    final private int nppClimYear1 = 2012;

    /**
     *  End year to use NPP as reference for climatology
     */
    final private int nppClimYear2 = 2022;

    /**
     *  Start month to use NPP as reference 
     */
    final private int nppClimMonth1 = 1;

    /**
     *  End month to use NPP as reference 
     */
    final private int nppClimMonth2 = 12;
  
    /**
     * Start year to use NPP data
     */
    final private int nppYearStart = 2012;

    /**
     * Start month to use NPP data
     */
    final private int nppMonthStart = 1;
    
    /**
     * Start year to use N20 data
     */
    final private int n20YearStart = 2017;
    
    /**
     * Start month to use N20 data
     */
    final private int n20MonthStart = 12;

    /**
     * Start year to use Aqua data
     */
    final private int aquaYearStart = 2002;

    /**
     * Start month to use Aqua data
     */
    final private int aquaMonthStart = 1;

    /**
     * end year to use Aqua data
     */
    final private int aquaYearEnd = 2019;

    /**
     * End month to use Aqua data
     */
    final private int aquaMonthEnd = 12;

    /** 
     * Start year of RTUT climatology base period
     */
    final private int rtutClimYear1 = nppClimYear1;

    /** 
     * Start month of RTUT climatology base period
     */
    final private int rtutClimMonth1 = nppClimMonth1;

    /** 
     * End year of RTUT climatology base period
     */
    final private int rtutClimYear2 = nppClimYear2;

    /** 
     * End month of RTUT climatology base period
     */
    final private int rtutClimMonth2 = nppClimMonth2;
      
    /** 
     * Start year of RTUT data
     */
    final private int yearStart = 2002;

    /** 
     * Start month of RTUT data
     */
    final private int monthStart = 8;
    
    /**
     * End year of RTUT
     */
    private int yearEnd;

    /**
     * End month of RTUT
     */
    private int monthEnd;

    /**
     * Number of months
     */
    private int numberMonths;
    
    /**
     * Input directory
     */
    private String dirIn;

    /**
     * Directory of LUTs
     */
    private String dirLut;
    
    /**
     * Directory of outputs 
     */
    private String dirOut;
       
    /**
     * List of years for data
     */
    private int[] fyear;

    /**
     * List of months for data
     */
    private int[] fmonth;
    
    /**
     *  Time for data
     */
    private double[] ft;
    
    /**
     * Anomaly data
     */
    private double[][][] Tanom;
    
    /**
     * climatology data, using SNPP as reference
     */
    private double[][][] Tclim;
    
    
    /**
     * Brightness temperature data
     */
    private double[][][] Tb;
        
    /**
     * NPP time series
     */
    private Geo2dTimeSeries g2tNppAnom = null;

    /**
     * N20 time series
     */
    private Geo2dTimeSeries g2tN20Anom = null;

    /**
     * Aqua time series
     */
    private Geo2dTimeSeries g2tAquaAnom = null;
    
    /**
     * Adjusted Aqua time series
     */
    private Geo2dTimeSeries g2tAquaAdj = null;

    /**
     * Adjusted N20 time series
     */
    private Geo2dTimeSeries g2tN20Adj = null;
    
    /**
     * Sum for data merging
     */
    private double[][][] vsumMerg = null; 

    /**
     * Pixel number for data merging
     */
    private double[][][] pctMerg = null; 
        
    /**
     * Constructor with task configurations.
     * @param taskConf 
     */
    public ReferenceTUT(RefCdrTaskConfig taskConf) {
        this.yearEnd = taskConf.yearEnd;
        this.monthEnd = taskConf.monthEnd;
        this.numberMonths = (this.yearEnd - this.yearStart)*12 + (this.monthEnd - this.monthStart +1);
        
        this.initTime();
        
        this.Tanom = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tb = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.Tclim = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        this.dirIn = taskConf.dirL4;
        if (!this.dirIn.endsWith(File.separator)) {
            this.dirIn=this.dirIn + File.separator;
        }
        
        this.dirLut = taskConf.dirLut;
        if (!this.dirLut.endsWith(File.separator)) {
            this.dirLut=this.dirLut + File.separator;
        }
        
        this.dirOut = taskConf.dirRefCDR;
        if (!this.dirOut.endsWith(File.separator)) {
            this.dirOut=this.dirOut + File.separator;
        }               
    }
    
    
    
    /**
     * Initialize time 
     */
    private void initTime(){
        int i, j , k;
        int  km, kyear, kmonth;
        int year1 = this.yearStart;
        int month1 = this.monthStart  ;
        int year2 = this.yearEnd;
        int month2 = this.monthEnd;
        this.numberMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numberMonths];
        this.fmonth = new int[numberMonths];
        this.ft = new double[numberMonths];
                
        for(i=0; i<numberMonths; i++){
            km = i + month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
    }

    /**
     * Get anomaly data from L4 data file with specified normal period.
     * @param filenameL4, filenameL4, full path of L4 data file.
     * @param cyear1, start year of the normal period for climatology.
     * @param cmonth1, start month of the normal period for climatology.
     * @param cyear2, end year of the normal period for climatology.
     * @param cmonth2, end month of the normal period for climatology.
     * @param isRef, if it's reference data. If true, use it to get RTUT climatology. 
     * @return Geo2dTimeSeries for anomaly time series
     */
    private Geo2dTimeSeries loadAnomalyData(String filenameL4, int cyear1, int cmonth1, int cyear2, int cmonth2, boolean isRef) {
        int i, j, k;
        double vtmp1, vtmp2;
        Geo2dTimeSeries g2tAsc = new Geo2dTimeSeries(filenameL4, "bt_asc");
        Geo2dTimeSeries g2tDsc = new Geo2dTimeSeries(filenameL4, "bt_dsc");
        g2tAsc.setDefaultValidLatIndex();  
        g2tAsc.setDeaultTemperatureRange();
        g2tDsc.setDefaultValidLatIndex();  
        g2tDsc.setDeaultTemperatureRange();

        int nt = g2tAsc.getTimeSeriesLength();
        
        double[][][] anomAsc = g2tAsc.getAnomaly(cyear1, cmonth1, cyear2, cmonth2);
        double[][][] anomDsc = g2tDsc.getAnomaly(cyear1, cmonth1, cyear2, cmonth2);

        double vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE;
        double vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE;
        
        
        if (isRef) {
            double[][][] climAsc = g2tAsc.getClimatology(cyear1, cmonth1, cyear2, cmonth2);
            double[][][] climDsc = g2tDsc.getClimatology(cyear1, cmonth1, cyear2, cmonth2);
            for(k=0; k<12; k++) {
                for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                    for(j=0; j<GEO_GRID_NUM_LON; j++) {
                        vtmp1 = climAsc[k][i][j];
                        vtmp2 = climDsc[k][i][j];
                        if ((vtmp1 >= vmin) & (vtmp1 <= vmax) & (vtmp2 >= vmin) & (vtmp2 <= vmax)) {
                            this.Tclim[k][i][j] = (climAsc[k][i][j] + climDsc[k][i][j]) / 2.0d;
                        } else {
                            this.Tclim[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                        }
                    }
                }
            }
        }
        
        double[][][] anomAll = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        
        for(k=0; k<nt; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp1 = anomAsc[k][i][j];
                    vtmp2 = anomDsc[k][i][j];
                    if ( (vtmp1>=vmin)&(vtmp1<=vmax) & (vtmp2>=vmin) & (vtmp2 <= vmax)) {
                            anomAll[k][i][j] = (anomAsc[k][i][j] + anomDsc[k][i][j])/2.0d; 
                    } else {
                        anomAll[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }
        int[] tyear = g2tAsc.getFyear();
        int[] tmonth = g2tAsc.getFmonth();
        Geo2dTimeSeries g2t = new  Geo2dTimeSeries(tyear, tmonth, anomAll);
        g2t.setDefaultValidLatIndex();
        g2t.setDeaultTemperatureAnomalyRange();
        return g2t;
    }
    
    /**
     * Get anomaly data from L4 data file.
     * @param filenameL4, full path of L4 data file.
     * @return time series of anomaly data
     */
    private Geo2dTimeSeries loadAnomalyData(String filenameL4) {
        int i, j, k;
        double vtmp1, vtmp2;
        Geo2dTimeSeries g2tAsc = new Geo2dTimeSeries(filenameL4, "bt_asc");
        Geo2dTimeSeries g2tDsc = new Geo2dTimeSeries(filenameL4, "bt_dsc");
        g2tAsc.setDefaultValidLatIndex();  
        g2tAsc.setDeaultTemperatureRange();
        g2tDsc.setDefaultValidLatIndex();  
        g2tDsc.setDeaultTemperatureRange();

        double[][][] anomAsc = g2tAsc.getAnomaly();
        double[][][] anomDsc = g2tDsc.getAnomaly();
        double[][][] anomAll = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        double vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        double vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;

        int nt = g2tAsc.getTimeSeriesLength();
        
        for(k=0; k<nt; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp1 = anomAsc[k][i][j];
                    vtmp2 = anomDsc[k][i][j];
                    if ( (vtmp1>=vmin)&(vtmp1<=vmax) & (vtmp2>=vmin) & (vtmp2 <= vmax)) {
                            anomAll[k][i][j] = (anomAsc[k][i][j] + anomDsc[k][i][j])/2.0d; 
                    } else {
                        anomAll[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }
        int[] tyear = g2tAsc.getFyear();
        int[] tmonth = g2tAsc.getFmonth();
        Geo2dTimeSeries g2t = new  Geo2dTimeSeries(tyear, tmonth, anomAll);
        g2t.setDefaultValidLatIndex();
        g2t.setDeaultTemperatureAnomalyRange();
        return g2t;
    }

    /**
     * The workflow to generate RTMT
     */
    public void processCDR() {
        this.loadAllL4Data();
        this.adjustAquaL4Data();
        this.adjustN20L4Data();
        this.mergeToRTUT();
        this.outputToFile();
    }
    
    
    /**
     * Merge data to RTUT CDR
     */
    private void mergeToRTUT(){
        int i,j,k,km, kyear, kmonth;
        double vtmp, vmin, vmax;
        vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        
        infoLogger.info("Start merging RTUT ...");
        // start merging
        this.prepareMerge();
        
        // add Aqua data
        int[] tyear = this.g2tAquaAdj.getFyear();
        int[] tmonth = this.g2tAquaAdj.getFmonth();
        double[][][] vdata = this.g2tAquaAdj.getData();
        this.addData(tyear, tmonth, vdata);
                
        // add SNPP
        tyear = this.g2tNppAnom.getFyear();
        tmonth = this.g2tNppAnom.getFmonth();
        vdata = this.g2tNppAnom.getData();
        this.addData(tyear, tmonth, vdata);
        
        // add N20
        tyear = this.g2tN20Adj.getFyear();
        tmonth = this.g2tN20Adj.getFmonth();
        vdata = this.g2tN20Adj.getData();
        this.addData(tyear, tmonth, vdata);
                
        // do average
        for (k=0; k<this.numberMonths; k++) {
            kmonth = this.fmonth[k];
            km = kmonth-1;
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    if (this.pctMerg[k][i][j] > 0) {
                        this.Tanom[k][i][j] = this.vsumMerg[k][i][j]/this.pctMerg[k][i][j];
                        this.Tb[k][i][j] = this.Tanom[k][i][j]  + this.Tclim[km][i][j];
                    } else {
                        this.Tanom[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                        this.Tb[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }       
                
    }

    
/**
 * add data for merging
 * @param tyear, year list of the data
 * @param tmonth, month list of the data
 * @param vdata, data time series 
 */
    private void addData(int[] tyear, int[] tmonth, double[][][] vdata){
        double vmin, vmax, vtmp;
        int i,j,k, km, kmonth, kyear;
        vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        int mt = tyear.length;
        for(k=0; k<mt; k++){
            kyear = tyear[k];
            kmonth = tmonth[k];
            km = (kyear - this.yearStart)*12 + (kmonth - this.monthStart); // time index
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp = vdata[k][i][j];
                    if ((vtmp >=vmin) &(vtmp <= vmax)) {
                        this.vsumMerg[km][i][j] = this.vsumMerg[km][i][j] + vtmp; 
                        this.pctMerg[km][i][j] = this.pctMerg[km][i][j] + 1.0d; 
                    }
                }
            }
        }     
    }
    
    /**
     *  Prepare to merge data
     */
    private void prepareMerge(){
        int k, i, j;
        this.vsumMerg = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.pctMerg = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];      
        for(k=0; k<this.numberMonths; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    this.vsumMerg[k][i][j] = 0.0;
                    this.pctMerg[k][i][j] = 0.0;
                }
            }
        }        
    }    
    /**
     * Adjust Aqua data to reference
     */
    private void adjustAquaL4Data(){
        int i, j, k, km, kmonth;
        int indAqua, indSnpp;
        double vtmp1, vtmp2;
        double vmin, vmax;
        int cyear1 = this.rtutClimYear1;
        int cmonth1 = this.rtutClimMonth1;
        int cyear2 = this.aquaYearEnd;
        int cmonth2 = this.aquaMonthEnd;
        int[] fmonthAqua = this.g2tAquaAnom.getFmonth();
        int[] fyearAqua = this.g2tAquaAnom.getFyear();
        int[] fmonthSnpp = this.g2tNppAnom.getFmonth();
        int[] fyearSnpp = this.g2tNppAnom.getFyear();
        
        double[][][] vanomAqua = this.g2tAquaAnom.getData();
        double[][][] vanomSnpp = this.g2tNppAnom.getData();
        
        // calculate climatology of anomaly difference
        double[][][] vclim = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        double[][][] pct = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        // find common period
        int indexAqua1=(cyear1 - fyearAqua[0])*12 + (cmonth1 - fmonthAqua[0]);
        int indexAqua2=(cyear2 - fyearAqua[0])*12 + (cmonth2 - fmonthAqua[0]);
        int indexSnpp1=(cyear1 - fyearSnpp[0])*12 + (cmonth1 - fmonthSnpp[0]);
        int indexSnpp2=(cyear2 - fyearSnpp[0])*12 + (cmonth2 - fmonthSnpp[0]);
        
        vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        
        int mt = indexSnpp2 - indexSnpp1 + 1;
        for(k=0; k<mt; k++) {
            indAqua = k + indexAqua1;
            indSnpp = k + indexSnpp1;
            kmonth = fmonthSnpp[indSnpp];
            km = kmonth-1;
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp1 = vanomAqua[indAqua][i][j];
                    vtmp2 = vanomSnpp[indSnpp][i][j];
                    if ( (vtmp1>=vmin)&(vtmp1<=vmax) & (vtmp2>=vmin) & (vtmp2 <= vmax)) {
                        vclim[km][i][j] = vclim[km][i][j]  +  (vtmp1 - vtmp2);
                        pct[km][i][j] = pct[km][i][j]  +  1.0d;                        
                    }
                }
            }            
        }
        
        for (k=0; k<12; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    if (pct[k][i][j]>0) {
                        vclim[k][i][j] = vclim[k][i][j]/pct[k][i][j];
                    } else {
                        vclim[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }    
        }
        
        mt = fyearAqua.length;
        for (k=0; k<mt; k++) {
            kmonth = fmonthAqua[k];
            km = kmonth - 1;
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp1 = vclim[km][i][j];
                    vtmp2 = vanomAqua[k][i][j];
                    if ( (vtmp1>=vmin)&(vtmp1<=vmax) & (vtmp2>=vmin) & (vtmp2 <= vmax)) {
                        vanomAqua[k][i][j] = vtmp2 - vtmp1; 
                    } else {
                        vanomAqua[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }
        
        this.g2tAquaAdj = new Geo2dTimeSeries(fyearAqua, fmonthAqua, vanomAqua);
        this.g2tAquaAdj.setDefaultValidLatIndex();
        this.g2tAquaAdj.setDeaultTemperatureAnomalyRange();
    }
    
    /**
     * Adjust N20 data to reference
     */
    private void adjustN20L4Data(){
        int i, j, k, km, kmonth;
        int indAqua, indSnpp;
        double vtmp1, vtmp2;
        double vmin, vmax;
        int cyear1 = this.n20YearStart;
        int cmonth1 = this.n20MonthStart;
        int cyear2 = this.rtutClimYear2;
        int cmonth2 = this.rtutClimMonth2;
        int[] fmonthN20 = this.g2tN20Anom.getFmonth();
        int[] fyearN20 = this.g2tN20Anom.getFyear();
        int[] fmonthSnpp = this.g2tNppAnom.getFmonth();
        int[] fyearSnpp = this.g2tNppAnom.getFyear();
        
        double[][][] vanomN20 = this.g2tN20Anom.getData();
        double[][][] vanomSnpp = this.g2tNppAnom.getData();
        
        // calculate climatology of anomaly difference
        double[][][] vclim = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        double[][][] pct = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        // find common period
        int indexN201=(cyear1 - fyearN20[0])*12 + (cmonth1 - fmonthN20[0]);
        int indexN202=(cyear2 - fyearN20[0])*12 + (cmonth2 - fmonthN20[0]);
        int indexSnpp1=(cyear1 - fyearSnpp[0])*12 + (cmonth1 - fmonthSnpp[0]);
        int indexSnpp2=(cyear2 - fyearSnpp[0])*12 + (cmonth2 - fmonthSnpp[0]);
        
        //System.out.println("Index_N20:" + indexN201 +","+indexN202+", snpp:"+ indexSnpp1 +"," + indexSnpp2);
        vmin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        vmax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
        
        int mt = indexSnpp2 - indexSnpp1 + 1;
        for(k=0; k<mt; k++) {
            indAqua = k + indexN201;
            indSnpp = k + indexSnpp1;
            kmonth = fmonthSnpp[indSnpp];
            km = kmonth-1;
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp1 = vanomN20[indAqua][i][j];
                    vtmp2 = vanomSnpp[indSnpp][i][j];
                    if ( (vtmp1>=vmin)&(vtmp1<=vmax) & (vtmp2>=vmin) & (vtmp2 <= vmax)) {
                        vclim[km][i][j] = vclim[km][i][j]  +  (vtmp1 - vtmp2);
                        pct[km][i][j] = pct[km][i][j]  +  1.0d;                        
                    }
                }
            }            
        }
        
        for (k=0; k<12; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    if (pct[k][i][j]>0) {
                        vclim[k][i][j] = vclim[k][i][j]/pct[k][i][j];
                    } else {
                        vclim[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }    
        }
        
        mt = fyearN20.length;
        for (k=0; k<mt; k++) {
            kmonth = fmonthN20[k];
            km = kmonth - 1;
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp1 = vclim[km][i][j];
                    vtmp2 = vanomN20[k][i][j];
                    if ( (vtmp1>=vmin)&(vtmp1<=vmax) & (vtmp2>=vmin) & (vtmp2 <= vmax)) {
                        vanomN20[k][i][j] = vtmp2 - vtmp1; 
                    } else {
                        vanomN20[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }
        
        this.g2tN20Adj = new Geo2dTimeSeries(fyearN20, fmonthN20, vanomN20);
        this.g2tN20Adj.setDefaultValidLatIndex();
        this.g2tN20Adj.setDeaultTemperatureAnomalyRange();        
    }
    

    /**
     * Load all L4 data 
     */
    private void loadAllL4Data(){
        this.loadAquaL4Data();
        this.loadSnppL4Data();
        this.loadN20L4Data();
        
        //this.testDataLoading();
    }
    
    /**
     * Load SNPP ATMS L4 data and use it as reference
     */
    private void loadSnppL4Data() {
        String filename = this.dirIn + "ATMS_L4_npp_ch08.nc";
        int cyear1 = this.nppClimYear1;
        int cmonth1 = this.nppClimMonth1;
        int cyear2 = this.nppClimYear2;
        int cmonth2 = this.nppClimMonth2;        
        this.g2tNppAnom  = this.loadAnomalyData(filename, cyear1, cmonth1, cyear2, cmonth2, true);       
        infoLogger.info("SNPP data loaded for RTUT merging");
    }    
    
    
    /**
     * Load N20 ATMS L4 data
     */
    private void loadN20L4Data() {
        String filename = this.dirIn + "ATMS_L4_j01_ch08.nc";        
        this.g2tN20Anom  = this.loadAnomalyData(filename);        
        infoLogger.info("N20 data loaded for RTUT merging");
    }    
    
    /**
     * Load Aqua L4 data for bias correction and merging
     */
    private void loadAquaL4Data(){
        int i,j,k;
        double vtmp1, vtmp2;
        String filename = this.dirIn + "AMSUA_L4_aqua_ch07.nc";
        this.g2tAquaAnom = this.loadAnomalyData(filename);
        infoLogger.info("Aqua data loaded for RTUT merging");
    }

    /**
     * Out put the RTMT data to a netcdf file
     */
    private void outputToFile(){
        int i, j, k;
        String filename = this.dirOut + getCDRFilename(this.yearEnd, this.monthEnd);
        infoLogger.info("Writing RTUT data to file " + filename);
        NetcdfFileWriter ncFileOut = null;
        Variable dataVar = null;
        ArrayInt.D1 iarr1 = null;
        ArrayDouble.D3 darr3 = null;
        ArrayDouble.D1 darr1 = null;
        try{
            ncFileOut = NetcdfFileWriter.createNew(NetcdfFileWriter.Version.netcdf3, filename);
            
            // create Dimensions
            Dimension dimTime = ncFileOut.addDimension(null, "time", this.numberMonths);
            Dimension dimLon = ncFileOut.addDimension(null, "longitude",GEO_GRID_NUM_LON);
            Dimension dimLat = ncFileOut.addDimension(null, "latitude", GEO_GRID_NUM_LAT);
            Dimension dimMonth = ncFileOut.addDimension(null, "month",12);

            List<Dimension> dimTimeD = new ArrayList<Dimension>();
            dimTimeD.add(dimTime);

            List<Dimension> dimLatD = new ArrayList<Dimension>();
            dimLatD.add(dimLat);

            List<Dimension> dimLonD = new ArrayList<Dimension>();
            dimLonD.add(dimLon);
            
            List<Dimension> dimMonD = new ArrayList<Dimension>();
            dimMonD.add(dimMonth);
            
            List<Dimension> dimGridD = new ArrayList<Dimension>();
            dimGridD.add(dimTime);
            dimGridD.add(dimLat);
            dimGridD.add(dimLon);

            List<Dimension> dimGridClim = new ArrayList<Dimension>();
            dimGridClim.add(dimMonth);
            dimGridClim.add(dimLat);
            dimGridClim.add(dimLon);

            // define corridinates variable variables
            Variable vlon = ncFileOut.addVariable(null, "longitude", DataType.FLOAT, dimLonD);
            Variable vlat = ncFileOut.addVariable(null, "latitude", DataType.FLOAT, dimLatD);
            Variable vmon = ncFileOut.addVariable(null, "month", DataType.INT, dimTimeD);
            Variable vyear = ncFileOut.addVariable(null, "fyear", DataType.INT, dimTimeD);
            Variable vmonth = ncFileOut.addVariable(null, "fmonth", DataType.INT, dimTimeD);
            Variable vft = ncFileOut.addVariable(null, "ft", DataType.DOUBLE, dimTimeD);
            vlat.addAttribute(new Attribute("units", "degree"));
            vlon.addAttribute(new Attribute("units", "degree"));
            
            // define data variables and attributes
            String varName = MltCdrConstant.VARNAME_BRIGHTNESS_TEMPERATURE_RTUT;
            Variable varBt= ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridD);
            varBt.addAttribute(new Attribute("units","Kelvin" ));
            varBt.addAttribute(new Attribute("valid_max", GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE));
            varBt.addAttribute(new Attribute("valid_min", GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE));
            varBt.addAttribute(new Attribute("_FillValue", GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varBt.addAttribute(new Attribute("coordinates","time latitude longitude" ));

            varName = MltCdrConstant.VARNAME_BRIGHTNESS_TEMPERATURE_RTUT_ANOMALY;
            Variable varAnom = ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridD);
            varAnom.addAttribute(new Attribute("units","Kelvin" ));
            varAnom.addAttribute(new Attribute("valid_max", GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE));
            varAnom.addAttribute(new Attribute("valid_min", GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE));
            varAnom.addAttribute(new Attribute("_FillValue", GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varAnom.addAttribute(new Attribute("coordinates","time latitude longitude" ));

            // define data variables and attributes
            varName = MltCdrConstant.VARNAME_BRIGHTNESS_TEMPERATURE_RTUT_CLIMATOLOGY;
            Variable varClim= ncFileOut.addVariable(null, varName , DataType.DOUBLE, dimGridClim);
            varClim.addAttribute(new Attribute("units","Kelvin" ));
            varClim.addAttribute(new Attribute("valid_max", GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE));
            varClim.addAttribute(new Attribute("valid_min", GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE));
            varClim.addAttribute(new Attribute("_FillValue", GeoGridConstant.GEO_INVALID_FILL_VALUE));          
            varClim.addAttribute(new Attribute("coordinates","month latitude longitude" ));
            
            ncFileOut.create();
            // write data
            ArrayFloat.D1 dataLat = new ArrayFloat.D1(GEO_GRID_NUM_LAT);
            ArrayFloat.D1 dataLon = new ArrayFloat.D1(GEO_GRID_NUM_LON);      
            ArrayInt.D1 dataCmonth = new ArrayInt.D1(12);
            ArrayInt.D1 dataFyear = new ArrayInt.D1(this.numberMonths);
            ArrayInt.D1 dataFmonth = new ArrayInt.D1(this.numberMonths);
            ArrayDouble.D1 dataFt = new ArrayDouble.D1(this.numberMonths);
            // write latitude   
            float[] plat = GeoGrid2D.getGridLatitudeFloat();
            for (i=0; i<GEO_GRID_NUM_LAT; i++) {
                dataLat.set(i, plat[i]);
            }
            ncFileOut.write(vlat, dataLat);
            
            // write longitude
            float[] plon = GeoGrid2D.getGridLongitudeFloat();
            for(i=0; i<GEO_GRID_NUM_LON; i++){
                dataLon.set(i, plon[i]);
            }
            ncFileOut.write(vlon, dataLon);
            
            // write year 
            for(i=0; i<this.numberMonths; i++) {
                dataFyear.set(i, this.fyear[i]);
            }
            ncFileOut.write(vyear, dataFyear);
            
            // write fmonth
            for(i=0; i<this.numberMonths; i++) {
                dataFmonth.set(i, fmonth[i]);
            }
            ncFileOut.write(vmonth, dataFmonth);

            
            // write time
            for(i=0; i<this.numberMonths; i++) {
                dataFt.set(i, ft[i]);
            }
            ncFileOut.write(vft, dataFt);
            
            // write month
            for(i=0; i<12; i++) {
                dataCmonth.set(i, i+1);
            }
            ncFileOut.write(vmon, dataCmonth);

            // write RTMT
            ArrayDouble.D3 vdata = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<this.numberMonths; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.Tb[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varBt, vdata);

            // write Anomaly
            ArrayDouble.D3 vdata2 = new ArrayDouble.D3(this.numberMonths, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<this.numberMonths; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata2.set(i,j,k, this.Tanom[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varAnom, vdata2);

            // write climatology
            vdata = new ArrayDouble.D3(12, GEO_GRID_NUM_LAT, GEO_GRID_NUM_LON);
            for (i=0; i<12; i++) {
                for(j=0; j<GEO_GRID_NUM_LAT; j++) {
                    for(k=0; k<GEO_GRID_NUM_LON; k++) {
                        vdata.set(i,j,k, this.Tclim[i][j][k]);
                    }
                }
            }
            ncFileOut.write(varClim, vdata);
            
        }catch(Exception e) {
            e.printStackTrace();
            errLogger.error("Failed to write RTMT data to netcdf file " + filename);
        }

        if (ncFileOut != null) {
            try {
                    ncFileOut.close();
                } catch (Exception exp) {
                    errLogger.error("Failed to close netcdf file" + filename);
                    exp.printStackTrace();
                }
        }

    }

 
    /**
     * Get short name of the RTUT file
     */
    public static String  getCDRFilename(int year2, int month2){
        String filename = String.format("NESDIS-STAR_RTUT_Anomaly_AMSUA_ATMS_V01_S200208_E%4d%02d.nc", year2, month2);        
        return filename;                
    }
    
    /**
     * Check if data are loaded correctly
     */
    private void testDataLoading(){
        double[] y = this.g2tAquaAnom.getGlobalMean();
        int n = y.length;
        for(int k=0; k<n; k++){
            System.out.println("Aqua mean:" + y[k]);
        }
        
        y = this.g2tNppAnom.getGlobalMean();
        n = y.length;
        for(int k=0; k<n; k++){
            System.out.println("SNPP mean:" + y[k]);
        }
        
        y = this.g2tN20Anom.getGlobalMean();
        n = y.length;
        for(int k=0; k<n; k++){
            System.out.println("N20 mean:" + y[k]);
        }
        
    }
    
    
    /**
     *  For testing only
     * @param args 
     */
    public static void main(String[] args){
        int year2 =  2023;
        int month2 = 12;
        String dirLut = "/home/xhao/data/mltcdrv5/lut/ref/";
        String dirL4 = "/home/xhao/data/mltcdrv5/data/L4/";
        String dirRefCDR = "/home/xhao/data/mltcdrv5/data/rfcdr/";
        RefCdrTaskConfig taskConf = new RefCdrTaskConfig(year2, month2,dirLut,dirL4, dirRefCDR);
        ReferenceTUT rtut = new ReferenceTUT(taskConf);
        rtut.processCDR();
    }    
}
