/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.geogrid.Geo2dTimeSeries.java
*
* NAME
*   atms.geogrid.Geo2dTimeSeries.java
*
* PURPOSE
*   To build time series for general processing
* 
* DESCRIPTION
*	This class is used to process time series of geo-gridded data
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package geogrid;

import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;
import static geogrid.GeoGridConstant.GEO_INVALID_FILL_VALUE;

import geogrid.GeoGridConstant;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

// For testing only, import NetCDF API
import ucar.ma2.ArrayInt;
import ucar.ma2.ArrayFloat;
import ucar.ma2.ArrayDouble;
import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;
   
/**
 * Time series of Geogrid2D data
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class Geo2dTimeSeries {

    /**
     * Logger for errors
     */
    final static private  Logger  errLogger = LogManager.getLogger("ErrorLog");
    
    /**
     * Logger for information
     */
    final static private  Logger  infoLogger = LogManager.getLogger("InfoLog");
    
    /**
     * List of years for data
     */
    private int[] fyear;

    /**
     * List of months for data
     */
    private int[] fmonth;
    
    /**
     * List of time for data
     */
    private double[] ft;
    
    /**
     * start year
     */
    private int yearStart;
    
    /**
     * start month
     */
    private int monthStart;

    /**
     * End year
     */
    private int yearEnd;
    
    /**
     * End month
     */
    private int monthEnd;
    
    /**
     * Number of months
     */
    private int numerMonths;
    
    /**
     * data
     */
    private double[][][] data = null;
    
    /**
     * Land/sea mask
     */
    private int[][] landseaMask = null;
    
    /**
     * Flag for mask
     */
    private boolean isMaskReady = false;
    
    /**
     * Valid min
     */
    private double validMin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE; // default valid minimum

    /*
     * Valid max
    */
    private double validMax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE; // default valid maximum 
    
    /**
     * Filled value for invalid data
     */
    private double invalidFill = GEO_INVALID_FILL_VALUE; // default filled value for invalid data
    
    /*
     * Starting index for global mean
    */
    private int index_valid_lat_start = 0;
    
    /**
     * Ending index for global mean
     */
    private int index_valid_lat_end = GEO_GRID_NUM_LAT-1;
        
    
    /**
     * Constructor.
     * 
     * @param fyear, year, month, data
     * @param fmonth
     * @param vdata 
     */
    public Geo2dTimeSeries(int[] fyear, int[] fmonth, double[][][] vdata) {
        
        int i, j , k;
        this.numerMonths = fyear.length;        
        this.fyear = new int[numerMonths];
        this.fmonth = new int[numerMonths];
        this.ft = new double[numerMonths];
        this.yearStart = fyear[0];
        this.monthStart = fmonth[0];
        this.yearEnd = fyear[numerMonths-1];
        this.monthEnd = fmonth[numerMonths-1];
        
        this.data = new double[numerMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        for(i=0; i<numerMonths; i++){
            this.fyear[i] = fyear[i];
            this.fmonth[i] = fmonth[i];
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
            for(j=0; j<GEO_GRID_NUM_LAT; j++){
                for(k=0; k<GEO_GRID_NUM_LON; k++) {
                    this.data[i][j][k] = vdata[i][j][k];
                }
            }
        }        
        
    }

    
    

    /**
     * Constructor
     * @param fyear, year list
     * @param fmonth, month list 
     */
    public Geo2dTimeSeries(int[] fyear, int[] fmonth) {
        int i;
        this.numerMonths = fyear.length;        
        this.fyear = new int[numerMonths];
        this.fmonth = new int[numerMonths];
        this.ft = new double[numerMonths];
        this.yearStart = fyear[0];
        this.monthStart = fmonth[0];
        this.yearEnd = fyear[numerMonths-1];
        this.monthEnd = fmonth[numerMonths-1];

        this.data = new double[numerMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        for(i=0; i<numerMonths; i++){
            this.fyear[i] = fyear[i];
            this.fmonth[i] = fmonth[i];
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
        this.fillInvalid();
    }
 
    /**
     * Constructor
     * @param year1,  start year
     * @param month1, start month; 
     * @param year2,  end year
     * @param month2, end month; 
     */
    public Geo2dTimeSeries(int year1, int month1, int year2, int month2) {
        int i;
        this.yearStart = year1;
        this.monthStart = month1;
        this.yearEnd = year2;
        this.monthEnd = month2;
        this.numerMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numerMonths];
        this.fmonth = new int[numerMonths];
        this.ft = new double[numerMonths];

        this.data = new double[numerMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        for(i=0; i<numerMonths; i++){
            this.fyear[i] = fyear[i];
            this.fmonth[i] = fmonth[i];
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
        this.fillInvalid();
    }

    /**
     * Constructor
     * @param year1,  start year
     * @param month1, start month 
     * @param year2,  end year
     * @param month2, end month
     * @param vdata, data
     */
    public Geo2dTimeSeries(int year1, int month1, int year2, int month2, double[][][] vdata) {
        int i, j , k;
        int  km, kyear, kmonth;
        this.yearStart = year1;
        this.monthStart = month1;
        this.yearEnd = year2;
        this.monthEnd = month2;
        this.numerMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numerMonths];
        this.fmonth = new int[numerMonths];
        this.ft = new double[numerMonths];

        this.data = new double[numerMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
                
        for(i=0; i<numerMonths; i++){
            km = i + month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
            for(j=0; j<GEO_GRID_NUM_LAT; j++){
                for(k=0; k<GEO_GRID_NUM_LON; k++) {
                    this.data[i][j][k] = vdata[i][j][k];
                }
            }
        }        
        
    }
    

    /**
     * Constructor
     * @param year1,  start year
     * @param month1, start month 
     * @param year2,  end year
     * @param month2, end month
     * @param vdata, data
     */
    public Geo2dTimeSeries(int year1, int month1, int year2, int month2, float[][][] vdata) {
        int i, j , k;
        int  km, kyear, kmonth;
        this.yearStart = year1;
        this.monthStart = month1;
        this.yearEnd = year2;
        this.monthEnd = month2;
        this.numerMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numerMonths];
        this.fmonth = new int[numerMonths];
        this.ft = new double[numerMonths];

        this.data = new double[numerMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
                
        for(i=0; i<numerMonths; i++){
            km = i + month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
            for(j=0; j<GEO_GRID_NUM_LAT; j++){
                for(k=0; k<GEO_GRID_NUM_LON; k++) {
                    this.data[i][j][k] = vdata[i][j][k];
                }
            }
        }        
        
    }
    
    
    /**
     * Create Geo2dTimeSeries with L4 data file
     * @param fileL4Netcdf
     * @param varname 
     */
    public Geo2dTimeSeries(String fileL4Netcdf, String varName) {
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayInt.D1 iarr1 = null;
        ArrayDouble.D3 darr3 = null;
        ArrayFloat.D3 farr3 = null;
        int[] pyear = null;
        int[] pmonth = null;
        int[] shape = new int[3];
        int[] origin = new int[3];
        int[] shape1 = new int[1];
        int[] origin1 = new int[1];
        int dim1, dim2, dim3;
        int i, j, k;
        int mt = 0;
        int km=0;
        boolean isDouble = true;

        origin1[0] = 0;
        for (i = 0; i < 3; i++) {
            origin[i] = 0;
        }

        try {
            fileNc = NetcdfFile.open(fileL4Netcdf);
            
            // read year
            dataVar = fileNc.findVariable("fyear");
            if (dataVar == null) {
                errLogger.error("failed to find variable fyear from file " + fileL4Netcdf);
            } else {
                shape1 = dataVar.getShape();
                iarr1 = (ArrayInt.D1) dataVar.read(origin1, shape1);
                mt = shape1[0];
                pyear = new int[mt];
                for(i=0; i<mt;i++){
                    pyear[i] = iarr1.get(i);
                }
            }
            // read month
            dataVar = fileNc.findVariable("fmonth");
            if (dataVar == null) {
                errLogger.error("failed to find variable fmonth from file " + fileL4Netcdf);
            } else {
                shape1 = dataVar.getShape();
                mt = shape1[0];
                pmonth = new int[mt];
                iarr1 = (ArrayInt.D1) dataVar.read(origin1, shape1);
                for(i=0; i<mt;i++){
                    pmonth[i] = iarr1.get(i);
                }
            }

            this.yearStart = pyear[0];
            this.yearEnd = pyear[mt-1];
            this.monthStart = pmonth[0];            
            this.monthEnd =pmonth[mt-1];
            this.initData();
            
            dataVar = fileNc.findVariable(varName);
            if (dataVar == null) {
                errLogger.error("failed to find variable from file " + fileL4Netcdf);
                this.data = null;
                this.numerMonths = 0;
            } else {
                shape = dataVar.getShape();
                try {
                    darr3 = (ArrayDouble.D3) dataVar.read(origin, shape);
                    isDouble = true;
                } catch (Exception exp) {
                    farr3 = (ArrayFloat.D3) dataVar.read(origin, shape);                    
                    isDouble = false;
                }
                mt = shape[0];
                              
                if (isDouble) {
                    for (k = 0; k < mt; k++) {
                        km = (pyear[k] - this.yearStart) * 12 + (pmonth[k] - this.monthStart);
                        for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                            for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                                this.data[km][i][j] = darr3.get(k, i, j);
                            }
                        }
                    }
                } else {
                    for (k = 0; k < mt; k++) {
                        km = (pyear[k] - this.yearStart) * 12 + (pmonth[k] - this.monthStart);
                        for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                            for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                                this.data[km][i][j] = farr3.get(k, i, j);
                            }
                        }
                    }
                }
            }            
        } catch (Exception e) {
            e.printStackTrace();
            errLogger.error("failed to read " + fileL4Netcdf + ", filled with invalid values");
            this.data = null;
            this.numerMonths = 0;
        }
                   
    }
    

        /**
     * Create Geo2dTimeSeries with L4 data file
     * @param fileL4Netcdf
     * @param varname 
     * @param option, time is saved as day since 
     */
    public Geo2dTimeSeries(String fileL4Netcdf, String varName, int option) {
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayInt.D1 iarr1 = null;
        ArrayDouble.D3 darr3 = null;
        ArrayFloat.D3 farr3 = null;
        int[] pyear = null;
        int[] pmonth = null;
        int[] shape = new int[3];
        int[] origin = new int[3];
        int[] shape1 = new int[1];
        int[] origin1 = new int[1];
        int dim1, dim2, dim3;
        int i, j, k;
        int mt = 0;
        int km=0;
        boolean isDouble = true;

        origin1[0] = 0;
        for (i = 0; i < 3; i++) {
            origin[i] = 0;
        }

        try {
            fileNc = NetcdfFile.open(fileL4Netcdf);
            
            // read year
            dataVar = fileNc.findVariable("fyear");
            if (dataVar == null) {
                errLogger.error("failed to find variable fyear from file " + fileL4Netcdf);
            } else {
                shape1 = dataVar.getShape();
                iarr1 = (ArrayInt.D1) dataVar.read(origin1, shape1);
                mt = shape1[0];
                pyear = new int[mt];
                for(i=0; i<mt;i++){
                    pyear[i] = iarr1.get(i);
                }
            }
            // read month
            dataVar = fileNc.findVariable("fmonth");
            if (dataVar == null) {
                errLogger.error("failed to find variable fmonth from file " + fileL4Netcdf);
            } else {
                shape1 = dataVar.getShape();
                mt = shape1[0];
                pmonth = new int[mt];
                iarr1 = (ArrayInt.D1) dataVar.read(origin1, shape1);
                for(i=0; i<mt;i++){
                    pmonth[i] = iarr1.get(i);
                }
            }

            this.yearStart = pyear[0];
            this.yearEnd = pyear[mt-1];
            this.monthStart = pmonth[0];            
            this.monthEnd =pmonth[mt-1];
            this.initData();
            
            dataVar = fileNc.findVariable(varName);
            if (dataVar == null) {
//                System.out.println("failed to find variable BT_mean from file " + filenameL3);
                System.out.println("failed to find variable from file " + fileL4Netcdf);
                errLogger.error("failed to find variable from file " + fileL4Netcdf);
                this.data = null;
                this.numerMonths = 0;
            } else {
                shape = dataVar.getShape();
                try {
                    darr3 = (ArrayDouble.D3) dataVar.read(origin, shape);
                    isDouble = true;
                } catch (Exception exp) {
                    farr3 = (ArrayFloat.D3) dataVar.read(origin, shape);                    
                    isDouble = false;
                }
                //System.out.println("Data dimension:" + shape[0] + "," + shape[1]+ "," + shape[2]);
                //System.out.println("Data dimension:" + darr3.shapeToString());
                mt = shape[0];
                              
                if (isDouble) {
                    for (k = 0; k < mt; k++) {
                        km = (pyear[k] - this.yearStart) * 12 + (pmonth[k] - this.monthStart);
                      /* for debug only  
                        if (mt != this.numerMonths) {
                            String msg = String.format("checking, k=%d, %4d, %02d: km=%d", k, pyear[k], pmonth[k], km);
                            System.out.println(msg);
                        }
                      */  
                        for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                            for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                                this.data[km][i][j] = darr3.get(k, i, j);
                            }
                        }
                    }
                } else {
                    for (k = 0; k < mt; k++) {
                        km = (pyear[k] - this.yearStart) * 12 + (pmonth[k] - this.monthStart);
                        for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                            for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                                this.data[km][i][j] = farr3.get(k, i, j);
                            }
                        }
                    }
                }
            }            
        } catch (Exception e) {
            e.printStackTrace();
            errLogger.error("failed to read " + fileL4Netcdf + ", filled with invalid values");
            this.data = null;
            this.numerMonths = 0;
        }
                   
    }

    /**
     * Initialize time 
     */
    private void initData(){
        int i, j , k;
        int  km, kyear, kmonth;
        int year1 = this.yearStart;
        int month1 = this.monthStart  ;
        int year2 = this.yearEnd;
        int month2 = this.monthEnd;
        this.numerMonths = (year2 - year1)*12 + (month2-month1+1);        
        this.fyear = new int[numerMonths];
        this.fmonth = new int[numerMonths];
        this.ft = new double[numerMonths];

        this.data = new double[numerMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
                
        for(i=0; i<numerMonths; i++){
            km = i + month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
        this.fillInvalid();
    }
    

    /**
     * Fill data array with invalid values
     */
    private void fillInvalid(){
        for(int i=0; i<this.numerMonths; i++) {
            for(int j=0; j<GEO_GRID_NUM_LAT; j++) {
                for(int k=0; k<GEO_GRID_NUM_LON; k++) {
                    this.data[i][j][k] = GEO_INVALID_FILL_VALUE;
                }
            }
        }
    }
    
    /**
     *  Quality control, remove invalid data of a specified month
     * @param qyear, year of the data to be removed
     * @param qmonth , month of the data to be removed
     */
    public void qaControl(int qyear, int qmonth){
        int km = (qyear - this.yearStart)*12 + (qmonth - this.monthStart);
        if ( (km>=0)&(km<this.numerMonths) ){
            for(int i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(int j=0; j<GEO_GRID_NUM_LON; j++) {
                    this.data[km][i][j] = GEO_INVALID_FILL_VALUE;
                }
            }        
        }
    }

    /**
     *  Quality control, remove invalid data for specified months
     * @param qyear, year of the data to be removed
     * @param qmonth , month of the data to be removed
     */
    public void qaControl(int[] qyears, int[] qmonths){
        if (qyears != null) {
            int nqa = qyears.length;
            for (int i = 0; i < nqa; i++) {
                this.qaControl(qyears[i], qmonths[i]);
            }
        }
    }
    
    /**
     * Get length of the time series
     * @return length of the time series 
     */
    public int getTimeSeriesLength(){
        return this.numerMonths;
    }
    
    /**
     * Get time of the series
     * @return time of the series
     */
    public double[] getFt(){
        return this.ft;
    }
    
    /**
     * Get years of the time series
     * @return years of the time series
     */
    public int[] getFyear(){
       return this.fyear; 
    }
    
    /**
     * Get months of the time series
     * @return months of the time series
     */
    public int[] getFmonth(){
        return this.fmonth;
    }
    
    
    /**
     * 
     * Generate year list for specified period 
     * @param year1, start year
     * @param month1, start month
     * @param year2, end year
     * @param month2, end month
     * @return year list
     */
    public static int[] getFyear(int year1, int month1, int year2, int month2) {
        int i, km, kmonth, kyear;
        int[] tyear = null;
        int mt = (year2 - year1)*12 + (month2 - month1 +1);
        
        if (mt>0){
            tyear = new int[mt];
        }
            for (i = 0; i < mt; i++) {
                km = i + month1;
                kmonth = km % 12;
                if (kmonth == 0) {
                    kyear = year1 + (km - 1) / 12;
                    kmonth = 12;
                } else {
                    kyear = year1 + km / 12;
                }
                tyear[i] = kyear;            
        } 
        
        return tyear;
    }
    
    /**
     * 
     * Generate month list for specified period 
     * @param year1, start year
     * @param month1, start month
     * @param year2, end year
     * @param month2, end month
     * @return month list
     */
    public static int[] getFmonth(int year1, int month1, int year2, int month2) {
        int i, km, kmonth, kyear;
        int[] tmonth = null;
        int mt = (year2 - year1)*12 + (month2 - month1 +1);
        
        if (mt>0){
            tmonth = new int[mt];
        }
            for (i = 0; i < mt; i++) {
                km = i + month1;
                kmonth = km % 12;
                if (kmonth == 0) {
                    kyear = year1 + (km - 1) / 12;
                    kmonth = 12;
                } else {
                    kyear = year1 + km / 12;
                }
                tmonth[i] = kmonth;            
        } 
        
        return tmonth;
    }

    /**
     * 
     * Generate year list for specified period 
     * @param year1, start year
     * @param month1, start month
     * @param year2, end year
     * @param month2, end month
     * @return year list
     */
    public static double[] getFt(int year1, int month1, int year2, int month2) {
        int i, km, kmonth, kyear;
        double[] tm = null;
        int mt = (year2 - year1)*12 + (month2 - month1 +1);
        
        if (mt>0){
            tm = new double[mt];
        }
            for (i = 0; i < mt; i++) {
                km = i + month1;
                kmonth = km % 12;
                if (kmonth == 0) {
                    kyear = year1 + (km - 1) / 12;
                    kmonth = 12;
                } else {
                    kyear = year1 + km / 12;
                }
                tm[i] = (double)kyear + ((double)kmonth - 0.5)/12.0d;            
        } 
        
        return tm;
    }
    
    /**
     * Get data time series
     * @return data time series
     */
    public double[][][] getData(){
        return this.data;
    }
    
    /**
     * Get global area-weighted mean of the data 
     * @return 
     */
    public double[] getGlobalMean(){
        int i, j, k;
        
        double[] ym = new double[this.numerMonths];
        double[][] vtmp = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        GeoGrid2D geo2d = new GeoGrid2D();
        geo2d.setValidRange(this.validMin, this.validMax);
        geo2d.setValidLatIndex(this.index_valid_lat_start, this.index_valid_lat_end);
        for(k=0; k<this.numerMonths; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp[i][j] = this.data[k][i][j];
                }
            }
            geo2d.setData(vtmp);
            ym[k] = geo2d.getGlobalMean();
        }        
        return ym;
    }

    /**
     * Get global area-weighted mean of the specified data
     * @return global mean time series
     */
    public static double[] getGlobalMean(double[][][] vdata, double vmin, double vmax){
        int i, j, k;
        int mt = vdata.length;
        double[] ym = new double[mt];
        double[][] vtmp = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        GeoGrid2D geo2d = new GeoGrid2D();
        geo2d.setValidRange(vmin, vmax);
        geo2d.setValidLatIndex(GeoGridConstant.GEO_INDEX_GLOBAL_MEAN_LATITUDE_START, GeoGridConstant.GEO_INDEX_GLOBAL_MEAN_LATITUDE_END);

        for(k=0; k<mt; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp[i][j] = vdata[k][i][j];
                }
            }
            geo2d.setData(vtmp);
            ym[k] = geo2d.getGlobalMean();
        }        
        return ym;
    }
    
    

    /**
     * Get climatology with specified normal period from cmonth1/cyear1 to cmonth2/cyear2
     * @param cyear1, start year of the normal period
     * @param cmonth1, start month of the normal period
     * @param cyear2, end year of the normal period 
     * @param cmonth2, end month of the normal period
     * @return climatology data
     */
    public double[][][] getClimatology(int cyear1, int cmonth1, int cyear2, int cmonth2) {
        int i, j, k, km, kmonth;
        double ct1 = (double)cyear1 + ((double)cmonth1-1.0d)/12.0d;
        double ct2 = (double)cyear2 + ((double)cmonth2)/12.0d;

        double[][][] vclim = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        double[][][] pct = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        for(k=0; k<12; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vclim[k][i][j] = 0.0d;
                    pct[k][i][j] = 0.0d;
                }
            }
        }
        
        for(km=0; km<12; km++) {
            kmonth = km + 1; // month;
            for (k=0; k<this.numerMonths; k++) {
                if ( (this.fmonth[k] == kmonth) & (this.ft[k]>ct1) & (this.ft[k] < ct2)) {
                    for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                        for(j=0; j<GEO_GRID_NUM_LON; j++) {
                            if ((data[k][i][j] >= this.validMin) & ((data[k][i][j] <= this.validMax))) {
                               vclim[km][i][j] = vclim[km][i][j] + this.data[k][i][j];
                               pct[km][i][j] = pct[km][i][j] + 1.0d;
                            }
                        }
                    }
                }
            }
        }
        
        for (k = 0; k < 12; k++) {
            for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                    if (pct[k][i][j]>0) {
                        vclim[k][i][j] = vclim[k][i][j]/pct[k][i][j];
                    } else {
                        vclim[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }
        
        return vclim;        
    }
    
    /**
     * Get climatology for whole time period
     * @return  climatology
     */
    public double[][][] getClimatology() {
        int cyear1 = this.yearStart;
        int cmonth1 = this.monthStart;
        int cyear2 = this.yearEnd;
        int cmonth2 = this.monthEnd;
        double[][][] vclim = this.getClimatology(cyear1, cmonth1, cyear2, cmonth2);
        return vclim;
    }


    /**
     * Calculate climatology of  data for a specified normal period
     * @param vdata, data time series with a dimension of  (nt, 72,144).
     * @param tyear, year for the data time series
     * @param tmonth, month for the data time series
     * @param ct1, start time of the normal period
     * @param ct2, end time of the normal period
     * @param vmin, minimum value of valid data
     * @param vmax, maximum value of valid data
     * @return, climatology 
     */
    public static double[][][] getClimatology(double[][][]vdata,  int[] tyear, int[] tmonth, double ct1, double ct2, double vmin, double vmax) {
        int i, j, k, km, kmonth;
        int mt = tyear.length;
        double[] tm = new double[mt];
        for(k=0; k<mt; k++){
            tm[k] = (double)tyear[k] + ((double)tmonth[k] - 0.5d)/12.0d;
        }
        
        double[][][] vclim = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        double[][][] pct = new double[12][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        
        for(k=0; k<12; k++) {
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vclim[k][i][j] = 0.0d;
                    pct[k][i][j] = 0.0d;
                }
            }
        }
        
        for(km=0; km<12; km++) {
            kmonth = km + 1; // month;
            for (k=0; k<mt; k++) {
                if ( (tmonth[k] == kmonth) & (tm[k]>ct1) & (tm[k] < ct2)) {
                    for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                        for(j=0; j<GEO_GRID_NUM_LON; j++) {
                            if ((vdata[k][i][j] >= vmin) & ((vdata[k][i][j] <= vmax))) {
                               vclim[km][i][j] = vclim[km][i][j] + vdata[k][i][j];
                               pct[km][i][j] = pct[km][i][j] + 1.0d;
                            }
                        }
                    }
                }
            }
        }
        
        for (k = 0; k < 12; k++) {
            for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                    if (pct[k][i][j]>0) {
                        vclim[k][i][j] = vclim[k][i][j]/pct[k][i][j];
                    } else {
                        vclim[k][i][j] = GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }
        
        return vclim;
    }
    
    /**
     * Get anomaly with specified normal period from cmonth1/cyear1 to cmonth2/cyear2
     * @param cyear1, start year of the normal period
     * @param cmonth1, start month of the normal period
     * @param cyear2, end year of the normal period 
     * @param cmonth2, end month of the normal period
     * @return anomaly data time series
     */
    public double[][][] getAnomaly(int cyear1, int cmonth1, int cyear2, int cmonth2) {
        int i, j, k, km;
        double vtmp1, vtmp2;
        double[][][] anom = new double[this.numerMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        double[][][] vclim = getClimatology(cyear1, cmonth1, cyear2, cmonth2);
        
        for(k=0; k<this.numerMonths; k++) {
            km = this.fmonth[k] - 1 ; // index for month
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp1 = this.data[k][i][j];
                    vtmp2 =  vclim[km][i][j];
                    if ( (vtmp1>=this.validMin) & (vtmp2>=this.validMin) &(vtmp1<=this.validMax) & (vtmp2<=this.validMax)) {
                        anom[k][i][j] = this.data[k][i][j] - vclim[km][i][j]; 
                    } else {
                        anom[k][i][j] = this.invalidFill;
                    }
                }
            }
        }        
        return anom;
    }

    /**
     * Get anomaly using the the whole period as normal 
     * @return 
     */
    public double[][][] getAnomaly(){
        int cyear1 = this.yearStart;
        int cmonth1 = this.monthStart;
        int cyear2 = this.yearEnd;
        int cmonth2 = this.monthEnd;
        double[][][] vclim = this.getAnomaly(cyear1, cmonth1, cyear2, cmonth2);
        return vclim;
    }
    
    /**
     * Get global land area-weighted mean of the data
     *
     * @return
     */
    public double[] getGlobalLandMean() {
        int i, j, k;

        double[] ym = new double[this.numerMonths];
        double[][] vtmp = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];

        GeoGrid2D geo2d = new GeoGrid2D();
        geo2d.setValidRange(this.validMin, this.validMax);
        geo2d.setValidLatIndex(this.index_valid_lat_start, this.index_valid_lat_end);

        if (this.isMaskReady) {
            geo2d.setLandseaMask(this.landseaMask);
            for (k = 0; k < this.numerMonths; k++) {
                for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                    for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                        vtmp[i][j] = this.data[k][i][j];
                    }
                }
                geo2d.setData(vtmp);
                ym[k] = geo2d.getGlobalLandMean();
            }
        } else {
            for (k = 0; k < this.numerMonths; k++) {
                ym[k] = this.invalidFill;
            }                
        }
        return ym;
    }

    /**
     * Setup valid index for global mean
     *
     * @param index1, starting index
     * @param index2, ending index
     */
    public void setValidLatIndex(int index1, int index2) {
        this.index_valid_lat_start = index1;
        this.index_valid_lat_end = index2;
    }

    /**
     * Setup valid index for global mean to default values
     */
    public void setDefaultValidLatIndex(){
        this.index_valid_lat_start = GeoGridConstant.GEO_INDEX_GLOBAL_MEAN_LATITUDE_START;
        this.index_valid_lat_end = GeoGridConstant.GEO_INDEX_GLOBAL_MEAN_LATITUDE_END;        
    }
    
    /**
     * Get global area-weighted mean of the data
     *
     * @return
     */
    public double[][] getGlobalMeanAll() {
        int i, j, k;

        double[] vm = new double[3];

        double[][] ym = new double[3][this.numerMonths];
        double[][] vtmp = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];

        GeoGrid2D geo2d = new GeoGrid2D();
        geo2d.setValidRange(this.validMin, this.validMax);
        geo2d.setValidLatIndex(this.index_valid_lat_start, this.index_valid_lat_end);

        geo2d.setLandseaMask(this.landseaMask);
        for (k = 0; k < this.numerMonths; k++) {
            for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                    vtmp[i][j] = this.data[k][i][j];
                }
            }
            geo2d.setData(vtmp);
            if (this.isMaskReady) {
                vm = geo2d.getGlobalMeanAll();
                for(j=0; j<3; j++) {
                    ym[j][k] = vm[j];
                }
            } else {
                ym[0][k] = this.invalidFill;
                ym[1][k] = this.invalidFill;
                ym[2][k] = geo2d.getGlobalMean();                
            }
        }
        return ym;
    }
    
    /**
     * Get global ocean area-weighted mean of the data
     *
     * @return
     */
    public double[] getGlobalOceanMean() {
        int i, j, k;

        double[] ym = new double[this.numerMonths];
        double[][] vtmp = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];

        GeoGrid2D geo2d = new GeoGrid2D();
        geo2d.setValidRange(this.validMin, this.validMax);
        geo2d.setValidLatIndex(this.index_valid_lat_start, this.index_valid_lat_end);

        if (this.isMaskReady) {
            geo2d.setLandseaMask(this.landseaMask);
            for (k = 0; k < this.numerMonths; k++) {
                for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                    for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                        vtmp[i][j] = this.data[k][i][j];
                    }
                }
                geo2d.setData(vtmp);
                ym[k] = geo2d.getGlobalOceanMean();
            }
        } else {
            for (k = 0; k < this.numerMonths; k++) {
                ym[k] = this.invalidFill;
            }                
        }
        return ym;
    }
    
    
    /**
     * Set land/sea mask for geo-grid
     *
     * @param mask
     */
    public void setLandseaMask(int[][] mask) {
        this.landseaMask = new int[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            for (int j = 0; j < GEO_GRID_NUM_LON; j++) {
                this.landseaMask[i][j] = mask[i][j];
            }
        }
        this.isMaskReady = true;
    }
    
    /**
     * Set land/sea mask for geo-grid with land fraction data file
     * @param landfracFilename 
     */
    public void setLandseaMask(String landfracFilename) {
        this.landseaMask = GeoGridConstant.getLandSeaMask(landfracFilename);
        this.isMaskReady = true;
    }

    /**
     * Set valid range
     *
     * @param vmin, valid minimum value
     * @param vmax, valid maximum value
     */
    public void setValidRange(double vmin, double vmax) {
        this.validMin = vmin;
        this.validMax = vmax;
    }
    
    /**
     * Get valid minimum value
     * @return the valid minimum value
     */
    public double getValidMin(){
        return this.validMin;
    }
    
    /**
     * Get valid maximum value
     * @return the valid maximum value
     */
    public double getValidMax(){
        return this.validMax;
    }
    
    
    
    /**
     * Set default valid range for temperature
     */
    public void setDeaultTemperatureRange() {
        this.validMin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_VALUE;
        this.validMax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_VALUE;
    }
    
    /**
     * Set default valid range for temperature
     */
    public void setDeaultTemperatureAnomalyRange() {
        this.validMin = GeoGridConstant.GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE;
        this.validMax = GeoGridConstant.GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE;
    }
    
    /**
     * read time series data with a dimension of (nt, 72, 144) from data file
     * @param fileNetcdf, netcdf file
     * @param varName, variable name
     * @return  data for the specified variable
     */
    public static double[][][] readGeo2dTimeSeriesDouble(String fileNetcdf, String varName) {
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayDouble.D3 darr3 = null;
        int[] shape = new int[3];
        int[] origin = new int[3];
        double [][][] vdata = null;
        int dim1, dim2, dim3;
        int i, j, k;
        int mt = 0;

        for (i = 0; i < 3; i++) {
            origin[i] = 0;
        }

        try {
            fileNc = NetcdfFile.open(fileNetcdf);
                        
            dataVar = fileNc.findVariable(varName);
            if (dataVar == null) {
//                System.out.println("failed to find variable BT_mean from file " + filenameL3);
                System.out.println("failed to find variable from file " + fileNetcdf);
                errLogger.error("failed to find variable from file " + fileNetcdf);
                vdata = null;
            } else {
                shape = dataVar.getShape();
                darr3 = (ArrayDouble.D3) dataVar.read(origin, shape);
                //System.out.println("Data dimension:" + shape[0] + "," + shape[1]+ "," + shape[2]);
                //System.out.println("Data dimension:" + darr3.shapeToString());
                mt = shape[0];
                vdata = new double[mt][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];                
                for (k = 0; k < mt; k++) {
                    for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                        for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                            vdata[k][i][j] = darr3.get(k, i, j);
                        }
                    }
                }
            }
            
        } catch (Exception e) {
            e.printStackTrace();
            errLogger.error("failed to read " + fileNetcdf + ", filled with invalid values");
            vdata = null;            
        }
        return vdata;           
    }
    
    /**
     * read time series data with a dimension of (nt, 72, 144) from data file
     * @param fileNetcdf, netcdf file
     * @param varName, variable name
     * @return  data for the specified variable
     */
    public static double[][][] readGeo2dTimeSeriesFloat(String fileNetcdf, String varName) {
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayFloat.D3 darr3 = null;
        int[] shape = new int[3];
        int[] origin = new int[3];
        double [][][] vdata = null;
        int dim1, dim2, dim3;
        int i, j, k;
        int mt = 0;

        for (i = 0; i < 3; i++) {
            origin[i] = 0;
        }

        try {
            fileNc = NetcdfFile.open(fileNetcdf);
                        
            dataVar = fileNc.findVariable(varName);
            if (dataVar == null) {
//                System.out.println("failed to find variable BT_mean from file " + filenameL3);
                System.out.println("failed to find variable from file " + fileNetcdf);
                errLogger.error("failed to find variable from file " + fileNetcdf);
                vdata = null;
            } else {
                shape = dataVar.getShape();
                darr3 = (ArrayFloat.D3) dataVar.read(origin, shape);
                //System.out.println("Data dimension:" + shape[0] + "," + shape[1]+ "," + shape[2]);
                //System.out.println("Data dimension:" + darr3.shapeToString());
                mt = shape[0];
                vdata = new double[mt][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];                                                
                for (k = 0; k < mt; k++) {
                    for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                        for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                            vdata[k][i][j] = darr3.get(k, i, j);
                        }
                    }
                }
            }
            
        } catch (Exception e) {
            e.printStackTrace();
            errLogger.error("failed to read " + fileNetcdf + ", filled with invalid values");
            vdata = null;            
        }
        return vdata;           
    }
    
    /** 
     * Testing only
     * @param args 
     */
    public static void main(String[] args) {
        
        String fileNc="D:\\work2024\\Projects\\CDRNextGen\\package\\data\\L4\\AMSUA_L4_aqua_ch05.nc";
        String varName = "bt_anom";
        Geo2dTimeSeries g2t = new Geo2dTimeSeries(fileNc, varName);
        double vmin=-50;
        double vmax=50;
        g2t.setValidRange(vmin, vmax);
        double[] ft = g2t.getFt();
        int[] fyear = g2t.getFyear();
        int[] fmonth = g2t.getFmonth();
        double[] ym = g2t.getGlobalMean();
        int nt = fyear.length;
        for (int k = 0; k < nt; k++) {
            String str = String.format("%4d, %2d, %16.10f, %16.10f", fyear[k], fmonth[k], ft[k], ym[k]);
            System.out.println(str);
        }

    }
   
}
