/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.geogrid.GeoGrid2D.java
*
* NAME
*   atms.geogrid.GeoGrid2D.java
*
* PURPOSE
*   To build geodata for general processing
* 
* DESCRIPTION
*	This class is used to process geo-gridded data
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package geogrid;

import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;
import static geogrid.GeoGridConstant.GEO_INVALID_FILL_VALUE;

/**
 * 2D gridded geographic data
 *
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class GeoGrid2D {

    /**
     * Starting latitude of the grid.
     */
    final public static double GEO_GRID_LATITUDE_START = -90.0d;

    /**
     * End latitude of the grid.
     */
    final public static double GEO_GRID_LATITUDE_END = 90.0d;

    /**
     * Starting longitude of the grid.
     */
    final public static double GEO_GRID_LONGITUDE_START = -180.0f;

    /**
     * End longitude of the grid.
     */
    final public static float GEO_GRID_LONGITUDE_END = 180.0f;

    /**
     * Step of latitude for gridding.
     */
    final public static float GEO_GRID_STEP_LATITUDE = 2.5f;

    /**
     * Step of longitude for gridding.
     */
    final public static float GEO_GRID_STEP_LONGITUDE = 2.5f;
    
    /**
     * data
     */
    private double[][] data;
    
    /**
     * Valid min
     */
    private double validMin = Double.MIN_VALUE;
    
    /**
     * Valid max
     */
    private double validMax = Double.MAX_VALUE;
    
    /**
     *  Filled value
     */
    private double invalidFill = GEO_INVALID_FILL_VALUE; // default filled value for invalid data
    
    /**
     * Land/sea mask
     */
    private int[][] landseaMask = null;
    
    /**
     * Start index for global mean
     */
    private int index_valid_lat_start = 0;

    /**
     * End index for global mean
     */
    private int index_valid_lat_end = GEO_GRID_NUM_LAT - 1;

    /**
     * Constructor
     */
    public GeoGrid2D() {
        this.data = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        // Initialize with NaN values by default
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            for (int j = 0; j < GEO_GRID_NUM_LON; j++) {
                this.data[i][j] = this.invalidFill;
            }
        }
    }

    /**
     * Constructor.
     * @param v, 2d array
     */
    public GeoGrid2D(double[][] v) {
        this(v, Double.MIN_VALUE, Double.MAX_VALUE, Double.NaN);
    }

    /**
     * Constructor.
     * @param v, 2d array
     */
    public GeoGrid2D(float[][] v) {
        this(v, Float.MIN_VALUE, Float.MAX_VALUE, Float.NaN);
    }

    /**
     * Constructor
     *
     * @param v
     * @param vmin
     * @param vmax
     */
    public GeoGrid2D(double[][] v, double vmin, double vmax) {
        this(v, vmin, vmax, GEO_INVALID_FILL_VALUE);
    }

    /**
     *  Constructor
     * @param v, 2D data
     * @param vmin, valid min
     * @param vmax, valid max 
     */
    public GeoGrid2D(float[][] v, float vmin, float vmax) {
        this(v, vmin, vmax, (float) GEO_INVALID_FILL_VALUE);
    }

    /**
     *  Constructor
     * @param v, 2D data
     * @param vmin, valid min
     * @param vmax, valid max 
     * @param invalidValue, invalid value
     */
    public GeoGrid2D(double[][] v, double vmin, double vmax, double invalidValue) {
        double vtmp;
        this.validMin = vmin;
        this.validMax = vmax;
        this.invalidFill = invalidValue;

        this.data = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        // Initialize with NaN values by default
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            for (int j = 0; j < GEO_GRID_NUM_LON; j++) {
                vtmp = v[i][j];
                if ((vtmp >= vmin) & (vtmp <= vmax)) {
                    this.data[i][j] = v[i][j];
                } else {
                    this.data[i][j] = this.invalidFill;
                }
            }
        }
    }

    /**
     *  Constructor
     * @param v, 2D data
     * @param vmin, valid min
     * @param vmax, valid max 
     */ 
    public GeoGrid2D(float[][] v, float vmin, float vmax, float invalidValue) {
        double vtmp;
        this.validMin = vmin;
        this.validMax = vmax;
        this.invalidFill = invalidValue;

        this.data = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        // Initialize with NaN values by default
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            for (int j = 0; j < GEO_GRID_NUM_LON; j++) {
                vtmp = v[i][j];
                if ((vtmp >= vmin) & (vtmp <= vmax)) {
                    this.data[i][j] = v[i][j];
                } else {
                    this.data[i][j] = this.invalidFill;
                }
            }
        }
    }

    /**
     * Set land/sea mask for geo-grid
     *
     * @param mask
     */
    public void setLandseaMask(int[][] mask) {
        this.landseaMask = new int[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            for (int j = 0; j < GEO_GRID_NUM_LON; j++) {
                this.landseaMask[i][j] = mask[i][j];
            }
        }
    }

    /**
     * Set valid range
     *
     * @param vmin, valid minimum value
     * @param vmax, valid maximum value
     */
    public void setValidRange(double vmin, double vmax) {
        this.validMin = vmin;
        this.validMax = vmax;
    }

    /**
     * Set valid range
     *
     * @param vmin, valid minum value
     * @param vmax, valid maximum value
     */
    public void setValidRange(float vmin, float vmax) {
        this.validMin = vmin;
        this.validMax = vmax;
    }

    /**
     * Setup valid index for global mean
     *
     * @param index1, starting index
     * @param index2, ending index
     */
    public void setValidLatIndex(int index1, int index2) {
        this.index_valid_lat_start = index1;
        this.index_valid_lat_end = index2;
    }

    /**
     * Get global mean
     *
     * @return global area-weighted mean if there are valid data, NaN otherwise.
     */
    public double getGlobalMean() {
        double vmean = 0.0d; //
        int i, j;
        double sumValue = 0.0d;
        double sumWt = 0.0d;
        double vwt, vdata;
        int numPix = 0;

        double[][] wt = GeoGridConstant.getAreaWeight();

        for (i = this.index_valid_lat_start; i <= this.index_valid_lat_end; i++) {
            for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                vwt = wt[i][j];
                vdata = this.data[i][j];
                if ((vdata >= this.validMin) & (vdata <= this.validMax)) {
                    numPix++;
                    sumWt = sumWt + vwt;
                    sumValue = sumValue + vdata*vwt;
                }
            }
        }

        if (numPix > 0) {
            vmean = sumValue / sumWt;
        } else {
            vmean = Double.NaN;
        }

        return vmean;
    }

    /**
     * Get global land mean
     *
     * @return global area-weighted mean if there are valid data, NaN otherwise.
     */
    public double getGlobalLandMean() {
        double vmean = 0.0d; //
        int i, j;
        double sumValue = 0.0d;
        double sumWt = 0.0d;
        double vwt, vdata;
        int numPix = 0;

        if (this.landseaMask == null) {
            // didn't set land/sea mask, return NaN
            vmean = Double.NaN;
        } else {
            double[][] wt = GeoGridConstant.getAreaWeight();
            for (i = this.index_valid_lat_start; i <= this.index_valid_lat_end; i++) {
                for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                    vwt = wt[i][j];
                    vdata = this.data[i][j];
                    if ((vdata >= this.validMin) & (vdata <= this.validMax) & (this.landseaMask[i][j] == GeoGridConstant.LCMASK_LAND)) {
                        numPix++;
                        sumWt = sumWt + vwt;
                        sumValue = sumValue + vdata*vwt;
                    }
                }
            }
            if (numPix > 0) {
                vmean = sumValue / sumWt;
            } else {
                vmean = Double.NaN;
            }
        }
        return vmean;
    }

    /**
     * Get global ocean mean
     *
     * @return global area-weighted mean if there are valid data, NaN otherwise.
     */
    public double getGlobalOceanMean() {
        double vmean = 0.0d; //
        int i, j;
        double sumValue = 0.0d;
        double sumWt = 0.0d;
        double vwt, vdata;
        int numPix = 0;

        if (this.landseaMask == null) {
            // didn't set land/sea mask, return NaN
            vmean = Double.NaN;
        } else {
            double[][] wt = GeoGridConstant.getAreaWeight();
            for (i = this.index_valid_lat_start; i <= this.index_valid_lat_end; i++) {
                for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                    vwt = wt[i][j];
                    vdata = this.data[i][j];
                    if ((vdata >= this.validMin) & (vdata <= this.validMax) & (this.landseaMask[i][j] == GeoGridConstant.LCMASK_OCEAN)) {
                        numPix++;
                        sumWt = sumWt + vwt;
                        sumValue = sumValue + vdata*vwt;
                    }
                }
            }
            if (numPix > 0) {
                vmean = sumValue / sumWt;
            } else {
                vmean = Double.NaN;
            }
        }
        return vmean;
    }

    /**
     * Get global ocean mean
     *
     * @return global area-weighted mean if there are valid data, NaN otherwise.
     */
    public double[] getGlobalMeanAll() {
        double[] vmean = new double[3]; //0: ocean, 1:land, 2:global
        int i, j, k;
        double[] sumValue = new double[3];
        double[] sumWt = new double[3];
        double vwt, vdata;
        int[] numPix = new int[3];
        int lcmask = 0;

        for(i=0; i<3; i++) {
            vmean[i] = Double.NaN;
            sumValue[i] = 0.0d;
            sumWt[i] = 0.0d;
            numPix[i] = 0;
        }
        
        if (this.landseaMask == null) {
            // didn't set land/sea mask, return NaN      
        } else {
            double[][] wt = GeoGridConstant.getAreaWeight();
            for (i = this.index_valid_lat_start; i <= this.index_valid_lat_end; i++) {
                for (j = 0; j < GEO_GRID_NUM_LON; j++) {
                    vwt = wt[i][j];
                    vdata = this.data[i][j];
                    lcmask = this.landseaMask[i][j];
                    if ((vdata >= this.validMin) & (vdata <= this.validMax)) {
                        //ocean
                        k = GeoGridConstant.INDEX_LANDSEA_MASK_OCEAN;
                        if (lcmask == GeoGridConstant.LCMASK_OCEAN ) {
                            numPix[k]++;
                            sumWt[k] = sumWt[k] + vwt;
                            sumValue[k] = sumValue[k] + vdata*vwt;
                        } 
                        
                        //land
                        k = GeoGridConstant.INDEX_LANDSEA_MASK_LAND;
                        if (lcmask == GeoGridConstant.LCMASK_LAND ) {
                            numPix[k]++;
                            sumWt[k] = sumWt[k] + vwt;
                            sumValue[k] = sumValue[k] + vdata*vwt;
                        } 
                        
                        // land and ocean
                        k = GeoGridConstant.INDEX_LANDSEA_MASK_GLOBAL;
                        numPix[k]++;
                        sumWt[k] = sumWt[k] + vwt;
                        sumValue[k] = sumValue[k] + vdata*vwt;                        
                    }
                }
            }
            
            for(k=0; k<3; k++) {
                if (numPix[k] > 0) {
                    vmean[k] = sumValue[k] / sumWt[k];
                } else {
                    vmean[k] = Double.NaN;
                }
            }
        }
        return vmean;
    }

    /**
     * Set data 
     * @param vdata, 2D gridded data 
     */
    public void setData(double[][] vdata){
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            for (int j = 0; j < GEO_GRID_NUM_LON; j++) {
                this.data[i][j] = vdata[i][j];
            }
        }        
    }
    
    /**
     * Set data 
     * @param vdata, 2D gridded data 
     */
    public void setData(float[][] vdata){
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            for (int j = 0; j < GEO_GRID_NUM_LON; j++) {
                this.data[i][j] = vdata[i][j];
            }
        }        
    }
    
    /**
     * Get grid latitude in float
     * @return grid latitude
     */
    public static float[] getGridLatitudeFloat(){
        double vtmp = 0.0d;
        float[] plat = new float[GEO_GRID_NUM_LAT];
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            vtmp = GEO_GRID_LATITUDE_START + (i+0.5)*GEO_GRID_STEP_LATITUDE;
            plat[i] = (float) vtmp;
        }
        return plat;
    }

    /**
     * Get grid latitude in double
     * @return grid latitude
     */
    public static double[] getGridLatitude(){
        double[] plat = new double[GEO_GRID_NUM_LAT];
        for (int i = 0; i < GEO_GRID_NUM_LAT; i++) {
            plat[i] = GEO_GRID_LATITUDE_START + (i+0.5)*GEO_GRID_STEP_LATITUDE;
        }
        return plat;
    }
    
    /**
     * Get grid latitude in float
     * @return grid longitude
     */
    public static float[] getGridLongitudeFloat(){
        double vtmp =0.0d;
        float[] plon = new float[GEO_GRID_NUM_LON];
        for (int i = 0; i < GEO_GRID_NUM_LON; i++) {
                vtmp = GEO_GRID_LONGITUDE_START + (i+0.5)*GEO_GRID_STEP_LONGITUDE;
                plon[i] = (float)vtmp;
        }
        return plon;
    }

    /**
     * Get grid latitude in double
     * @return grid longitude
     */
    public static double[] getGridLongitude(){
        double[] plon = new double[GEO_GRID_NUM_LON];
        for (int i = 0; i < GEO_GRID_NUM_LON; i++) {
                plon[i] = GEO_GRID_LONGITUDE_START + (i+0.5)*GEO_GRID_STEP_LONGITUDE;
        }
        return plon;
    }
    
    // testing only
   public static void main(String[] args) {
       int i,j;
       double[][] z = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];

       String filename="C:\\Users\\xhao\\Documents\\MATLAB\\src\\etc\\landfrac.nc";
       int[][] mask=GeoGridConstant.getLandSeaMask(filename);
       
       for(i=0; i<GEO_GRID_NUM_LAT; i++) {
           for(j=0; j<GEO_GRID_NUM_LON; j++) {
               if (mask[i][j]==0) {
                   z[i][j] = 0.0d;
               } else {
                   z[i][j] = 1.0d;                   
               }
           }
       }
       
       GeoGrid2D vd2 = new GeoGrid2D(z);
       vd2.setLandseaMask(mask);
       vd2.setValidRange(-10d, 10d);
               
       double[] vm1 = new double[3];
       double[] vm2 = vd2.getGlobalMeanAll();
       vm1[0] = vd2.getGlobalOceanMean();
       vm1[1] = vd2.getGlobalLandMean();
       vm1[2] = vd2.getGlobalMean();
       
       for (int k=0; k<3; k++){
           String str = String.format("Global mean: %10.5f, %10.5f, %10.5f", vm1[k], vm2[k], vm1[k] - vm2[k]); 
           System.out.println(str);                   
       }
   }    
    
}

