/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.geogrid.GeoGridConstant.java
*
* NAME
*   atms.geogrid.GeoGridConstant.java
*
* PURPOSE
*   To define Constants for CDR geographic grid 
* 
* DESCRIPTION
*	This class is used to define Constants for CDR geographic grid 
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package geogrid;

import ucar.ma2.ArrayDouble;
import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 * Constants for CDR geographic grid 
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class GeoGridConstant {

   /**
    * Logger for errors
    */
   final static private Logger errLogger = LogManager.getLogger("ErrorLog");
    
   /**
    * Number of grids in latitude
    */
   final public static int GEO_GRID_NUM_LAT = 72;

   /**
    * Number of grids in longitude
    */
   final public static int GEO_GRID_NUM_LON = 144;

   /**
    * Starting latitude for grid
    */
   final public static double GEO_GRID_LATITUDE_START = -90d; 

   /**
    * Ending latitude for grid
    */
   final public static double GEO_GRID_LATITUDE_END = 90d; 

   /**
    * Latitude grid size
    */
   final public static double GEO_GRID_LATITUDE_STEP = 2.5d; 

   /**
    * Starting longitude for grid
    */
   final public static double GEO_GRID_LONGITUDE_START = -180d; 

   /**
    * Ending longitude for grid
    */
   final public static double GEO_GRID_LONGITUDE_END = 180d; 

   /**
    * Longitude grid size
    */
   final public static double GEO_GRID_LONGITUDE_STEP = 2.5d; 

   /**
    * Filled value for invalid data
    */
   final public static double GEO_INVALID_FILL_VALUE = -9999.0d; 
   
   /**
    * Valid minimum temperature
    */
   final public static double GEO_MINIMUM_TEMPERATURE_VALUE = 150.0d; 
   /**
    * Valid maximum temperature
    */
   final public static double GEO_MAXIMUM_TEMPERATURE_VALUE = 350.0d; 

   /**
    * Valid minimum temperature anomaly
    */
   final public static double GEO_MINIMUM_TEMPERATURE_ANOMALY_VALUE = -50.0d; 

   /**
    * Valid maximum temperature anomaly
    */
   final public static double GEO_MAXIMUM_TEMPERATURE_ANOMALY_VALUE = 50.0d; 

   /**
    * Starting latitude index for global mean
    */
   final public static int GEO_INDEX_GLOBAL_MEAN_LATITUDE_START = 3;

   /**
    * Ending latitude index for global mean
    */
   final public static int GEO_INDEX_GLOBAL_MEAN_LATITUDE_END = 68;
   
   /**
    * Mask index value for ocean
    */
   final public static int INDEX_LANDSEA_MASK_OCEAN = 0;

   /**
    * Mask index value for land
    */
   final public static int INDEX_LANDSEA_MASK_LAND = 1;

   /**
    * Mask index value to include both land and ocean
    */
   final public static int INDEX_LANDSEA_MASK_GLOBAL = 2;
   
   /**
    * Index for all masks
    */
   final public static int[] INDEX_LANDSEA_MASK_ALL = {0,1,2};
   
   /**
    * Mask value for ocean
    */
   final public static int LCMASK_OCEAN = 0;
   
   /**
    * Mask value for land
    */
   final public static int LCMASK_LAND = 1;

   /**
    * Get grid area weight
    * @return area weight
    */
   public static double[][] getAreaWeight(){
       double[][] wt = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        int k1, k2;
        double tmpv = 0.0d;
        double tmplat = 0.0d;
        double tmplat1 = 0.0d;
        double tmplat2 = 0.0d;
        double sc = Math.PI/180.0d;
        double grid_minimum_lat = -90.0d;
        double grid_step_lat = -2.5d;

        for(k1=0; k1< GEO_GRID_NUM_LAT; k1++) {
            tmplat1 = (GEO_GRID_LATITUDE_START + k1*GEO_GRID_LATITUDE_STEP)*sc;   
            tmplat2 = (GEO_GRID_LATITUDE_START + (k1+1)*GEO_GRID_LATITUDE_STEP)*sc;   
            tmpv = Math.abs(Math.sin(tmplat2)-Math.sin(tmplat1));
            for(k2=0; k2<GEO_GRID_NUM_LON; k2++){
                wt[k1][k2] = tmpv;
            }
        }                
        return wt;
   }
   
   /**
    * Get land/sea mask
    * @return land/sea mask, 0: ocean, 1: land 
    */
   public static int[][] getLandSeaMask(String landfracFilename) {
       int[][] mask = new int[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayDouble.D2 landfrac = null;
        int[] shape = new int[2];
        int[] origin = new int[2];
        int dim1, dim2;
        int i, j;
        
        for(i=0; i<2; i++){
            origin[i] = 0;
        }
       
        try{
            fileNc = NetcdfFile.open(landfracFilename);
            dataVar = fileNc.findVariable("landfrac");
            if (dataVar==null) {
                errLogger.error("failed to read landfrac data");            
            } else {
                shape = dataVar.getShape();
                landfrac = (ArrayDouble.D2) dataVar.read(origin, shape);   
                //System.out.println("Landfrac dimension:" + shape[0] + "," + shape[1]);
                for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                    for(j=0; j<GEO_GRID_NUM_LON; j++) {
                        if (landfrac.get(i,j) > 0.5d) { // land
                            mask[i][j] = LCMASK_LAND;
                        } else {  // ocean
                            mask[i][j] = LCMASK_OCEAN;
                        }
                    }
                }
            }
            
        }catch (Exception e) {
            errLogger.error("failed to read landfrac data");
            e.printStackTrace();
        }
       return mask;
   } 

   /**
    * Get land/sea mask
    * @return land/sea mask, 0: ocean, 1: land 
    */
   public static double[][] getLandfrac(String landfracFilename) {
        double[][] landfrac = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayDouble.D2 darr2 = null;
        int[] shape = new int[2];
        int[] origin = new int[2];
        int dim1, dim2;
        int i, j;
        
        for(i=0; i<2; i++){
            origin[i] = 0;
        }
       
        try{
            fileNc = NetcdfFile.open(landfracFilename);
            dataVar = fileNc.findVariable("landfrac");
            if (dataVar==null) {
                errLogger.error("failed to read landfrac data");            
            } else {
                shape = dataVar.getShape();
                darr2 = (ArrayDouble.D2) dataVar.read(origin, shape);   
                //System.out.println("Landfrac dimension:" + shape[0] + "," + shape[1]);
                for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                    for(j=0; j<GEO_GRID_NUM_LON; j++) {
                        landfrac[i][j] = darr2.get(i,j);
                    }
                }
            }
            
        }catch (Exception e) {
            errLogger.error("failed to read landfrac data");
            e.printStackTrace();
        }
       return landfrac;
   } 
   
   /**
    * Load area weight from file
    * @param filename
    * @return area weight of grid 
    */
   public static double[][] loadGridAreaWeight(String filename) {
        double[][] wt = new double[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        NetcdfFile fileNc = null;
        Variable dataVar = null;
        ArrayDouble.D2 darr2 = null;
        int[] shape = new int[2];
        int[] origin = new int[2];
        int dim1, dim2;
        int i, j;
        
        for(i=0; i<2; i++){
            origin[i] = 0;
        }
       
        try{
            fileNc = NetcdfFile.open(filename);
            dataVar = fileNc.findVariable("wt");
            if (dataVar==null) {
                errLogger.error("failed to read weight data");            
            } else {
                shape = dataVar.getShape();
                darr2 = (ArrayDouble.D2) dataVar.read(origin, shape);   
                //System.out.println("Landfrac dimension:" + shape[0] + "," + shape[1]);
                for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                    for(j=0; j<GEO_GRID_NUM_LON; j++) {
                        wt[i][j] = darr2.get(i,j);
                    }
                }
            }
            
        }catch (Exception e) {
            errLogger.error("failed to read landfrac data");
            e.printStackTrace();
        }
       return wt;       
   }
      
   
}
