/*  Robodoc Markup Header
*@***h* MLT_CDR_V5/atms.geogrid.GeoTimeseriesMerger.java
*
* NAME
*   atms.geogrid.GeoTimeseriesMerger.java
*
* PURPOSE
*   To merge multiple time series 
* 
* DESCRIPTION
*	This class is used to merge time series of geo-gridded data
*
* AUTHOR
*   Xianjun Hao, George Mason University 
*
* COPYRIGHT 
*   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
*   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
*   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
*   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
*   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
*   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
*   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
*   SUPPORT TO USERS.
*
* REVISION HISTORY
*   10/16/2023 Initial Version prepared for NCEI
*
*@*****
 */
package geogrid;

import geogrid.GeoGridConstant;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LAT;
import static geogrid.GeoGridConstant.GEO_GRID_NUM_LON;
 

/**
 * Utility class to merge geo-gridded data time series 
 * @author <Xianjun Hao <xhao1@gmu.edu>
 */
public class GeoTimeseriesMerger {


    /**
     * Start year
     */
    private int year1;
    
    /**
     * Start month
     */
    private int month1;
    
    /**
     * Ending year
     */
    private int year2;
    
    /**
     * Ending month
     */
    private int month2;
    
    /**
     * Number of months
     */
    private int numberMonths;
    
    /**
     * List of 
     */
    /**
     * List of years for data
     */
    private int[] fyear;

    /**
     * List of months for data
     */
    private int[] fmonth;
    
    /**
     *  Time for data
     */
    private double[] ft;
    
    /**
     * Mean value
     */
    double[][][] vmean;
    
    /**
     * Sum of data
     */
    double[][][] vsum;
    
    /**
     *  Pixel number
     */
    int[][][] pct;
    
    /**
     * Valid min
     */
    private double validMin;
    
    /**
     * Valid Max
     */
    private double validMax;
    
    /**
     * Land/sea mask
     */
    private int[][] mask = null;
    
    /**
     * Initialize a time series for merging
     * @param year1
     * @param month1
     * @param year2
     * @param month2 
     */
    public GeoTimeseriesMerger(int year1, int month1, int year2, int month2, double vmin, double vmax) {
        this.year1 = year1;
        this.month1 = month1;
        this.year2 = year2;
        this.month2 = month2;
        this.numberMonths = (year2 - year1)*12 + (month2-month1+1);
        this.validMin = vmin;
        this.validMax = vmax;
        this.vmean = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.vsum = new double[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.pct = new int[this.numberMonths][GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON];
        this.initTime();
        this.initData();
    }
            
    /**
     * Initialize the merger
     */
    public void initData(){
        for(int k=0; k<this.numberMonths; k++) {
            for(int i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(int j=0; j<GEO_GRID_NUM_LON; j++) {
                    this.vmean[k][i][j] = 0.0d;
                    this.vsum[k][i][j] = 0.0d;
                    this.pct[k][i][j] = 0;
                }
            }
        }        
    }

    /**
     * Initialize time 
     */
    private void initTime(){
        int i, j , k;
        int  km, kyear, kmonth;
        this.fyear = new int[this.numberMonths];
        this.fmonth = new int[this.numberMonths];
        this.ft = new double[this.numberMonths];
                
        for(i=0; i<this.numberMonths; i++){
            km = i + this.month1;
            kmonth = km%12;
            if (kmonth==0) {
                kyear = this.year1 + (km-1)/12;
                kmonth=12;
            } else {
                kyear = this.year1 + km/12;
            }
            this.fyear[i] = kyear;
            this.fmonth[i] = kmonth;
            this.ft[i] = (double)(this.fyear[i]) + ( (double)(this.fmonth[i]) - 0.5)/12.0d;
        }
    }

    
    /**
     * Get time of the series
     * @return time of the series
     */
    public double[] getFt(){
        return this.ft;
    }
    
    /**
     * Get years of the time series
     * @return years of the time series
     */
    public int[] getFyear(){
       return this.fyear; 
    }
    
    /**
     * Get months of the time series
     * @return months of the time series
     */
    public int[] getFmonth(){
        return this.fmonth;
    }
    
    /**
     * Get mean data after merging.
     * @return 
     */
    public double[][][] getMean(){
        return this.vmean;
    }

    /**
     * Set valid range
     *
     * @param vmin, valid minimum value
     * @param vmax, valid maximum value
     */
    public void setValidRange(double vmin, double vmax) {
        this.validMin = vmin;
        this.validMax = vmax;
    }
    
/**
 * add data for merging
 * @param tyear, year list of the data
 * @param tmonth, month list of the data
 * @param vdata, data time series 
 */
    public void addData(int[] tyear, int[] tmonth, double[][][] vdata){
        double vmin, vmax, vtmp;
        int i,j,k, km, kmonth, kyear;
        int mt = tyear.length;
        for(k=0; k<mt; k++){
            kyear = tyear[k];
            kmonth = tmonth[k];
            km = (kyear - this.year1)*12 + (kmonth - this.month1); // time index
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    vtmp = vdata[k][i][j];
                    if ((vtmp >=validMin) &(vtmp <= validMax)) {
                        this.vsum[km][i][j] = this.vsum[km][i][j] + vtmp; 
                        this.pct[km][i][j] = this.pct[km][i][j] + 1; 
                    }
                }
            }
        }     
    }

    /**
     * Merge data to get the mean values
     * @return  merged data time series
     */
    public double[][][] mergeData() {
        int k, i, j;
        int mp;
        
        for(k=0; k<this.numberMonths; k++){
            for(i=0; i<GEO_GRID_NUM_LAT; i++) {
                for(j=0; j<GEO_GRID_NUM_LON; j++) {
                    mp =this.pct[k][i][j];
                    if (mp>0) {
                        this.vmean[k][i][j] = (this.vsum[k][i][j])/( (double)mp); 
                    } else {
                        this.vmean[k][i][j] =  GeoGridConstant.GEO_INVALID_FILL_VALUE;
                    }
                }
            }
        }    
        
        
        return this.vmean;
    }
    
    
    /**
     * Return sum for testing
     * @return sum at grid
     */
    public double[][][] getSum(){
        return this.vsum;
    }
    
    /**
     * Get pixel number at each grid 
     * @return pixel number
     */
    public int[][][] getPixelCount(){
        return this.pct;
    }
    /**
     * Set land/sea mask
     * @param landMask  land/sea mask, 0 for ocean, 1 for land.
     */
    public void setLandseaMask(int[][] landMask) {
       int i, j;
       this.mask = new int[GEO_GRID_NUM_LAT][GEO_GRID_NUM_LON]; 
       for(i=0; i<GEO_GRID_NUM_LAT; i++) {
           for(j=0; j<GEO_GRID_NUM_LON; j++) {
               this.mask[i][j] = landMask[i][j];
           }
       }
    }            
    
    /**
     * Get constant bias of two data time series at each latitude
     * @param gt1, data time series 1
     * @param gt2, data time series 2
     * @return constant bias at each latitude
     */
    public static double[] getConstBiasByLatitude(Geo2dTimeSeries gt1, Geo2dTimeSeries gt2){
        int k, i, j, k1, k2;
        int cyear1, cmonth1, cyear2, cmonth2;
        double ct1, ct2;
        double vtmp1, vtmp2, vtmp;
        double[] bias = new double[GEO_GRID_NUM_LAT];
        double vmin = gt1.getValidMin();
        double vmax = gt1.getValidMax();
        
        int[] fyear1 = gt1.getFyear();
        int[] fmonth1 = gt1.getFmonth();
        int[] fyear2 = gt2.getFyear();
        int[] fmonth2 = gt2.getFmonth();
        double[] ft1 = gt1.getFt();
        double[] ft2 = gt2.getFt();
        int nt1= ft1.length;
        int nt2 = ft2.length;
        
        ct1 = ft1[0];
        if (ct1 < ft2[0]) {
           ct1 = ft2[0];            
        }
        ct2 = ft1[nt1-1];
        if (ct2 > ft2[nt2-1]) {
            ct2 = ft2[nt2-1];
        }
        
        cyear1 = (int)Math.floor(ct1);
        cmonth1 = (int) Math.floor( (ct1 - cyear1)*12) +1;
        cyear2 = (int)Math.floor(ct2);
        cmonth2 = (int) Math.floor( (ct2 - cyear2)*12) +1;
        
        int mt = (cyear2 - cyear1 )*12 + (cmonth2 - cmonth1 +1);
        
        int indg1 = (cyear1 - fyear1[0])*12 + (cmonth1 - fmonth1[0]);
        int indg2 = (cyear1 - fyear2[0])*12 + (cmonth1 - fmonth2[0]);
        
        int[] pix = new int[GEO_GRID_NUM_LAT];
       
        for(k=0; k<GEO_GRID_NUM_LAT; k++) {
            pix[k] = 0;
            bias[k] = 0;
        }
        
        double[][][] data1 = gt1.getData();
        double[][][] data2 = gt2.getData();
        
        
        for (k=0; k<mt; k++){
           k1 = indg1 + k;
           k2 = indg2 + k;
           for(i=0; i<GEO_GRID_NUM_LAT; i++) {
               for(j=0; j<GEO_GRID_NUM_LON; j++) {
                   vtmp1 = data1[k1][i][j];
                   vtmp2 = data2[k2][i][j];
                   if ( (vtmp1>=vmin) &(vtmp1<=vmax) & (vtmp2>=vmin) & (vtmp2<=vmax)) {
                       pix[i] ++;
                       bias[i] = bias[i] + (vtmp1-vtmp2);
                   }
               }
           }
        }
        
        for(i=0; i<GEO_GRID_NUM_LAT; i++) {
            vtmp1 = bias[i];
            if (pix[i]>0) {
                bias[i] = vtmp1 / pix[i];
            } else {
                bias[i] = 0.0d; //GeoGridConstant.GEO_INVALID_FILL_VALUE;
            }
        }
        
        return bias;
    }
    
    /**
     * Get constant bias of two data time series at each latitude
     * @param gt1, data time series 1
     * @param gt2, data time series 2
     * @param mask, land/sea mask, 0 for ocean and 1 for land
     * @return constant bias at each latitude for ocean, land and both
     */
    public static double[][] getAllConstBiasByLatitude(Geo2dTimeSeries gt1, Geo2dTimeSeries gt2, int[][]mask){
        int k, i, j, k1, k2, ks;
        int cyear1, cmonth1, cyear2, cmonth2;
        double ct1, ct2;
        double vtmp1, vtmp2, vtmp;
        double[][] bias = new double[3][GEO_GRID_NUM_LAT];
        double vmin = gt1.getValidMin();
        double vmax = gt1.getValidMax();
        
        int[] fyear1 = gt1.getFyear();
        int[] fmonth1 = gt1.getFmonth();
        int[] fyear2 = gt2.getFyear();
        int[] fmonth2 = gt2.getFmonth();
        double[] ft1 = gt1.getFt();
        double[] ft2 = gt2.getFt();
        int nt1= ft1.length;
        int nt2 = ft2.length;
        
        ct1 = ft1[0];
        if (ct1 < ft2[0]) {
           ct1 = ft2[0];            
        }
        ct2 = ft1[nt1-1];
        if (ct2 > ft2[nt2-1]) {
            ct2 = ft2[nt2-1];
        }
        
        cyear1 = (int)Math.floor(ct1);
        cmonth1 = (int) Math.floor( (ct1 - cyear1)*12) +1;
        cyear2 = (int)Math.floor(ct2);
        cmonth2 = (int) Math.floor( (ct2 - cyear2)*12) +1;
        
        int mt = (cyear2 - cyear1 )*12 + (cmonth2 - cmonth1 +1);
        
        int indg1 = (cyear1 - fyear1[0])*12 + (cmonth1 - fmonth1[0]);
        int indg2 = (cyear1 - fyear2[0])*12 + (cmonth1 - fmonth2[0]);
        
        int[][] pix = new int[3][GEO_GRID_NUM_LAT];
 
        for (k = 0; k < 3; k++) {
            for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                pix[k][i] = 0;
                bias[k][i] = 0;
            }
        }
        
        
        double[][][] data1 = gt1.getData();
        double[][][] data2 = gt2.getData();
        
        
        for (k=0; k<mt; k++){
           k1 = indg1 + k;
           k2 = indg2 + k;
           for(i=0; i<GEO_GRID_NUM_LAT; i++) {
               for(j=0; j<GEO_GRID_NUM_LON; j++) {
                   vtmp1 = data1[k1][i][j];
                   vtmp2 = data2[k2][i][j];
                   if ( (vtmp1>=vmin) &(vtmp1<=vmax) & (vtmp2>=vmin) & (vtmp2<=vmax)) {
                       // for all
                       pix[2][i] ++;
                       bias[2][i] = bias[2][i] + (vtmp1-vtmp2);
                       // For locean or land
                       ks = mask[i][j];
                       pix[ks][i] ++;
                       bias[ks][i] = bias[ks][i] + (vtmp1-vtmp2);
                   }
               }
           }
        }

        for (k = 0; k < 3; k++) {
            for (i = 0; i < GEO_GRID_NUM_LAT; i++) {
                vtmp1 = bias[k][i];
                if (pix[k][i] > 0) {
                    bias[k][i] = vtmp1 / pix[k][i];
                } else {
                    bias[k][i] = 0.0d; // GeoGridConstant.GEO_INVALID_FILL_VALUE;
                }
            }
        }
        
        return bias;
    }
    
}

